// ----------------------------------
// Filename      : ModelerGUI.java
// Created       : 10/12/1999
// Last modified : 08/04/2003
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.modeler;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import mitrasim.exceptions.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class ModelerGUI extends Frame implements WindowListener, ActionListener, KeyListener
{
  // internal datastructures
  private Resources     fResources;
  private int           fWinXSize;
  private int           fWinYSize;
  private Messages      fMessages;
  private boolean       fDrawNetwork;
  private EditorCanvas  fEditorCanvas;
  private MessageCanvas fMessageCanvas;
  private int           fOriginXOffset;
  private int           fOriginYOffset;
  private float         fScaleFactor;
  private float         fZoomFactor;
  private int           fMapXOffset;
  private int           fMapYOffset;
  private String        fMapFilename;
  private int           fMapWidth;
  private int           fMapHeight;
  private boolean       fMapAvailable;
  private Image         fMap;
  private Network       fNetwork;
  private int           fLastNodeID;
  private short         fLastNodeType;
  private int           fLastLinkID;
  private MenuBar       fMenuBar;
  private Button[]      fControlButtons;
  private final int     kNrOfControlButtons = 10;
  private final int     kAddNodeButton      = 0;
  private final int     kAddLinkButton      = 1;
  private final int     kMoveLeftButton     = 2;
  private final int     kMoveRightButton    = 3;
  private final int     kZoomInButton       = 4;
  private final int     kEditNodeButton     = 5;
  private final int     kEditLinkButton     = 6;
  private final int     kMoveUpButton       = 7;
  private final int     kMoveDownButton     = 8;
  private final int     kZoomOutButton      = 9;
  private MLegend       fLegend;

  // the editor-canvas's draw-area
  private int cx1;
  private int cy1;
  private int cx2;
  private int cy2;
  private int cwidth;
  private int cheight;

  public ModelerGUI()
  {
    super();

    fResources = new Resources();
    fWinXSize = fResources.fnmXSize;
    fWinYSize = fResources.fnmYSize;
    fMessages = fResources.fMessages;

    fOriginXOffset = 0;
    fOriginYOffset = 0;
    fScaleFactor = 1.0f;
    fZoomFactor = 1.0f;
    fMapXOffset = 0;
    fMapYOffset = 0;
    fMapFilename = "";
    fMapWidth = 0;
    fMapHeight = 0;
    fMapAvailable = false;
    fMap = null;
    fResources.fNetwork = new Network(fMessages);
    fNetwork = fResources.fNetwork;
    fLastNodeID = -1;
    fLastNodeType = Node.kOrigin;
    fLastLinkID = -1;

    // adjust the windowlayout
    fDrawNetwork = false;
    setFont(new Font(fMessages.lookup("textFontName"),Font.PLAIN,14));
    setSize(fWinXSize,fWinYSize);
    setResizable(false);
    setLocation(0,0);
    setTitle(fMessages.lookup("textNetworkModelerWindowTitle"));
    setBackground(Color.black);
    setIconImage(Toolkit.getDefaultToolkit().getImage("M2000Icon.jpg"));

    setLayout(new BorderLayout());

    Panel CanvasPanel = new Panel();
    CanvasPanel.setLayout(new BorderLayout());

    // create and add the edit-canvas to the window
    fEditorCanvas = new EditorCanvas(fWinXSize,fWinYSize - 120);
    CanvasPanel.add("North",fEditorCanvas);
    fMessageCanvas = new MessageCanvas(fWinXSize,40);
    CanvasPanel.add("South",fMessageCanvas);
    add("North",CanvasPanel);

    // create panel containing editor-control commands
    Panel ControlPanel = new Panel();
    ControlPanel.setLayout(new GridLayout(2,5));

    fControlButtons = new Button[kNrOfControlButtons];

    fControlButtons[kAddNodeButton] = new Button(fMessages.lookup("buttonAddNode"));
    fControlButtons[kAddLinkButton] = new Button(fMessages.lookup("buttonAddLink"));
    fControlButtons[kMoveLeftButton] = new Button(fMessages.lookup("buttonMoveLeft"));
    fControlButtons[kMoveRightButton] = new Button(fMessages.lookup("buttonMoveRight"));
    fControlButtons[kZoomInButton] = new Button(fMessages.lookup("buttonZoomIn"));
    fControlButtons[kEditNodeButton] = new Button(fMessages.lookup("buttonEditNode"));
    fControlButtons[kEditLinkButton] = new Button(fMessages.lookup("buttonEditLink"));
    fControlButtons[kMoveDownButton] = new Button(fMessages.lookup("buttonMoveDown"));
    fControlButtons[kMoveUpButton] = new Button(fMessages.lookup("buttonMoveUp"));
    fControlButtons[kZoomOutButton] = new Button(fMessages.lookup("buttonZoomOut"));

    for (int i = 0; i < kNrOfControlButtons; ++i) {
      fControlButtons[i].addActionListener(this);
      // add the keyboardlistener
      fControlButtons[i].addKeyListener(this);
      ControlPanel.add(fControlButtons[i]);
    }

    // add the controlpanel to the window
    add("South",ControlPanel);

    // create the menubar
    fMenuBar = new MenuBar();

    GeneralMenu generalMenu = new GeneralMenu(this,fMessages.lookup("menuGeneral"));
    fMenuBar.add(generalMenu);

    ModelMenu modelMenu = new ModelMenu(this,fMessages.lookup("menuModel"));
    fMenuBar.add(modelMenu);

    EditorMenu editorMenu = new EditorMenu(this,fMessages.lookup("menuEditor"));
    fMenuBar.add(editorMenu);

    NetworkMenu networkMenu = new NetworkMenu(this,fMessages.lookup("menuNetwork"));
    fMenuBar.add(networkMenu);

    NavigateMenu navigateMenu = new NavigateMenu(this,fMessages.lookup("menuNavigate"));
    fMenuBar.add(navigateMenu);

    setMenuBar(fMenuBar);

    // add the windowlistener
    addWindowListener(this);

    pack();
    show();

    if (fResources.fnmShowAboutBox) {
      // show the about-box
      AboutBox about = new AboutBox(this,"Network Modeler");
    }
    fDrawNetwork = true;

    // show the legend
    fLegend = new MLegend();
    fLegend.setVisible(false);
  }

  private void disableControls()
  {
    for (int i = 0; i < fMenuBar.getMenuCount(); ++i) {
      Menu menu = fMenuBar.getMenu(i);
      for (int j = 0; j < menu.getItemCount(); ++j) {
        MenuItem item = menu.getItem(j);
        item.setEnabled(false);
      }
    }
    for (int i = 0; i < kNrOfControlButtons; ++i) {
      fControlButtons[i].setEnabled(false);
    }
  }

  private void enableControls()
  {
    for (int i = 0; i < fMenuBar.getMenuCount(); ++i) {
      Menu menu = fMenuBar.getMenu(i);
      for (int j = 0; j < menu.getItemCount(); ++j) {
        MenuItem item = menu.getItem(j);
        item.setEnabled(true);
      }
    }
    for (int i = 0; i < kNrOfControlButtons; ++i) {
      fControlButtons[i].setEnabled(true);
    }
  }

  private void moveLeft()
  {
    fOriginXOffset += fResources.fnmPixelsToMove;
    fEditorCanvas.repaint();
  }

  private void moveRight()
  {
    fOriginXOffset -= fResources.fnmPixelsToMove;
    fEditorCanvas.repaint();
  }

  private void moveUp()
  {
    fOriginYOffset -= fResources.fnmPixelsToMove;
    fEditorCanvas.repaint();
  }

  private void moveDown()
  {
    fOriginYOffset += fResources.fnmPixelsToMove;
    fEditorCanvas.repaint();
  }

  private void zoomIn()
  {
    if (fZoomFactor < 32.0f) {
      // keep the window centered, i.e.: the point in the middle of
      // the screen should remain location-invariant under the
      // zoom-transformation
      int MiddleX = (int) Math.round((cx1 + cx2) / 2.0f);
      fOriginXOffset = (int) Math.round((MiddleX - cx1) - ((MiddleX - fOriginXOffset - cx1) * 2.0f));
      int MiddleY = (int) Math.round((cy1 + cy2) / 2.0f);
      fOriginYOffset = (int) Math.round((cy2 - MiddleY) - ((MiddleY - fOriginYOffset - cy1) * 2.0f));
      fZoomFactor *= 2.0f;
      fEditorCanvas.repaint();
      if (!fControlButtons[kZoomOutButton].isEnabled()) {
        fControlButtons[kZoomOutButton].setEnabled(true);
      }
      if (fZoomFactor == 32.0f) {
        fControlButtons[kZoomInButton].setEnabled(false);
      }
    }
  }

  private void zoomOut()
  {
    if (fZoomFactor > (1.0f / 32.0f)) {
      // keep the window centered, i.e.: the point in the middle of
      // the screen should remain location-invariant under the
      // zoom-transformation
      int MiddleX = (int) Math.round((cx1 + cx2) / 2.0f);
      fOriginXOffset = (int) Math.round((MiddleX - cx1) - ((MiddleX - fOriginXOffset - cx1) / 2.0f));
      int MiddleY = (int) Math.round((cy1 + cy2) / 2.0f);
      fOriginYOffset = (int) Math.round((cy2 - MiddleY) - ((MiddleY - fOriginYOffset - cy1) / 2.0f));
      fZoomFactor /= 2.0f;
      fEditorCanvas.repaint();
      if (!fControlButtons[kZoomInButton].isEnabled()) {
        fControlButtons[kZoomInButton].setEnabled(true);
      }
      if (fZoomFactor == (1.0f / 32.0f)) {
        fControlButtons[kZoomOutButton].setEnabled(false);
      }
    }
  }

  // methods from the WindowListener-interface
  public void windowOpened(WindowEvent e) { }
  public void windowClosing(WindowEvent e)
  {
    ConfirmationBox confirmationBox = new ConfirmationBox(this,
                                                          fMessages.lookup("messageConfirmQuit"),
                                                          fMessages.lookup("buttonYes"),
                                                          fMessages.lookup("buttonNo"));
    fDrawNetwork = false;
    if (confirmationBox.confirmed()) {
      // quit the running program
      System.exit(0);
    }
    fDrawNetwork = true;
  }
  public void windowClosed(WindowEvent e) { }
  public void windowIconified(WindowEvent e) { }
  public void windowDeiconified(WindowEvent e) { }
  public void windowActivated(WindowEvent e) { }
  public void windowDeactivated(WindowEvent e) { }

  // method from the ActionListener-interface
  public void actionPerformed(ActionEvent e)
  {
    String Command = e.getActionCommand();

    if (Command.equals(fMessages.lookup("menuAbout"))) {
      AboutBox about = new AboutBox(this,"Network Modeler");
    }

    if (Command.equals(fMessages.lookup("menuQuit"))) {
      windowClosing(null);
    }

    if (Command.equals(fMessages.lookup("menuLoadMap"))) {
      FileDialog fd = new FileDialog(this,
                                     fMessages.lookup("textLoadMapDialogTitle"),
                                     FileDialog.LOAD);
      fd.show();
      String Dir = fd.getDirectory();
      String Filename = fd.getFile();
      if ((Dir != null) && (Filename != null)) {
        fMapFilename = Dir + Filename;
        fMap = Toolkit.getDefaultToolkit().getImage(fMapFilename);
        fMapAvailable = true;
        fEditorCanvas.repaint();
      }
    }

    if (Command.equals(fMessages.lookup("menuClearMap"))) {
      fMapFilename = "";
      fMapAvailable = false;
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuRelocateMap"))) {
      // let the user specify a point
      fMessageCanvas.showMessage(fMessages.lookup("messageSelectMapPoint"));
      Point p = fEditorCanvas.specifyPoint();
      fMessageCanvas.clearMessage();

      fMapXOffset = (int) Math.round(((p.x - cx1) - fOriginXOffset) / fZoomFactor);
      fMapYOffset = (int) Math.round(((cy2 - p.y) - fOriginYOffset) / fZoomFactor);

      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuSetOrigin"))) {
      // only allow setting of the origin when the zoomfactor is 1:1
      if (fZoomFactor == 1.0f) {
        // let the user specify a point
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectOriginPoint"));
        Point p = fEditorCanvas.specifyPoint();
        fMessageCanvas.clearMessage();

        int OldOriginXOffset = fOriginXOffset;
        int OldOriginYOffset = fOriginYOffset;

        fOriginXOffset = p.x - cx1;
        fOriginYOffset = cy2 - p.y;

        fMapXOffset = fMapXOffset + (OldOriginXOffset - fOriginXOffset);
        fMapYOffset = fMapYOffset + (OldOriginYOffset - fOriginYOffset);

        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorSetOriginRequiresNoZoom"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuSetScale"))) {
      // let the user specify two points
      fMessageCanvas.showMessage(fMessages.lookup("messageSelectFirstPointForScaling"));
      Point p1 = fEditorCanvas.specifyPoint();

      fMessageCanvas.showMessage(fMessages.lookup("messageSelectSecondPointForScaling"));
      Point p2 = fEditorCanvas.specifyPoint();
      fMessageCanvas.clearMessage();

      SetScaleBox setScaleBox = new SetScaleBox(this,(int) Math.round(Math.abs(p2.x - p1.x) / fZoomFactor));
      if (!setScaleBox.canceled()) {
        fScaleFactor = setScaleBox.getScaleFactor();
      }
      fEditorCanvas.repaint();      
    }

    if (Command.equals(fMessages.lookup("menuGeneralSettings"))) {
      GeneralSettingsBox generalSettingsBox = new GeneralSettingsBox(this,fNetwork.fLaneWidth,fNetwork.fRailroadWidth,fNetwork.fTitle);
      if (!generalSettingsBox.canceled()) {
        fNetwork.fLaneWidth = generalSettingsBox.fLaneWidth;
        fNetwork.fRailroadWidth = generalSettingsBox.fRailroadWidth;
        fNetwork.fTitle = generalSettingsBox.fTitle;
        if (!fNetwork.fTitle.equalsIgnoreCase("")) {
          setTitle(fMessages.lookup("textNetworkModelerWindowTitle") +
                   " [" + fNetwork.fTitle + "]");
        }
        else {
          setTitle(fMessages.lookup("textNetworkModelerWindowTitle"));
        }
      }
    }

    if (Command.equals(fMessages.lookup("menuAddNode")) ||
        Command.equals(fMessages.lookup("buttonAddNode"))) {
      // let the user specify a point
      fMessageCanvas.showMessage(fMessages.lookup("messageSelectNodePoint"));
      Point p = fEditorCanvas.specifyPoint();
      fMessageCanvas.clearMessage();

      Node node = new Node();
      // find the next free nodenumber > fLastNodeID
      node.fID = fLastNodeID + 1;
      node.fType = fLastNodeType;
      Node FindNode = (Node) fNetwork.fNodes.get(new Integer(node.fID));
      while (FindNode != null) {
        ++node.fID;
        FindNode = (Node) fNetwork.fNodes.get(new Integer(node.fID));
      }
      node.fXPos = (int) Math.round((((p.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
      node.fYPos = (int) Math.round((((cy2 - p.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);
      NodeBox AddNodeBox = new NodeBox(this,NodeBox.kAddMode,node);
      if (!AddNodeBox.canceled()) {
        node = AddNodeBox.getNode();
        // add the node to the network
        fNetwork.fNodes.put(new Integer(node.fID),node);
        fLastNodeID = node.fID;
        fLastNodeType = node.fType;
      }
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuEditNode")) ||
        Command.equals(fMessages.lookup("buttonEditNode"))) {
      if (fNetwork.fNodes.size() != 0) {
        // let the user specify a node
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNodeToEdit"));
        Node SpecifiedNode = fEditorCanvas.specifyNode();
        fMessageCanvas.clearMessage();

        if (SpecifiedNode == null) {
          return;
        }

        Node EditNode = new Node();
        SpecifiedNode.cloneTo(EditNode);
        NodeBox EditNodeBox = new NodeBox(this,NodeBox.kEditMode,EditNode);
        if (!EditNodeBox.canceled()) {
          EditNode = EditNodeBox.getNode();
          EditNode.cloneTo(SpecifiedNode);
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoNodesDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuMoveNode"))) {
      if (fNetwork.fNodes.size() != 0) {
        // let the user specify a node
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNodeToMove"));
        Node SpecifiedNode = fEditorCanvas.specifyNode();
        fMessageCanvas.clearMessage();

        if (SpecifiedNode == null) {
          return;
        }

        // let the user specify a new location
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNewLocationForNode"));
        Point TranslatePoint = fEditorCanvas.specifyPoint();
        int NewXPos = (int) Math.round((((TranslatePoint.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
        int NewYPos = (int) Math.round((((cy2 - TranslatePoint.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);
        fMessageCanvas.clearMessage();

        SpecifiedNode.fXPos = NewXPos;
        SpecifiedNode.fYPos = NewYPos;

        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoNodesDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuDeleteNode"))) {
      if (fNetwork.fNodes.size() != 0) {
        // let the user specify a node
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNodeToDelete"));
        Node SpecifiedNode = fEditorCanvas.specifyNode();
        fMessageCanvas.clearMessage();

        if (SpecifiedNode == null) {
          return;
        }

        ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageDeleteNode"),
                                                                   fMessages.lookup("buttonYes"),
                                                                   fMessages.lookup("buttonNo"));
        if (confirmationBox.confirmed()) {

          // scan the links
          Enumeration LinkEnum = fNetwork.fLinks.elements();
          while (LinkEnum.hasMoreElements()) {
            Link link = (Link) LinkEnum.nextElement();

            // remove all the nodes that are equal to the node to delete
            int CurrentIndex = 0;
            while (CurrentIndex < link.fNrOfNodes) {
              if (link.fNodes[CurrentIndex].fID == SpecifiedNode.fID) {
                // compact the arrays
                for (int i = CurrentIndex; i < (link.fNrOfNodes - 1); ++i) {
                  link.fNodes[i] = link.fNodes[i + 1];
                  if (i < (link.fNrOfNodes - 2)) {
                    link.fLengths[i] = link.fLengths[i + 1];
                    link.fGrades[i] = link.fGrades[i + 1];
                  }
                }
                --link.fNrOfNodes;
              }
              else {
                ++CurrentIndex;
              }
            }

            // if the link isn't compliant anymore then delete it
            if (!link.nodeCompliant()) {
              fNetwork.fLinks.remove(new Integer(link.fID)); // ignore returned value
            }
            else {
              // truncate the arrays
              Node[] Nodes = new Node[link.fNrOfNodes];
              int[] Lengths = new int[link.fNrOfNodes - 1];
              float[] Grades = new float[link.fNrOfNodes - 1];
              for (int i = 0; i < link.fNrOfNodes; ++i) {
                Nodes[i] = link.fNodes[i];
                if (i < (link.fNrOfNodes - 2)) {
                  Lengths[i] = link.fLengths[i];
                  Grades[i] = link.fGrades[i];
                }
              }
              link.fNodes = new Node[link.fNrOfNodes];
              link.fLengths = new int[link.fNrOfNodes - 1];
              link.fGrades = new float[link.fNrOfNodes - 1];
              for (int i = 0; i < link.fNrOfNodes; ++i) {
                link.fNodes[i] = Nodes[i];
                if (i < (link.fNrOfNodes - 2)) {
                  link.fLengths[i] = Lengths[i];
                  link.fGrades[i] = Grades[i];
                }
              }
            }
          }

          // scan the environment
          Enumeration EnvironmentEnum = fNetwork.fEnvironment.elements();
          while (EnvironmentEnum.hasMoreElements()) {
            Environment environment = (Environment) EnvironmentEnum.nextElement();

            // remove all the nodes that are equal to the node to delete
            int CurrentIndex = 0;
            while (CurrentIndex < environment.fNrOfNodes) {
              if (environment.fNodes[CurrentIndex].fID == SpecifiedNode.fID) {
                // compact the array
                for (int i = CurrentIndex; i < (environment.fNrOfNodes - 1); ++i) {
                  environment.fNodes[i] = environment.fNodes[i + 1];
                }
                --environment.fNrOfNodes;
              }
              else {
                ++CurrentIndex;
              }
            }
            // truncate the array
            Node[] Nodes = new Node[environment.fNrOfNodes];
            for (int i = 0; i < environment.fNrOfNodes; ++i) {
              Nodes[i] = environment.fNodes[i];
            }
            environment.fNodes = new Node[environment.fNrOfNodes];
            for (int i = 0; i < environment.fNrOfNodes; ++i) {
              environment.fNodes[i] = Nodes[i];
            }

            // if the environment isn't compliant anymore then delete it
            if (!environment.nodeCompliant()) {
              fNetwork.fEnvironment.removeElement(environment); // ignore returned value
            }
          }
          fNetwork.fNodes.remove(new Integer(SpecifiedNode.fID)); // ignore returned value
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoNodesDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuAddLink")) ||
        Command.equals(fMessages.lookup("buttonAddLink"))) {
      Link link = new Link();
      link.fType = Link.kHighway;
      link.fSpeedLimit = 100;
      // find the next free linknumber > fLastLinkID
      link.fID = fLastLinkID + 1;
      Link FindLink = (Link) fNetwork.fLinks.get(new Integer(link.fID));
      while (FindLink != null) {
        ++link.fID;
        FindLink = (Link) fNetwork.fLinks.get(new Integer(link.fID));
      }
      LinkBox AddLinkBox = new LinkBox(this,LinkBox.kAddMode,link);
      if (!AddLinkBox.canceled()) {
        link = AddLinkBox.getLink();
        // add the link to the network
        fNetwork.fLinks.put(new Integer(link.fID),link);
        fLastLinkID = link.fID;
      }
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuEditLink")) ||
        Command.equals(fMessages.lookup("buttonEditLink"))) {
      if (fNetwork.fLinks.size() != 0) {
/*
        // let the user specify a link (OLD METHOD)
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectLinkToEdit"));
        Link SpecifiedLink = fEditorCanvas.specifyLink();
        fMessageCanvas.clearMessage();

        if (SpecifiedLink == null) {
          return;
        }

*/
        // let the user specify a link
        LinkIDBox LinkID = new LinkIDBox(this);
        if (LinkID.canceled()) {
          return;
        }

        Link SpecifiedLink = (Link) fNetwork.fLinks.get(new Integer(LinkID.getLinkID()));

        Link EditLink = new Link();
        SpecifiedLink.cloneTo(EditLink);
        LinkBox EditLinkBox = new LinkBox(this,LinkBox.kEditMode,EditLink);
        if (!EditLinkBox.canceled()) {
          EditLink = EditLinkBox.getLink();
          EditLink.cloneTo(SpecifiedLink);
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoLinksDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuDeleteLink"))) {
      if (fNetwork.fLinks.size() != 0) {
        // let the user specify a link
        LinkIDBox LinkID = new LinkIDBox(this);
        if (LinkID.canceled()) {
          return;
        }

        Link SpecifiedLink = (Link) fNetwork.fLinks.get(new Integer(LinkID.getLinkID()));

        ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageDeleteLink"),
                                                                   fMessages.lookup("buttonYes"),
                                                                   fMessages.lookup("buttonNo"));
        if (confirmationBox.confirmed()) {
          fNetwork.fLinks.remove(new Integer(SpecifiedLink.fID)); // ignore returned value
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoLinksDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuSetLaneChangePermissions"))) {
      if (fNetwork.fLinks.size() != 0) {
        // let the user specify a link
        LinkIDBox LinkID = new LinkIDBox(this);
        if (LinkID.canceled()) {
          return;
        }

        Link SpecifiedLink = (Link) fNetwork.fLinks.get(new Integer(LinkID.getLinkID()));

        Link EditLink = new Link();
        SpecifiedLink.cloneTo(EditLink);

        LaneChangesBox laneChangesBox = new LaneChangesBox(this,EditLink);

        if (!laneChangesBox.canceled()) {
          EditLink = laneChangesBox.getLink();
          EditLink.cloneTo(SpecifiedLink);
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoLinksDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuAddEnvironment"))) {
      Environment environment = new Environment();
      EnvironmentBox AddEnvironmentBox = new EnvironmentBox(this,EnvironmentBox.kAddMode,environment);
      if (!AddEnvironmentBox.canceled()) {
        environment = AddEnvironmentBox.getEnvironment();
        // add the environment to the network
        fNetwork.fEnvironment.addElement(environment);
      }
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuEditEnvironment"))) {
      if (fNetwork.fEnvironment.size() != 0) {
        // let the user specify an environment
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectEnvironmentToEdit"));
        Environment SpecifiedEnvironment = fEditorCanvas.specifyEnvironment();
        fMessageCanvas.clearMessage();

        if (SpecifiedEnvironment == null) {
          return;
        }

        Environment EditEnvironment = new Environment();
        SpecifiedEnvironment.cloneTo(EditEnvironment);
        EnvironmentBox EditEnvironmentBox = new EnvironmentBox(this,EnvironmentBox.kEditMode,EditEnvironment);
        if (!EditEnvironmentBox.canceled()) {

          EditEnvironment = EditEnvironmentBox.getEnvironment();
          EditEnvironment.cloneTo(SpecifiedEnvironment);
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoEnvironmentsDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuDeleteEnvironment"))) {
      if (fNetwork.fEnvironment.size() != 0) {
        // let the user specify an environment
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectEnvironmentToDelete"));
        Environment SpecifiedEnvironment = fEditorCanvas.specifyEnvironment();
        fMessageCanvas.clearMessage();

        if (SpecifiedEnvironment == null) {
          return;
        }

        ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageDeleteEnvironment"),
                                                                   fMessages.lookup("buttonYes"),
                                                                   fMessages.lookup("buttonNo"));
        if (confirmationBox.confirmed()) {
          fNetwork.fEnvironment.removeElement(SpecifiedEnvironment);  // ignore returned value
        }
        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoEnvironmentsDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuLoadNetwork"))) {
      FileDialog fd = new FileDialog(this,
                                     fMessages.lookup("textLoadNetworkDialogTitle"),
                                     FileDialog.LOAD);
      fd.show();
      String Dir = fd.getDirectory();
      String Filename = fd.getFile();
      if ((Dir != null) && (Filename != null)) {
        fNetwork.clear();
        try {
          fDrawNetwork = false;
          fNetwork.loadFromFile(Dir + Filename);
          if (fNetwork.modelSectionRead()) {
            ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageUseModelSection"),
                                                                       fMessages.lookup("buttonYes"),
                                                                       fMessages.lookup("buttonNo"));
            if (confirmationBox.confirmed()) {
              fOriginXOffset = fNetwork.fOriginXOffset;
              fOriginYOffset = fNetwork.fOriginYOffset;
              fScaleFactor = fNetwork.fScaleFactor;
              fZoomFactor = fNetwork.fZoomFactor;
              fMapXOffset = fNetwork.fMapXOffset;
              fMapYOffset = fNetwork.fMapYOffset;
              fMapFilename = fNetwork.fMapFilename;
              fMap = Toolkit.getDefaultToolkit().getImage(fMapFilename);
              fMapAvailable = true;
              if (!fNetwork.fTitle.equalsIgnoreCase("")) {
                setTitle(fMessages.lookup("textNetworkModelerWindowTitle") +
                         " [" + fNetwork.fTitle + "]");
              }
              else {
                setTitle(fMessages.lookup("textNetworkModelerWindowTitle"));
              }
            }
          }
          fDrawNetwork = true;
        }
        catch (GeneralError err) {
          MessageBox ErrorBox = new MessageBox(this,err.message(),MessageBox.kError);
          fNetwork.clear();
        }
        fEditorCanvas.repaint();
      }
    }

    if (Command.equals(fMessages.lookup("menuSaveNetwork"))) {
      FileDialog fd = new FileDialog(this,
                                     fMessages.lookup("textSaveNetworkDialogTitle"),
                                     FileDialog.SAVE);
      fd.show();
      String Dir = fd.getDirectory();
      String Filename = fd.getFile();

      if ((Dir != null) && (Filename != null)) {
        fNetwork.fOriginXOffset = fOriginXOffset;
        fNetwork.fOriginYOffset = fOriginYOffset;
        fNetwork.fScaleFactor = fScaleFactor;
        fNetwork.fZoomFactor = fZoomFactor;
        fNetwork.fMapXOffset = fMapXOffset;
        fNetwork.fMapYOffset = fMapYOffset;
        fNetwork.fMapFilename = fMapFilename;
        if (!fNetwork.saveToFile(Dir + Filename)) {
          MessageBox errorBox = new MessageBox(this,
                                               fMessages.lookup("errorNetworkNotSaved"),
                                               MessageBox.kError);
        }
      }
    }

    if (Command.equals(fMessages.lookup("menuClearNetwork"))) {
      ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageClearNetwork"),
                                                                 fMessages.lookup("buttonYes"),
                                                                 fMessages.lookup("buttonNo"));
      if (confirmationBox.confirmed()) {
        fNetwork.clear();
        fEditorCanvas.repaint();
        setTitle(fMessages.lookup("textNetworkModelerWindowTitle"));
      }
    }

    if (Command.equals(fMessages.lookup("menuAnalyzeNetwork"))) {
      // check the network
      Report ResultReport = fNetwork.analyse();
      boolean Ok = true;
      if (ResultReport.isEmpty()) {
        fNetwork.calcGeometry();

        // create summary for bounding-box
        ResultReport.addLine(fMessages.lookup("textReportBoundingBoxCaption"));
        ResultReport.addLine(fMessages.lookup("textReportBoundingBoxCaptionLine"));
        ResultReport.addLine("  " + fMessages.lookup("textReportFrom") + " (" +
                             String.valueOf(fNetwork.fBoundingBox.fMinX) +
                             "," +
                             String.valueOf(fNetwork.fBoundingBox.fMinY) +
                             "," +
                             String.valueOf(fNetwork.fBoundingBox.fMinZ) +
                             ") " + fMessages.lookup("textReportTo") + " (" +
                             String.valueOf(fNetwork.fBoundingBox.fMaxX) +
                             "," +
                             String.valueOf(fNetwork.fBoundingBox.fMaxY) +
                             "," +
                             String.valueOf(fNetwork.fBoundingBox.fMaxZ) +
                             ").");

        ResultReport.addLine(" ");

        // create summary for nodes
        int NrOfOriginNodes = 0;
        int NrOfDestinationNodes = 0;
        int NrOfJunctionNodes = 0;
        int NrOfBendNodes = 0;
        int NrOfEnvironmentNodes = 0;

        Enumeration NodeEnum = fNetwork.fNodes.elements();
        while (NodeEnum.hasMoreElements()) {
          Node node = (Node) NodeEnum.nextElement();

          if (node.fType == Node.kOrigin) {
            ++NrOfOriginNodes;
          }
          else if (node.fType == Node.kDestination) {
            ++NrOfDestinationNodes;
          }
          else if (node.fType == Node.kJunction) {
            ++NrOfJunctionNodes;
          }
          else if (node.fType == Node.kBend) {
            ++NrOfBendNodes;
          }
          else if (node.fType == Node.kEnvironment) {
            ++NrOfEnvironmentNodes;
          }
        }
        int TotalNrOfNodes = NrOfOriginNodes +
                             NrOfDestinationNodes +
                             NrOfJunctionNodes +
                             NrOfBendNodes +
                             NrOfEnvironmentNodes;

        ResultReport.addLine(fMessages.lookup("textReportNodesCaption"));
        ResultReport.addLine(fMessages.lookup("textReportNodesCaptionLine"));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportOrigins")      + " " + String.valueOf(NrOfOriginNodes));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportDestinations") + " " + String.valueOf(NrOfDestinationNodes));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportJunctions")    + " " + String.valueOf(NrOfJunctionNodes));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportBends")        + " " + String.valueOf(NrOfBendNodes));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportEnvironments") + " " + String.valueOf(NrOfEnvironmentNodes));
        ResultReport.addLine("=>  " + fMessages.lookup("textReportTotalNodes")    + " " + String.valueOf(TotalNrOfNodes));

        ResultReport.addLine(" ");

        // create summary for links
        int NrOfTunnelLinks = 0;
        int NrOfHighwayLinks = 0;
        int NrOfOnRampLinks = 0;
        int NrOfOffRampLinks = 0;
        int NrOfNarrowingLinks = 0;
        int NrOfWideningLinks = 0;
        int NrOfTerminators = 0;

        Enumeration LinkEnum = fNetwork.fLinks.elements();
        while (LinkEnum.hasMoreElements()) {
          Link link = (Link) LinkEnum.nextElement();

          if (link.fType == Link.kTunnel) {
            ++NrOfTunnelLinks;
          }
          else if (link.fType == Link.kHighway) {
            ++NrOfHighwayLinks;
          }
          else if ((link.fType == Link.kSeparatedOnRamp) ||
                   (link.fType == Link.kMergedOnRamp)) {
            ++NrOfOnRampLinks;
          }
          else if ((link.fType == Link.kSeparatedOffRamp) ||
                   (link.fType == Link.kMergedOffRamp)) {
            ++NrOfOffRampLinks;
          }
          else if (link.fType == Link.kNarrowing) {
            ++NrOfNarrowingLinks;
          }
          else if (link.fType == Link.kWidening) {
            ++NrOfWideningLinks;
          }
          if (link.fIsTerminator) {
            ++NrOfTerminators;
          }
        }
        int TotalNrOfLinks = NrOfTunnelLinks +
                             NrOfHighwayLinks +
                             NrOfOnRampLinks +
                             NrOfOffRampLinks +
                             NrOfNarrowingLinks +
                             NrOfWideningLinks;

        ResultReport.addLine(fMessages.lookup("textReportLinksCaption"));
        ResultReport.addLine(fMessages.lookup("textReportLinksCaptionLine"));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportTerminators") + " " + String.valueOf(NrOfTerminators));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportTunnels")     + " " + String.valueOf(NrOfTunnelLinks));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportHighways")    + " " + String.valueOf(NrOfHighwayLinks));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportOnRamps")     + " " + String.valueOf(NrOfOnRampLinks));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportOffRamps")    + " " + String.valueOf(NrOfOffRampLinks));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportNarrowings")  + " " + String.valueOf(NrOfNarrowingLinks));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportWidenings")   + " " + String.valueOf(NrOfWideningLinks));
        ResultReport.addLine("=>  " + fMessages.lookup("textReportTotalLinks")   + " " + String.valueOf(TotalNrOfLinks));

        ResultReport.addLine(" ");

        LinkEnum = fNetwork.fLinks.elements();
        while (LinkEnum.hasMoreElements()) {
          Link link = (Link) LinkEnum.nextElement();

          if (!link.fName.equalsIgnoreCase("")) {
            ResultReport.addLine(fMessages.lookup("textReportLinkID") + link.fID +
                                 " (" + link.fName + ") :");
          }
          else {
            ResultReport.addLine(fMessages.lookup("textReportLinkID") + link.fID + " :");
          }
          ResultReport.addLine(" * #" + fMessages.lookup("textReportLinkNodes") + " " + String.valueOf(link.fNrOfNodes));
          if (link.fIsTerminator) {
            ResultReport.addLine(" * " + fMessages.lookup("textReportIsTerminator"));
          }
          else {
            ResultReport.addLine(" * " + fMessages.lookup("textReportIsNoTerminator"));
          }
          int LinkLength = 0;
          for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
            LinkLength += link.fLengths[i];
          }
          ResultReport.addLine(" * " + fMessages.lookup("textReportTotalLinkLength") + " " +
                               String.valueOf(LinkLength) + " " +
                               fMessages.lookup("textDistanceUnit"));
          String LinkGrades = " * " + fMessages.lookup("textReportLinkGrades") + " ";
          for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
            LinkGrades += String.valueOf(link.fGrades[i]) + "%";
            if (i < (link.fNrOfNodes - 2)) {
              LinkGrades += ", ";
            }
          }
          ResultReport.addLine(LinkGrades);
        }

        ResultReport.addLine(" ");

        // create summary for environments
        int NrOfRectangleEnvironments = 0;
        int NrOfPolygonEnvironments = 0;
        int NrOfOvalEnvironments = 0;
        int NrOfRailroadEnvironments = 0;

        Enumeration EnvironmentEnum = fNetwork.fEnvironment.elements();
        while (EnvironmentEnum.hasMoreElements()) {
          Environment environment = (Environment) EnvironmentEnum.nextElement();

          if (environment.fType == Environment.kRectangle) {
            ++NrOfRectangleEnvironments;
          }
          else if (environment.fType == Environment.kPolygon) {
            ++NrOfPolygonEnvironments;
          }
          else if (environment.fType == Environment.kOval) {
            ++NrOfOvalEnvironments;
          }
          else if (environment.fType == Environment.kRailroad) {
            ++NrOfRailroadEnvironments;
          }
        }
        int TotalNrOfEnvironments = NrOfRectangleEnvironments +
                                    NrOfPolygonEnvironments +
                                    NrOfOvalEnvironments +
                                    NrOfRailroadEnvironments;

        ResultReport.addLine(fMessages.lookup("textReportEnvironmentsCaption"));
        ResultReport.addLine(fMessages.lookup("textReportEnvironmentsCaptionLine"));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportRectangles")       + " " + String.valueOf(NrOfRectangleEnvironments));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportPolygons")         + " " + String.valueOf(NrOfPolygonEnvironments));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportOvals")            + " " + String.valueOf(NrOfOvalEnvironments));
        ResultReport.addLine("  * #" + fMessages.lookup("textReportRailroads")        + " " + String.valueOf(NrOfRailroadEnvironments));
        ResultReport.addLine("=>  " + fMessages.lookup("textReportTotalEnvironments") + " " + String.valueOf(TotalNrOfEnvironments));
      }
      else {
        Ok = false;
      }

      // show the report
      if (Ok) {
        ReportBox reportBox = new ReportBox(this,ResultReport,
                                            fMessages.lookup("textAnalysisOkDialogTitle"),
                                            fMessages.lookup("messageSummaryReport") + " :\n\n",
                                            ReportBox.kShowTickMarks);
      }
      else {
        ReportBox reportBox = new ReportBox(this,ResultReport,
                                            fMessages.lookup("textAnalysisNotOkDialogTitle"),
                                            fMessages.lookup("messageSummaryReport") + " :\n\n",
                                            ReportBox.kShowTickMarks);
      }
    }

    if (Command.equals(fMessages.lookup("menuTranslateNetworkByNode"))) {
      if (fNetwork.fNodes.size() != 0) {
        // let the user specify a node
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNodeForTranslating"));
        Node TranslateNode = fEditorCanvas.specifyNode();
        fMessageCanvas.clearMessage();

        if (TranslateNode == null) {
          return;
        }

        // let the user specify a new location
        fMessageCanvas.showMessage(fMessages.lookup("messageSelectNewLocationForTranslating"));
        Point TranslatePoint = fEditorCanvas.specifyPoint();
        int NewXPos = (int) Math.round((((TranslatePoint.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
        int NewYPos = (int) Math.round((((cy2 - TranslatePoint.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);
        fMessageCanvas.clearMessage();

        ConfirmationBox confirmationBox = new ConfirmationBox(this,fMessages.lookup("messageTranslateNetwork"),
                                                                   fMessages.lookup("buttonYes"),
                                                                   fMessages.lookup("buttonNo"));
        if (confirmationBox.confirmed()) {
          int dx = NewXPos - TranslateNode.fXPos;
          int dy = NewYPos - TranslateNode.fYPos;
          fNetwork.translate(dx,dy,0);
        }

        fEditorCanvas.repaint();
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoNodesDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuTranslateNetworkManually"))) {
      if (fNetwork.fNodes.size() != 0) {
        TranslateNetworkBox translateNetworkBox = new TranslateNetworkBox(this);

        if (!translateNetworkBox.canceled()) {
          fNetwork.translate(translateNetworkBox.getdx(),
                             translateNetworkBox.getdy(),
                             translateNetworkBox.getdz());
          fEditorCanvas.repaint();
        }
      }
      else {
        MessageBox messageBox = new MessageBox(this,
                                               fMessages.lookup("errorNoNodesDefinedYet"),
                                               MessageBox.kError);
      }
    }

    if (Command.equals(fMessages.lookup("menuMoveLeft")) ||
        Command.equals(fMessages.lookup("buttonMoveLeft"))) {
      moveLeft();
    }

    if (Command.equals(fMessages.lookup("menuMoveRight")) ||
        Command.equals(fMessages.lookup("buttonMoveRight"))) {
      moveRight();
    }

    if (Command.equals(fMessages.lookup("menuMoveUp")) ||
        Command.equals(fMessages.lookup("buttonMoveUp"))) {
      moveUp();
    }

    if (Command.equals(fMessages.lookup("menuMoveDown")) ||
        Command.equals(fMessages.lookup("buttonMoveDown"))) {
      moveDown();
    }

    if (Command.equals(fMessages.lookup("menuGoto0"))) {
      fOriginXOffset = 0;
      fOriginYOffset = 0;
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuZoomIn")) ||
        Command.equals(fMessages.lookup("buttonZoomIn"))) {
      zoomIn();
    }

    if (Command.equals(fMessages.lookup("menuZoomOut")) ||
        Command.equals(fMessages.lookup("buttonZoomOut"))) {
      zoomOut();
    }

    if (Command.equals(fMessages.lookup("menuResetZoom"))) {
      fZoomFactor = 1.0f;
      fEditorCanvas.repaint();
    }

    if (Command.equals(fMessages.lookup("menuRedrawScreen"))) {
      fEditorCanvas.repaint();
    }
  }

  // methods from the KeyListener-interface
  public void keyTyped(KeyEvent e) { }
  public void keyPressed(KeyEvent e)
  {
    int KeyCode = e.getKeyCode();

    if (KeyCode == KeyEvent.VK_LEFT) {
      moveLeft();
    }
    else if (KeyCode == KeyEvent.VK_RIGHT) {
      moveRight();
    }
    else if (KeyCode == KeyEvent.VK_UP) {
      moveUp();
    }
    else if (KeyCode == KeyEvent.VK_DOWN) {
      moveDown();
    }
    else if (KeyCode == KeyEvent.VK_ADD) {
      zoomIn();
    }
    else if (KeyCode == KeyEvent.VK_SUBTRACT) {
      zoomOut();
    }
    else if (KeyCode == KeyEvent.VK_N) {
      fResources.fnmShowNodeIDs = !fResources.fnmShowNodeIDs;
      fEditorCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_L) {
      fResources.fnmShowLinkIDs = !fResources.fnmShowLinkIDs;
      fEditorCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_S) {
      fResources.fnmSpreadLinkIDs = !fResources.fnmSpreadLinkIDs;
      fEditorCanvas.repaint();
    }
  }
  public void keyReleased(KeyEvent e) { }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */  
  // the editor's canvas
  private class EditorCanvas extends Canvas implements MouseListener
  {
    public EditorCanvas(int width, int height)
    {
      super();

      setSize(width,height);

      // add the mouselistener
      addMouseListener(this);

      setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
      repaint();
    }

    public void update(Graphics g)
    {
      paint(g);
    }

    public boolean imageUpdate(Image img, int flags, int x, int y, int w, int h)
    {
      // only allow repainting of this canvas when the image is completely loaded
      // this prevents flickering
      if ((flags & ALLBITS) != 0) {
        repaint();
      }
      return ((flags & (ALLBITS|ERROR)) == 0);
    }

    public void paint(Graphics g)
    {
      // prevent drawing of the map when the aboutbox is shown at startup or
      // when the user is loading a network from file
      if (fDrawNetwork == false) {
        return;
      }

      // define the boundaries for drawing
      cx1 = 5;
      cy1 = 5;
      cx2 = (cx1 + getSize().width) - (getInsets().right + 5 + 5);
      cy2 = (cy1 + getSize().height) - (getInsets().bottom + 5 + 5);
      cwidth = cx2 - cx1;
      cheight = cy2 - cy1;

      // clear the draw-area
      g.clearRect(cx1,cy1,cwidth + 1,cheight + 1);

      // draw a border
      g.setColor(fResources.knmBorderColor);
      g.drawRoundRect(cx1 - 1,cy1 - 1,cwidth + 2,cheight + 2,10,10);

      // activate clipping of graphics primitives
      g.setClip(cx1 - 1,cy1 - 1,cwidth + 2,cheight + 2);

      // show the map at the correct position and with the correct size
      if (fMapAvailable) {
        fMapWidth = fMap.getWidth(this);
        fMapHeight = fMap.getHeight(this);
        g.drawImage(fMap,
                    (int) Math.round(cx1 + (fOriginXOffset + (fMapXOffset * fZoomFactor))),
                    (int) Math.round(cy2 - (fOriginYOffset + ((fMapHeight + fMapYOffset) * fZoomFactor))),
                    (int) Math.round(fMapWidth * fZoomFactor),
                    (int) Math.round(fMapHeight * fZoomFactor),
                    this);
      }

      // show the network's environment
      if (fNetwork.fEnvironment.size() != 0) {
        Enumeration EnvironmentEnum = fNetwork.fEnvironment.elements();

        while (EnvironmentEnum.hasMoreElements()) {
          Environment environment = (Environment) EnvironmentEnum.nextElement();

          g.setColor(environment.fColor);

          if (environment.fType == Environment.kRectangle) {
            // transform the points to the screen
            int x1 = (int) Math.round(cx1 + fOriginXOffset + ((environment.fNodes[0].fXPos / fScaleFactor) * fZoomFactor));
            int y1 = (int) Math.round(cy2 - (fOriginYOffset + ((environment.fNodes[0].fYPos / fScaleFactor) * fZoomFactor)));
            int x2 = (int) Math.round(cx1 + fOriginXOffset + ((environment.fNodes[1].fXPos / fScaleFactor) * fZoomFactor));
            int y2 = (int) Math.round(cy2 - (fOriginYOffset + ((environment.fNodes[1].fYPos / fScaleFactor) * fZoomFactor)));

            Polygon Rectangle = new Polygon();
            Rectangle.addPoint(x1,y1);
            Rectangle.addPoint(x1,y2);
            Rectangle.addPoint(x2,y2);
            Rectangle.addPoint(x2,y1);

            if (environment.fFilled) {
              g.fillPolygon(Rectangle);
            }
            else {
              // draw the outline
              g.drawPolygon(Rectangle);
            }
          }
          else if (environment.fType == Environment.kPolygon) {
            Polygon polygon = new Polygon();

            for (int i = 0; i < environment.fNrOfNodes; ++i) {
              // transform the points to the screen
              int x = (int) Math.round(cx1 + fOriginXOffset + ((environment.fNodes[i].fXPos / fScaleFactor) * fZoomFactor));
              int y = (int) Math.round(cy2 - (fOriginYOffset + ((environment.fNodes[i].fYPos / fScaleFactor) * fZoomFactor)));
              polygon.addPoint(x,y);
            }

            if (environment.fFilled) {
              g.fillPolygon(polygon);
            }
            else {
              // draw the outline
              g.drawPolygon(polygon);
            }
          }
          else if (environment.fType == Environment.kOval) {
            Polygon polygon = new Polygon();

            for (int i = 0; i < fResources.fnvOvalInterpolation; ++i) {
              float Angle = (float) (((2.0 * Math.PI) / fResources.fnvOvalInterpolation) * i);
              int OvalX = (int) (environment.fNodes[0].fXPos + Math.round((float) environment.fXRadius * Math.cos(Angle)));
              int OvalY = (int) (environment.fNodes[0].fYPos + Math.round((float) environment.fYRadius * Math.sin(Angle)));

              // transform the points to the screen
              int x = (int) Math.round(cx1 + fOriginXOffset + ((OvalX / fScaleFactor) * fZoomFactor));
              int y = (int) Math.round(cy2 - (fOriginYOffset + ((OvalY / fScaleFactor) * fZoomFactor)));
              polygon.addPoint(x,y);
            }

            if (environment.fFilled) {
              g.fillPolygon(polygon);
            }
            else {
              // draw the outline
              g.drawPolygon(polygon);
            }
          }
          else if (environment.fType == Environment.kRailroad) {
            // draw straight-lines between the railroadsegments
            for (int i = 0; i < (environment.fNrOfNodes - 1); ++i) {
              // transform the point to the screen
              int x1 = (int) Math.round(cx1 + fOriginXOffset + ((environment.fNodes[i].fXPos / fScaleFactor) * fZoomFactor));
              int y1 = (int) Math.round(cy2 - (fOriginYOffset + ((environment.fNodes[i].fYPos / fScaleFactor) * fZoomFactor)));
              int x2 = (int) Math.round(cx1 + fOriginXOffset + ((environment.fNodes[i + 1].fXPos / fScaleFactor) * fZoomFactor));
              int y2 = (int) Math.round(cy2 - (fOriginYOffset + ((environment.fNodes[i + 1].fYPos / fScaleFactor) * fZoomFactor)));

              g.setColor(Color.white);
              g.drawLine(x1 - 1,y1,x2 - 1,y2);
              g.drawLine(x1 + 1,y1,x2 + 1,y2);
              g.drawLine(x1,y1 - 1,x2,y2 - 1);
              g.drawLine(x1,y1 + 1,x2,y2 + 1);

              g.setColor(Color.black);
              g.drawLine(x1,y1,x2,y2);
            }
          }

          if (fResources.fnmShowEnvironmentIDs) {
            g.setColor(fResources.knmIDColor);

            int XPos = 0;
            int YPos = 0;

            if (environment.fType == Environment.kRectangle) {
              XPos = (environment.fNodes[0].fXPos + environment.fNodes[1].fXPos) / 2;
              YPos = (environment.fNodes[0].fYPos + environment.fNodes[1].fYPos) / 2;
            }
            else if ((environment.fType == Environment.kPolygon) ||
                     (environment.fType == Environment.kRailroad)) {
              float SumX = 0.0f;
              float SumY = 0.0f;
              for (int i = 0; i < environment.fNrOfNodes; ++i) {
                SumX += environment.fNodes[i].fXPos;
                SumY += environment.fNodes[i].fYPos;
              }
              XPos = (int) Math.round(SumX / environment.fNrOfNodes);
              YPos = (int) Math.round(SumY / environment.fNrOfNodes);
            }
            else if (environment.fType == Environment.kOval) {
              XPos = environment.fNodes[0].fXPos;
              YPos = environment.fNodes[0].fYPos;
            }

            // transform the point to the screen
            int x = (int) Math.round(cx1 + fOriginXOffset + ((XPos / fScaleFactor) * fZoomFactor));
            int y = (int) Math.round(cy2 - (fOriginYOffset + ((YPos / fScaleFactor) * fZoomFactor)));

            g.setColor(fResources.knmIDColor);
            g.drawString(environment.fName,x + 10,y + 10);
          }
        }
      }

      // show the network's links
      if (fNetwork.fLinks.size() != 0) {
        Enumeration LinkEnum = fNetwork.fLinks.elements();

        while (LinkEnum.hasMoreElements()) {
          Link link = (Link) LinkEnum.nextElement();

          // draw straight-lines between the linksegments
          for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {

            // transform the point to the screen
            int x1 = (int) Math.round(cx1 + fOriginXOffset + ((link.fNodes[i].fXPos / fScaleFactor) * fZoomFactor));
            int y1 = (int) Math.round(cy2 - (fOriginYOffset + ((link.fNodes[i].fYPos / fScaleFactor) * fZoomFactor)));
            int x2 = (int) Math.round(cx1 + fOriginXOffset + ((link.fNodes[i + 1].fXPos / fScaleFactor) * fZoomFactor));
            int y2 = (int) Math.round(cy2 - (fOriginYOffset + ((link.fNodes[i + 1].fYPos / fScaleFactor) * fZoomFactor)));

            if (fResources.fnmColorLinksAccordingToType) {
              switch (link.fType) {
                case Link.kTunnel           : g.setColor(fResources.knmTunnelLinkColor);
                                              break;
                case Link.kHighway          : g.setColor(fResources.knmHighwayLinkColor);
                                              break;
                case Link.kSeparatedOnRamp  : g.setColor(fResources.knmSeparatedOnRampLinkColor);
                                              break;
                case Link.kMergedOnRamp     : g.setColor(fResources.knmMergedOnRampLinkColor);
                                              break;
                case Link.kSeparatedOffRamp : g.setColor(fResources.knmSeparatedOffRampLinkColor);
                                              break;
                case Link.kMergedOffRamp    : g.setColor(fResources.knmMergedOffRampLinkColor);
                                              break;
                case Link.kNarrowing        : g.setColor(fResources.knmNarrowingLinkColor);
                                              break;
                case Link.kWidening         : g.setColor(fResources.knmWideningLinkColor);
                                              break;
              }
            }
            else {
              if (link.fSpeedLimit < 20) {
                g.setColor(fResources.knm0To19SpeedColor);
              }
              else if (link.fSpeedLimit < 40) {
                g.setColor(fResources.knm20To39SpeedColor);
              }
              else if (link.fSpeedLimit < 60) {
                g.setColor(fResources.knm40To59SpeedColor);
              }
              else if (link.fSpeedLimit < 80) {
                g.setColor(fResources.knm60To79SpeedColor);
              }
              else if (link.fSpeedLimit < 100) {
                g.setColor(fResources.knm80To99SpeedColor);
              }
              else if (link.fSpeedLimit < 120) {
                g.setColor(fResources.knm100To119SpeedColor);
              }
              else {
                g.setColor(fResources.knm120AndUpSpeedColor);
              }
            }
            g.drawLine(x1,y1,x2,y2);
            g.drawLine(x1 - 1,y1,x2 - 1,y2);
            g.drawLine(x1 + 1,y1,x2 + 1,y2);
            g.drawLine(x1,y1 - 1,x2,y2 - 1);
            g.drawLine(x1,y1 + 1,x2,y2 + 1);

            if (fResources.fnmShowLinkIDs && fResources.fnmSpreadLinkIDs) {
              g.setColor(fResources.knmIDColor);

              int mx = (x1 + x2) / 2;
              int my = (y1 + y2) / 2;

              g.drawString("l",mx,my);
              if (link.fName.length() != 0) {
                g.drawString(String.valueOf(link.fID) + " (" + link.fName + ")",mx + 5,my + 8);
              }
              else {
                g.drawString(String.valueOf(link.fID),mx + 5,my + 8);
              }
            }
          }

          if (fResources.fnmShowLinkIDs && !fResources.fnmSpreadLinkIDs) {
            g.setColor(fResources.knmIDColor);

            int mx = (link.fNodes[0].fXPos + link.fNodes[1].fXPos) / 2;
            int my = (link.fNodes[0].fYPos + link.fNodes[1].fYPos) / 2;

            mx = (int) Math.round(cx1 + fOriginXOffset + ((mx / fScaleFactor) * fZoomFactor));
            my = (int) Math.round(cy2 - (fOriginYOffset + ((my / fScaleFactor) * fZoomFactor)));

            g.drawString("l",mx,my);
            if (link.fName.length() != 0) {
              g.drawString(String.valueOf(link.fID) + " (" + link.fName + ")",mx + 5,my + 8);
            }
            else {
              g.drawString(String.valueOf(link.fID),mx + 5,my + 8);
            }
          }
        }
      }

      // show the network's nodes
      if ((fResources.fnmShowNodes) && (fNetwork.fNodes.size() != 0)) {
        Enumeration NodeEnum = fNetwork.fNodes.elements();

        while (NodeEnum.hasMoreElements()) {
          Node node = (Node) NodeEnum.nextElement();

          // transform the point to the screen
          int x = (int) Math.round(cx1 + fOriginXOffset + ((node.fXPos / fScaleFactor) * fZoomFactor));
          int y = (int) Math.round(cy2 - (fOriginYOffset + ((node.fYPos / fScaleFactor) * fZoomFactor)));

          // draw a big dot at the specified location
          switch (node.fType) {
            case Node.kOrigin      : g.setColor(fResources.knmOriginNodeColor);
                                     break;
            case Node.kDestination : g.setColor(fResources.knmDestinationNodeColor);
                                     break;
            case Node.kJunction    : g.setColor(fResources.knmJunctionNodeColor);
                                     break;
            case Node.kBend        : g.setColor(fResources.knmBendNodeColor);
                                     break;
            case Node.kEnvironment : g.setColor(fResources.knmEnvironmentNodeColor);
                                     break;
          }
          g.fillOval(x - 5,y - 5,10,10);
          g.setColor(Color.black);          
          g.drawOval(x - 5,y - 5,10,10);

          if (fResources.fnmShowNodeIDs) {
            g.setColor(fResources.knmIDColor);
            g.drawString("n",x + 10,y + 10);
            if (node.fName.length() != 0) {
              g.drawString(String.valueOf(node.fID) + " (" + node.fName + ")",x + 18,y + 18);
            }
            else {
              g.drawString(String.valueOf(node.fID),x + 18,y + 18);
            }
          }
        }
      }

      // draw a small marker at the origin
      g.setColor(fResources.knmAxisColor);
      g.drawLine(cx1 + fOriginXOffset - 10,
                 cy2 - fOriginYOffset,
                 cx1 + fOriginXOffset + 10,
                 cy2 - fOriginYOffset);
      g.drawLine(cx1 + fOriginXOffset,
                 cy2 - fOriginYOffset - 10,
                 cx1 + fOriginXOffset,
                 cy2 - fOriginYOffset + 10);
      if (fResources.fnmShowAxes) {
        // draw X-axis
        g.drawLine(cx1,cy2 - fOriginYOffset,cx2,cy2 - fOriginYOffset);
        g.drawLine(cx2 - 10,cy2 - (fOriginYOffset + 10),cx2,cy2 - fOriginYOffset);
        g.drawLine(cx2 - 10,cy2 - (fOriginYOffset - 10),cx2,cy2 - fOriginYOffset);
        g.drawString("X",cx2 - 20,cy2 - (fOriginYOffset - 15));
        // draw Y-axis
        g.drawLine(cx1 + fOriginXOffset,cy1,cx1 + fOriginXOffset,cy2);
        g.drawLine(cx1 + (fOriginXOffset - 10),cy1 + 10,cx1 + fOriginXOffset,cy1);
        g.drawLine(cx1 + (fOriginXOffset + 10),cy1 + 10,cx1 + fOriginXOffset,cy1);
        g.drawString("Y",cx1 + (fOriginXOffset + 5),cy1 + 25);
      }

      if (fResources.fnmShowZoomFactor) {
        // show the zoomfactor
        g.setColor(fResources.knmZoomFactorColor);
        if (fZoomFactor >= 1.0f) {
          g.drawString("Zoomfactor = " + String.valueOf((int) fZoomFactor) + ":1",cx1 + 5,cy1 + 15);
        }
        else {
          g.drawString("Zoomfactor = 1:" + String.valueOf((int) (1.0f / fZoomFactor)),cx1 + 5,cy1 + 15);
        }
      }
    }

    private Point specifyPoint()
    {
      disableControls();
      // prevent the AWT-framework from handling the incoming events
      EventQueue eventQueue = Toolkit.getDefaultToolkit().getSystemEventQueue();

      // retrieve and change the mousecursor
      Cursor OldCursor = getCursor();
      setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

      int kOffset = 5;

      while (true) {
        try {
          AWTEvent event = eventQueue.getNextEvent();
          if (event.getID() == MouseEvent.MOUSE_CLICKED) {
            MouseEvent mouseEvent = (MouseEvent) event;
            Point point = mouseEvent.getPoint();

            // check if the selected point lies on the map
            if ((point.x > (cx1 + kOffset)) && (point.y > (cy1 + kOffset)) &&
                (point.x < (cx2 - kOffset)) && (point.y < (cy2 - kOffset))) {
              // draw a big dot at the specified location
              Graphics g = getGraphics();
              g.setColor(Color.yellow);
              g.fillOval(point.x - 5,point.y - 5,10,10);

              // restore the mousecursor
              setCursor(OldCursor);
              enableControls();
              return point;
            }
          }
        }
        catch (InterruptedException exc) {
        }
      }
    }

    private Node specifyNode()
    {
      disableControls();
      // prevent the AWT-framework from handling the incoming events
      EventQueue eventQueue = Toolkit.getDefaultToolkit().getSystemEventQueue();

      // retrieve and change the mousecursor
      Cursor OldCursor = getCursor();
      setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

      int kOffset = 5;

      while (true) {
        try {
          AWTEvent event = eventQueue.getNextEvent();
          if (event.getID() == MouseEvent.MOUSE_CLICKED) {
            MouseEvent mouseEvent = (MouseEvent) event;

            if (mouseEvent.getClickCount() == 2) {
              // restore the mousecursor
              setCursor(OldCursor);
              enableControls();
              return null;
            }

            Point point = mouseEvent.getPoint();

            // check if the selected point lies on the map
            if ((point.x > (cx1 + kOffset)) && (point.y > (cy1 + kOffset)) &&
                (point.x < (cx2 - kOffset)) && (point.y < (cy2 - kOffset))) {

              // calculate the world-coordinates of the selected point on the screen
              int x = (int) Math.round((((point.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
              int y = (int) Math.round((((cy2 - point.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);

              // find the node closest to the user's selected coordinates
              Enumeration NodeEnum = fNetwork.fNodes.elements();
              Node ClosestNode = (Node) NodeEnum.nextElement();

              float MinDistance = (float) Math.sqrt(sqr(x - ClosestNode.fXPos) +
                                                    sqr(y - ClosestNode.fYPos));

              while (NodeEnum.hasMoreElements()) {
                Node node = (Node) NodeEnum.nextElement();
                float Distance = (float) Math.sqrt(sqr(x - node.fXPos) +
                                                   sqr(y - node.fYPos));
                if (Distance < MinDistance) {
                  MinDistance = Distance;
                  ClosestNode = node;
                }
              }

              // did the user select a point "close-enough" to the closest node ?
              if (((MinDistance / fScaleFactor) * fZoomFactor) <= Resources.fnmSelectionProximity) {
                // restore the mousecursor
                setCursor(OldCursor);
                enableControls();
                return ClosestNode;
              }
            }
          }
        }
        catch (InterruptedException exc) {
        }
      }
    }

    private Link specifyLink()
    {
      disableControls();
      // prevent the AWT-framework from handling the incoming events
      EventQueue eventQueue = Toolkit.getDefaultToolkit().getSystemEventQueue();

      // retrieve and change the mousecursor
      Cursor OldCursor = getCursor();
      setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

      int kOffset = 5;
      int kSelectOffset = 10;

      while (true) {
        try {
          AWTEvent event = eventQueue.getNextEvent();
          if (event.getID() == MouseEvent.MOUSE_CLICKED) {
            MouseEvent mouseEvent = (MouseEvent) event;

            if (mouseEvent.getClickCount() == 2) {
              // restore the mousecursor
              setCursor(OldCursor);
              enableControls();
              return null;
            }

            Point point = mouseEvent.getPoint();

            // check if the selected point lies on the map
            if ((point.x > (cx1 + kOffset)) && (point.y > (cy1 + kOffset)) &&
                (point.x < (cx2 - kOffset)) && (point.y < (cy2 - kOffset))) {

              // calculate the world-coordinates of the selected point on the screen
              int x = (int) Math.round((((point.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
              int y = (int) Math.round((((cy2 - point.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);

              // find the link closest to the user's selected coordinates
              Enumeration LinkEnum = fNetwork.fLinks.elements();
              Link ClosestLink = (Link) LinkEnum.nextElement();

              // calculate the minimum distance from
              // the selected point to the first segment in the first link
              float a = ClosestLink.fNodes[0].fYPos - ClosestLink.fNodes[1].fYPos;
              float b = ClosestLink.fNodes[1].fXPos - ClosestLink.fNodes[0].fXPos;
              float c = (ClosestLink.fNodes[0].fXPos * ClosestLink.fNodes[1].fYPos) -
                        (ClosestLink.fNodes[1].fXPos * ClosestLink.fNodes[0].fYPos);
              float Modulus = (float) Math.sqrt(sqr(a) + sqr(b));
              float MinSegmentDistance = (float) Math.abs(((a * x) + (b * y) + c) / Modulus);

              // since ax+by+c=0 defines an infinite line we have to check
              // wether the selected point is enclosed by the line-segments'
              // bounding-box
              int x1 = ClosestLink.fNodes[0].fXPos;
              int x2 = ClosestLink.fNodes[1].fXPos;
              int y1 = ClosestLink.fNodes[0].fYPos;
              int y2 = ClosestLink.fNodes[1].fYPos;
              if (x1 > x2) {
                int temp = x1;
                x1 = x2;
                x2 = temp;
              }
              if (y1 > y2) {
                int temp = y1;
                y1 = y2;
                y2 = temp;
              }
              // adjust the coordinates so we can still select (almost) horizontal
              // and vertical link-segments
              x1 -= kSelectOffset;
              y1 -= kSelectOffset;
              x2 += kSelectOffset;
              y2 += kSelectOffset;
              boolean DistanceDefined = false;
              if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
                DistanceDefined = true;
              }

              for (int i = 1; i < (ClosestLink.fNrOfNodes - 1); ++i) {
                // consider the segment from node i to node i + 1
                a = ClosestLink.fNodes[i].fYPos - ClosestLink.fNodes[i + 1].fYPos;
                b = ClosestLink.fNodes[i + 1].fXPos - ClosestLink.fNodes[i].fXPos;
                c = (ClosestLink.fNodes[i].fXPos * ClosestLink.fNodes[i + 1].fYPos) -
                    (ClosestLink.fNodes[i + 1].fXPos * ClosestLink.fNodes[i].fYPos);
                Modulus = (float) Math.sqrt(sqr(a) + sqr(b));
                float SegmentDistance = (float) Math.abs(((a * x) + (b * y) + c) / Modulus);

                x1 = ClosestLink.fNodes[0].fXPos;
                x2 = ClosestLink.fNodes[1].fXPos;
                y1 = ClosestLink.fNodes[0].fYPos;
                y2 = ClosestLink.fNodes[1].fYPos;
                if (x1 > x2) {
                  int temp = x1;
                  x1 = x2;
                  x2 = temp;
                }
                if (y1 > y2) {
                  int temp = y1;
                  y1 = y2;
                  y2 = temp;
                }
                x1 -= kSelectOffset;
                y1 -= kSelectOffset;
                x2 += kSelectOffset;
                y2 += kSelectOffset;
                if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
                  if (DistanceDefined) {
                    if (SegmentDistance < MinSegmentDistance) {
                      MinSegmentDistance = SegmentDistance;
                    }                  
                  }
                  else {
                    MinSegmentDistance = SegmentDistance;
                  }
                  DistanceDefined = true;
                }
              }

              // the first link was handled
              float MinDistance = MinSegmentDistance;

              // handle the other links
              while (LinkEnum.hasMoreElements()) {
                Link link = (Link) LinkEnum.nextElement();

                // consider all the segments in this link
                a = link.fNodes[0].fYPos - link.fNodes[1].fYPos;
                b = link.fNodes[1].fXPos - link.fNodes[0].fXPos;
                c = (link.fNodes[0].fXPos * link.fNodes[1].fYPos) -
                    (link.fNodes[1].fXPos * link.fNodes[0].fYPos);
                Modulus = (float) Math.sqrt(sqr(a) + sqr(b));
                MinSegmentDistance = (float) Math.abs(((a * x) + (b * y) + c) / Modulus);

                x1 = link.fNodes[0].fXPos;
                x2 = link.fNodes[1].fXPos;
                y1 = link.fNodes[0].fYPos;
                y2 = link.fNodes[1].fYPos;
                if (x1 > x2) {
                  int temp = x1;
                  x1 = x2;
                  x2 = temp;
                }
                if (y1 > y2) {
                  int temp = y1;
                  y1 = y2;
                  y2 = temp;
                }
                x1 -= kSelectOffset;
                y1 -= kSelectOffset;
                x2 += kSelectOffset;
                y2 += kSelectOffset;

                if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
                  DistanceDefined = true;
                }

                for (int i = 1; i < (link.fNrOfNodes - 1); ++i) {
                  // consider the segment from node i to node i + 1
                  a = link.fNodes[i].fYPos - link.fNodes[i + 1].fYPos;
                  b = link.fNodes[i + 1].fXPos - link.fNodes[i].fXPos;
                  c = (link.fNodes[i].fXPos * link.fNodes[i + 1].fYPos) -
                      (link.fNodes[i + 1].fXPos * link.fNodes[i].fYPos);
                  Modulus = (float) Math.sqrt(sqr(a) + sqr(b));
                  float SegmentDistance = (float) Math.abs(((a * x) + (b * y) + c) / Modulus);

                  x1 = link.fNodes[i].fXPos;
                  x2 = link.fNodes[i + 1].fXPos;
                  y1 = link.fNodes[i].fYPos;
                  y2 = link.fNodes[i + 1].fYPos;
                  if (x1 > x2) {
                    int temp = x1;
                    x1 = x2;
                    x2 = temp;
                  }
                  if (y1 > y2) {
                    int temp = y1;
                    y1 = y2;
                    y2 = temp;
                  }
                  x1 -= kSelectOffset;
                  y1 -= kSelectOffset;
                  x2 += kSelectOffset;
                  y2 += kSelectOffset;
                  if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
                    if (DistanceDefined) {
                      if (SegmentDistance < MinSegmentDistance) {
                        MinSegmentDistance = SegmentDistance;
                      }                  
                    }
                    else {
                      MinSegmentDistance = SegmentDistance;
                    }
                    DistanceDefined = true;
                  }
                }

                if (DistanceDefined) {
                  if (MinSegmentDistance < MinDistance) {
                     MinDistance = MinSegmentDistance;
                     ClosestLink = link;
                  }
                }
              }

              // did the user select a point "close-enough" to the closest link ?
              if (DistanceDefined) {
                if (((MinDistance / fScaleFactor) * fZoomFactor) <= Resources.fnmSelectionProximity) {
                  // restore the mousecursor
                  setCursor(OldCursor);
                  enableControls();
                  return ClosestLink;
                }
              }
            }
          }
        }
        catch (InterruptedException exc) {
        }
      }
    }

    private Environment specifyEnvironment()
    {
      disableControls();
      // prevent the AWT-framework from handling the incoming events
      EventQueue eventQueue = Toolkit.getDefaultToolkit().getSystemEventQueue();

      // retrieve and change the mousecursor
      Cursor OldCursor = getCursor();
      setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

      int kOffset = 5;

      while (true) {
        try {
          AWTEvent event = eventQueue.getNextEvent();
          if (event.getID() == MouseEvent.MOUSE_CLICKED) {
            MouseEvent mouseEvent = (MouseEvent) event;

            if (mouseEvent.getClickCount() == 2) {
              // restore the mousecursor
              setCursor(OldCursor);
              enableControls();
              return null;
            }

            Point point = mouseEvent.getPoint();

            // check if the selected point lies on the map
            if ((point.x > (cx1 + kOffset)) && (point.y > (cy1 + kOffset)) &&
                (point.x < (cx2 - kOffset)) && (point.y < (cy2 - kOffset))) {

              // calculate the world-coordinates of the selected point on the screen
              int x = (int) Math.round((((point.x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
              int y = (int) Math.round((((cy2 - point.y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);

              // find the node closest to the user's selected coordinates
              Enumeration NodeEnum = fNetwork.fNodes.elements();
              Node ClosestNode = (Node) NodeEnum.nextElement();

              float MinDistance = (float) Math.sqrt(sqr(x - ClosestNode.fXPos) +
                                                    sqr(y - ClosestNode.fYPos));

              while (NodeEnum.hasMoreElements()) {
                Node node = (Node) NodeEnum.nextElement();
                float Distance = (float) Math.sqrt(sqr(x - node.fXPos) +
                                                   sqr(y - node.fYPos));
                if (Distance < MinDistance) {
                  MinDistance = Distance;
                  ClosestNode = node;
                }
              }

              // did the user select a point "close-enough" to the closest node ?
              if (((MinDistance / fScaleFactor) * fZoomFactor) <= Resources.fnmSelectionProximity) {

                // find the environment to which the selected node belongs
                Enumeration EnvironmentEnum = fNetwork.fEnvironment.elements();

                while (EnvironmentEnum.hasMoreElements()) {
                  Environment environment = (Environment) EnvironmentEnum.nextElement();

                  // scan all the environment's nodes
                  for (int i = 0; i < environment.fNrOfNodes; ++i) {
                    if (environment.fNodes[i] == ClosestNode) {
                      // restore the mousecursor
                      setCursor(OldCursor);
                      enableControls();
                      return environment;
                    }
                  }
                }
              }
            }
          }
        }
        catch (InterruptedException exc) {
        }
      }
    }

    private float sqr(float x)
    {
      return (x * x);
    }

    // methods from the MouseListener-interface
    public void mouseClicked(MouseEvent e)
    {
      int x = e.getX();
      int y = e.getY();

      if (e.getClickCount() == 2) {
        repaint();
      }
      else {
        // check if the selected point lies on the map
        if ((x >= cx1) && (y >= cy1) && (x <= cx2) && (y <= cy2)) {

          // draw a big cross at the specified location
          Graphics g = getGraphics();
          g.setColor(Color.yellow);
          g.drawLine(x - 5,y - 5,x + 5,y + 5);
          g.drawLine(x + 5,y - 5,x - 5,y + 5);

          // calculate the world-coordinates of the selected point on the screen
          x = (int) Math.round((((x - cx1) - fOriginXOffset) / fZoomFactor) * fScaleFactor);
          y = (int) Math.round((((cy2 - y) - fOriginYOffset) / fZoomFactor) * fScaleFactor);

          // show the coordinates on the screen
          String Message = fMessages.lookup("messageSelectedPoint") + 
                           " : (" + String.valueOf(x) +
                           " m," + String.valueOf(y) + " m)";
          fMessageCanvas.showMessage(Message);
        }
      }
    }
    public void mousePressed(MouseEvent e) { }
    public void mouseReleased(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */  
  // the message-canvas
  private class MessageCanvas extends Canvas
  {
    // the canvas-draw-area
    private int cx1;
    private int cy1;
    private int cx2;
    private int cy2;
    private int cwidth;
    private int cheight;

    public MessageCanvas(int width, int height)
    {
      super();

      setSize(width,height);
    }

    public void update(Graphics g)
    {
      paint(g);
    }

    public void paint(Graphics g)
    {
      // define the boundaries for drawing
      cx1 = 5;
      cy1 = 5;
      cx2 = (cx1 + getSize().width) - (getInsets().right + 5 + 5);
      cy2 = (cy1 + getSize().height) - (getInsets().bottom + 5 + 5);
      cwidth = cx2 - cx1;
      cheight = cy2 - cy1;

      // activate clipping of graphics primitives
      g.setClip(cx1,cy1,cwidth,cheight);
    }

    public void showMessage(String message)
    {
      Graphics g = getGraphics();
      g.clearRect(cx1,cy1,cwidth,cheight);
      g.setColor(fResources.knmMessageColor);
      g.drawString(message,cx1 + 5,cy1 + 18);
    }

    public void clearMessage()
    {
      showMessage("");
    }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class GeneralMenu extends Menu
  {
    public GeneralMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;

      mi = new MenuItem(fMessages.lookup("menuAbout"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuQuit"));
      mi.addActionListener(listener);
      add(mi);
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class ModelMenu extends Menu implements ItemListener
  {
    public ModelMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;
      CheckboxMenuItem ci;

      mi = new MenuItem(fMessages.lookup("menuLoadMap"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuClearMap"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuRelocateMap"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuSetOrigin"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuSetScale"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowLegend"));
      ci.setState(false);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowAxes"));
      ci.setState(fResources.fnmShowAxes);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowZoomFactor"));
      ci.setState(fResources.fnmShowZoomFactor);
      ci.addItemListener(this);
      add(ci);

      addSeparator();

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowNodes"));
      ci.setState(fResources.fnmShowNodes);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowNodeIDs"));
      ci.setState(fResources.fnmShowNodeIDs);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowLinkIDs"));
      ci.setState(fResources.fnmShowLinkIDs);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuSpreadLinkIDs"));
      ci.setState(fResources.fnmSpreadLinkIDs);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuColorLinksAccordingToType"));
      ci.setState(fResources.fnmColorLinksAccordingToType);
      ci.addItemListener(this);
      add(ci);

      ci = new CheckboxMenuItem(fMessages.lookup("menuShowEnvironmentIDs"));
      ci.setState(fResources.fnmShowEnvironmentIDs);
      ci.addItemListener(this);
      add(ci);
    }

    // method from the ItemListener-interface
    public void itemStateChanged(ItemEvent e)
    {
      String Command = (e.getItem()).toString();

      if (Command.equals(fMessages.lookup("menuShowLegend"))) {
        if (fLegend.isVisible()) {
          fLegend.setVisible(false);
        }
        else {
          fLegend.setVisible(true);
        }
      }
      
      if (Command.equals(fMessages.lookup("menuShowAxes"))) {
        fResources.fnmShowAxes = !fResources.fnmShowAxes;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuShowZoomFactor"))) {
        fResources.fnmShowZoomFactor = !fResources.fnmShowZoomFactor;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuShowNodes"))) {
        fResources.fnmShowNodes = !fResources.fnmShowNodes;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuShowNodeIDs"))) {
        fResources.fnmShowNodeIDs = !fResources.fnmShowNodeIDs;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuShowLinkIDs"))) {
        fResources.fnmShowLinkIDs = !fResources.fnmShowLinkIDs;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuSpreadLinkIDs"))) {
        fResources.fnmSpreadLinkIDs = !fResources.fnmSpreadLinkIDs;
        fEditorCanvas.repaint();
      }

      if (Command.equals(fMessages.lookup("menuColorLinksAccordingToType"))) {
        fResources.fnmColorLinksAccordingToType = !fResources.fnmColorLinksAccordingToType;
        fEditorCanvas.repaint();
        fLegend.repaintLegend();
      }

      if (Command.equals(fMessages.lookup("menuShowEnvironmentIDs"))) {
        fResources.fnmShowEnvironmentIDs = !fResources.fnmShowEnvironmentIDs;
        fEditorCanvas.repaint();
      }
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class EditorMenu extends Menu
  {
    public EditorMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;

      mi = new MenuItem(fMessages.lookup("menuGeneralSettings"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuAddNode"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuEditNode"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuMoveNode"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuDeleteNode"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuAddLink"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuEditLink"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuDeleteLink"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuSetLaneChangePermissions"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuAddEnvironment"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuEditEnvironment"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuDeleteEnvironment"));
      mi.addActionListener(listener);
      add(mi);
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class NetworkMenu extends Menu
  {
    public NetworkMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;

      mi = new MenuItem(fMessages.lookup("menuLoadNetwork"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuSaveNetwork"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuClearNetwork"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuAnalyzeNetwork"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuTranslateNetworkByNode"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuTranslateNetworkManually"));
      mi.addActionListener(listener);
      add(mi);
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class NavigateMenu extends Menu
  {
    public NavigateMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;

      mi = new MenuItem(fMessages.lookup("menuMoveLeft"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuMoveRight"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuMoveUp"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuMoveDown"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuGoto0"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuZoomIn"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuZoomOut"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuResetZoom"));
      mi.addActionListener(listener);
      add(mi);

      addSeparator();

      mi = new MenuItem(fMessages.lookup("menuRedrawScreen"));
      mi.addActionListener(listener);
      add(mi);
    }
  } 
}
