// -------------------------------------
// Filename      : RemoteControlGUI.java
// Created       : 07/03/2000
// Last modified : 08/04/2003
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// -------------------------------------

package mitrasim.remotecontrol;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import mitrasim.exceptions.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.server.*;
import mitrasim.simulator.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class RemoteControlGUI extends Frame implements WindowListener, ActionListener
{
  // internal datastructures
  private Resources        fResources;
  private int              fWinXSize;
  private int              fWinYSize;
  private Messages         fMessages;
  private boolean          fAboutBoxShown;
  private Button[]         fControlButtons;
  private final int        kNrOfControlButtons            = 12;
  private final int        kConnectButton                 = 0;
  private final int        kDisconnectButton              = 1;
  private final int        kShowConnectionLogButton       = 2;
  private final int        kUploadScenarioButton          = 3;
  private final int        kSetCycleHoldTimeButton        = 4;
  private final int        kSetZeroCycleHoldTimeButton    = 5;
  private final int        kSetAverageCycleHoldTimeButton = 6;
  private final int        kStartContinueSimulatorButton  = 7;
  private final int        kStopSimulatorButton           = 8;
  private final int        kResetSimulatorButton          = 9;
  private final int        kShutdownServerButton          = 10;
  private final int        kQuitButton                    = 11;
  private boolean          fConnected;
  private Socket           fSocket;
  private DataInputStream  fIn;
  private DataOutputStream fOut;
  private String           fIPAddress;
  private int              fTCPPort;
  private FileParser       fScenarioFile;

  public RemoteControlGUI()
  {
    super();

    fResources = new Resources();
    fWinXSize = fResources.frcXSize;
    fWinYSize = fResources.frcYSize;
    fMessages = fResources.fMessages;
    fIPAddress = "127.0.0.1";
    fTCPPort = 4242;
    fScenarioFile = null;
    fResources.fNetwork = new Network(fMessages);

    // adjust the windowlayout
    fAboutBoxShown = false;
    setFont(new Font(fMessages.lookup("textFontName"),Font.PLAIN,14));
    setSize(fWinXSize,fWinYSize);
    setResizable(false);
    setLocation(0,0);
    setTitle(fMessages.lookup("textRemoteControlWindowTitle"));
    setBackground(Color.lightGray);
    setForeground(Color.black);
    setIconImage(Toolkit.getDefaultToolkit().getImage("M2000Icon.jpg"));

    setLayout(new BorderLayout());

    // create the panel containing the control-commands
    Panel ControlPanel = new Panel();
    ControlPanel.setLayout(new GridLayout(kNrOfControlButtons + 3,1));

    fControlButtons = new Button[kNrOfControlButtons];
    fControlButtons[kConnectButton] = new Button(fMessages.lookup("buttonConnect"));
    fControlButtons[kDisconnectButton] = new Button(fMessages.lookup("buttonDisconnect"));
    fControlButtons[kShowConnectionLogButton] = new Button(fMessages.lookup("buttonShowConnectionLog"));
    fControlButtons[kUploadScenarioButton] = new Button(fMessages.lookup("buttonUploadScenario"));
    fControlButtons[kSetCycleHoldTimeButton] = new Button(fMessages.lookup("buttonSetCycleHoldTime"));
    fControlButtons[kSetZeroCycleHoldTimeButton] = new Button(fMessages.lookup("buttonSetZeroCycleHoldTime"));
    fControlButtons[kSetAverageCycleHoldTimeButton] = new Button(fMessages.lookup("buttonSetAverageCycleHoldTime"));
    fControlButtons[kStartContinueSimulatorButton] = new Button(fMessages.lookup("buttonStartContinueSimulator"));
    fControlButtons[kStopSimulatorButton] = new Button(fMessages.lookup("buttonStopSimulator"));
    fControlButtons[kResetSimulatorButton] = new Button(fMessages.lookup("buttonResetSimulator"));
    fControlButtons[kShutdownServerButton] = new Button(fMessages.lookup("buttonShutdownServer"));
    fControlButtons[kQuitButton] = new Button(fMessages.lookup("buttonQuit"));
    fControlButtons[kQuitButton].setBackground(Color.red);

    for (int i = 0; i < kNrOfControlButtons; ++i) {
      fControlButtons[i].addActionListener(this);
    }

    Label label = new Label("Remote Control",Label.CENTER);
    label.setFont(new Font(fMessages.lookup("textFontName"),Font.BOLD,18));
    ControlPanel.add(label);
    ControlPanel.add(fControlButtons[kConnectButton]);
    ControlPanel.add(fControlButtons[kDisconnectButton]);
    ControlPanel.add(fControlButtons[kShowConnectionLogButton]);
    ControlPanel.add(new Label());
    ControlPanel.add(fControlButtons[kUploadScenarioButton]);
    ControlPanel.add(fControlButtons[kSetCycleHoldTimeButton]);
    ControlPanel.add(fControlButtons[kSetZeroCycleHoldTimeButton]);
    ControlPanel.add(fControlButtons[kSetAverageCycleHoldTimeButton]);
    ControlPanel.add(fControlButtons[kStartContinueSimulatorButton]);
    ControlPanel.add(fControlButtons[kStopSimulatorButton]);
    ControlPanel.add(fControlButtons[kResetSimulatorButton]);
    ControlPanel.add(new Label());
    ControlPanel.add(fControlButtons[kShutdownServerButton]);
    ControlPanel.add(fControlButtons[kQuitButton]);

    add("Center",ControlPanel);

    // add the windowlistener
    addWindowListener(this);

    pack();
    show();

    if (fResources.frcShowAboutBox) {
      // show the about-box
      AboutBox about = new AboutBox(this,"Remote Control");
    }
    fAboutBoxShown = true;

    fConnected = false;
    setDefaultControls();
  }

  private void setDefaultControls()
  {
    fControlButtons[kConnectButton].setEnabled(true);
    fControlButtons[kDisconnectButton].setEnabled(false);
    fControlButtons[kShowConnectionLogButton].setEnabled(false);
    fControlButtons[kUploadScenarioButton].setEnabled(false);
    fControlButtons[kSetCycleHoldTimeButton].setEnabled(false);
    fControlButtons[kSetZeroCycleHoldTimeButton].setEnabled(false);
    fControlButtons[kSetAverageCycleHoldTimeButton].setEnabled(false);
    fControlButtons[kStartContinueSimulatorButton].setEnabled(false);
    fControlButtons[kStopSimulatorButton].setEnabled(false);
    fControlButtons[kResetSimulatorButton].setEnabled(false);
    fControlButtons[kShutdownServerButton].setEnabled(false);
    fControlButtons[kQuitButton].setEnabled(true);
  }

  // methods from the WindowListener-interface
  public void windowOpened(WindowEvent e) { }
  public void windowClosing(WindowEvent e)
  {
    // quit the running program
    System.exit(0);
  }
  public void windowClosed(WindowEvent e) { }
  public void windowIconified(WindowEvent e) { }
  public void windowDeiconified(WindowEvent e) { }
  public void windowActivated(WindowEvent e) { }
  public void windowDeactivated(WindowEvent e) { }

  // method from the ActionListener-interface
  public void actionPerformed(ActionEvent e)
  {
    String Command = e.getActionCommand();

    if (Command.equals(fMessages.lookup("buttonConnect"))) {
      connect();
    }

    if (Command.equals(fMessages.lookup("buttonDisconnect"))) {
      disconnect();
    }

    if (Command.equals(fMessages.lookup("buttonShowConnectionLog"))) {
      showConnectionLog();
    }

    if (Command.equals(fMessages.lookup("buttonUploadScenario"))) {
      uploadScenario();
    }

    if (Command.equals(fMessages.lookup("buttonStartContinueSimulator"))) {
      sendSimulationCommandToServer(RemoteControlProtocol.kStartContinueSimulator,
                                    fMessages.lookup("errorFailedToStartContinueSimulator"));
    }

    if (Command.equals(fMessages.lookup("buttonSetCycleHoldTime"))) {
      setCycleHoldTime();
    }

    if (Command.equals(fMessages.lookup("buttonSetZeroCycleHoldTime"))) {
      setSpecialCycleHoldTime(0);
    }

    if (Command.equals(fMessages.lookup("buttonSetAverageCycleHoldTime"))) {
      setSpecialCycleHoldTime(100);
    }

    if (Command.equals(fMessages.lookup("buttonStopSimulator"))) {
      sendSimulationCommandToServer(RemoteControlProtocol.kStopSimulator,
                                    fMessages.lookup("errorFailedToStopSimulator"));
    }

    if (Command.equals(fMessages.lookup("buttonResetSimulator"))) {
      sendSimulationCommandToServer(RemoteControlProtocol.kResetSimulator,
                                    fMessages.lookup("errorFailedToResetSimulator"));
    }

    if (Command.equals(fMessages.lookup("buttonShutdownServer"))) {
      shutdownServer();
    }

    if (Command.equals(fMessages.lookup("buttonQuit"))) {
      windowClosing(null);
    }
  }

  private void connect()
  {
    try {
    	if (Resources.kDisableAutoConnect) {

	      ConnectBox connectBox = new ConnectBox(this,fIPAddress,fTCPPort);

  	    if (connectBox.canceled()) {
    	    return;
      	}

	      fIPAddress = connectBox.getIPAddress();
  	    fTCPPort = connectBox.getTCPPort();
			}

      fSocket = new Socket(fIPAddress,fTCPPort);
      fIn = new DataInputStream(fSocket.getInputStream());
      fOut = new DataOutputStream(fSocket.getOutputStream());

      // acquire the server's identification-string
      String ServerIDString = fIn.readUTF();

      // request to use the remote control-protocol
      fOut.writeInt(Protocol.kRemoteControlProtocol);
      fOut.flush();

      // check if the server agrees
      int Indicator = fIn.readInt();

      if (Indicator != Protocol.kProtocolOk) {
        if (Indicator == Protocol.kInvalidProtocolRequested) {
          MessageBox ErrorBox = new MessageBox(this,
                                               fMessages.lookup("errorInvalidProtocolRequested"),
                                               MessageBox.kError);
        }
        else if (Indicator == Protocol.kProtocolAlreadyInUse) {
          MessageBox ErrorBox = new MessageBox(this,
                                               fMessages.lookup("errorAlreadyRemoteControlled"),
                                               MessageBox.kError);
        }
        fSocket.close();
        return;
      }

    	if (Resources.kDisableAutoConnect) {
	      // indicate success
  	    MessageBox messageBox = new MessageBox(this,
    	                                         ServerIDString,
      	                                       MessageBox.kMessage);
			}

      fConnected = true;

      fControlButtons[kConnectButton].setEnabled(false);
      fControlButtons[kDisconnectButton].setEnabled(true);
      fControlButtons[kShowConnectionLogButton].setEnabled(true);
      fControlButtons[kUploadScenarioButton].setEnabled(true);
      fControlButtons[kSetCycleHoldTimeButton].setEnabled(true);
      fControlButtons[kSetZeroCycleHoldTimeButton].setEnabled(true);
      fControlButtons[kSetAverageCycleHoldTimeButton].setEnabled(true);
      fControlButtons[kShutdownServerButton].setEnabled(true);

      setSimulatorButtonStates();

      fControlButtons[kQuitButton].setEnabled(false);
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToConnectWithServer"),
                                           MessageBox.kError);
    }
  }

  private void disconnect()
  {
    if (!fConnected) {
      return;
    }

    try {
      fOut.writeInt(RemoteControlProtocol.kDisconnect);
      fOut.flush();
      boolean Ok = fIn.readBoolean();
      if (!Ok) {
        throw (new IOException());
      }

      fSocket.close();

      fConnected = false;
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToDisconnectFromServer"),
                                           MessageBox.kError);
    }

    setDefaultControls();
  }

  private void showConnectionLog()
  {
    try {
      fOut.writeInt(RemoteControlProtocol.kShowConnectionLog);
      fOut.flush();
      Report ConnectionLog = new Report();
      ConnectionLog.readFromStream(fIn);

      // show the connectionlog
      ReportBox reportBox = new ReportBox(this,ConnectionLog,
                                          fMessages.lookup("textConnectionLogDialogTitle"),
                                          "",
                                          ReportBox.kShowTickMarks);
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToShowConnectionLog"),
                                           MessageBox.kError);
      setDefaultControls();
    }
  }

  private void setCycleHoldTime()
  {
    SetCycleHoldTimeBox setCycleHoldTimeBox = new SetCycleHoldTimeBox(this);

    if (setCycleHoldTimeBox.canceled()) {
      return;
    }

    try {
      fOut.writeInt(RemoteControlProtocol.kSetCycleHoldTime);
      fOut.flush();

      fOut.writeInt(fResources.fsimCycleHoldTime);
      fOut.flush();
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToSetCycleHoldTime"),
                                           MessageBox.kError);
      setDefaultControls();
    }
  }

  private void setSpecialCycleHoldTime(int ms)
  {
    try {
      fOut.writeInt(RemoteControlProtocol.kSetCycleHoldTime);
      fOut.flush();

			fResources.fsimCycleHoldTime = ms;

      fOut.writeInt(fResources.fsimCycleHoldTime);
      fOut.flush();
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToSetCycleHoldTime"),
                                           MessageBox.kError);
      setDefaultControls();
    }
  }

  private void uploadScenario()
  {
    if (!loadScenarioFile()) {
      return;
    }

    try {
      fOut.writeInt(RemoteControlProtocol.kUploadScenario);
      fOut.flush();

      // transfer the network
      fResources.fNetwork.writeToStream(fOut);

      boolean UploadOk = fIn.readBoolean();
      if (!UploadOk) {
        throw (new IOException());
      }

      // transfer the simulation parameters
      fOut.writeInt(fResources.fsimStartTime);
      fOut.writeInt(fResources.fsimPercentageMiddleLaneDrivers);
      fOut.writeFloat(fResources.fsimReactionTimeMean);
      fOut.writeFloat(fResources.fsimReactionTimeStdDev);
      fOut.writeFloat(fResources.fsimReactionTimeMin);
      fOut.writeFloat(fResources.fsimReactionTimeMax);
      fOut.writeFloat(fResources.fsimInitialVehicleSpeed);
      fOut.writeInt(fResources.fsimCycleHoldTime);
      fOut.writeFloat(fResources.fsimMinSpaceAtEntrance);
      fOut.writeBoolean(fResources.fsimAggressionAndExcessCoupled);
      fOut.writeFloat(fResources.fsimGravitationalConstant);
      fOut.writeFloat(fResources.fsimCarFollowingMinSpaceHeadway);
      fOut.writeFloat(fResources.fsimCarFollowingMinSpaceHeadwayExcess);
      fOut.writeFloat(fResources.fsimCarFollowingMinTimeHeadway);
      fOut.writeFloat(fResources.fsimCarFollowingMinTimeHeadwayExcess);
      fOut.writeFloat(fResources.fsimCarFollowingMaxTimeHeadway);
      fOut.writeFloat(fResources.fsimCarFollowingMaxTimeHeadwayExcess);
      fOut.writeFloat(fResources.fsimCarFollowingAlphaMin);
      fOut.writeFloat(fResources.fsimCarFollowingBetaMin);
      fOut.writeFloat(fResources.fsimCarFollowingGammaMin);
      fOut.writeFloat(fResources.fsimCarFollowingAlphaPlus);
      fOut.writeFloat(fResources.fsimCarFollowingBetaPlus);
      fOut.writeFloat(fResources.fsimCarFollowingGammaPlus);
      fOut.writeFloat(fResources.fsimCriticalRampDistance);
      fOut.writeInt(fResources.fsimPowerOfMLCProbability);
      fOut.writeFloat(fResources.fsimLaneChangingMinLeadGap);
      fOut.writeFloat(fResources.fsimLaneChangingMinLeadGapExcess);
      fOut.writeFloat(fResources.fsimLaneChangingMinLagGap);
      fOut.writeFloat(fResources.fsimLaneChangingMinLagGapExcess);
      fOut.writeInt(fResources.fsimAllowTrucksOnLanes);
      fOut.writeInt(fResources.fsimMinTimeLeftDuringOvertaking);

      // transfer the routing table
      fOut.writeInt(fResources.fsimRoutingTable.size());
      Enumeration RoutingTableEnum = fResources.fsimRoutingTable.elements();
      while (RoutingTableEnum.hasMoreElements()) {
        RoutingTableEntry routingTableEntry = (RoutingTableEntry) RoutingTableEnum.nextElement();

        fOut.writeInt(routingTableEntry.fOriginNode.fID);
        fOut.writeInt(routingTableEntry.fFlowRate);
        fOut.writeInt(routingTableEntry.fDistributionFactor);
        fOut.writeInt(routingTableEntry.fFleetMix);
        fOut.writeInt(routingTableEntry.fNrOfRoutes);
        for (int i = 0; i < routingTableEntry.fNrOfRoutes; ++i) {
          fOut.writeInt(routingTableEntry.fPercentages[i]);
        }
        for (int i = 0; i < routingTableEntry.fNrOfRoutes; ++i) {
          fOut.writeInt(routingTableEntry.fRoutes[i].length);
          for (int j = 0; j < routingTableEntry.fRoutes[i].length; ++j) {
            fOut.writeInt(routingTableEntry.fRoutes[i][j].fID);
          }
        }
      }

      // transfer the loopdetectors
      fOut.writeInt(fResources.fsimLoopDetectors.size());
      Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
      while (LoopDetectorEnum.hasMoreElements()) {
        LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

        fOut.writeUTF(loopDetector.fName);
        fOut.writeInt(loopDetector.fLocation.fID);
        fOut.writeInt(loopDetector.fTimeForMeasurement);
      }

      // transfer the view-specifications
      fOut.writeInt(fResources.fvisViewSpecifications.size());
      Enumeration ViewSpecificationEnum = fResources.fvisViewSpecifications.elements();
      while (ViewSpecificationEnum.hasMoreElements()) {
        ViewSpecification viewSpecification = (ViewSpecification) ViewSpecificationEnum.nextElement();

        fOut.writeUTF(viewSpecification.fName);
        fOut.writeBoolean(viewSpecification.fPerspectiveMode);
        fOut.writeInt(viewSpecification.fViewPointX);
        fOut.writeInt(viewSpecification.fViewPointY);
        fOut.writeInt(viewSpecification.fViewPointZ);
        fOut.writeInt(viewSpecification.fEyePointRho);
        fOut.writeInt(viewSpecification.fEyePointPhi);
        fOut.writeInt(viewSpecification.fEyePointTheta);
        fOut.writeInt(viewSpecification.fBoundingBoxMinX);
        fOut.writeInt(viewSpecification.fBoundingBoxMinY);
        fOut.writeInt(viewSpecification.fBoundingBoxMaxX);
        fOut.writeInt(viewSpecification.fBoundingBoxMaxY);
        fOut.writeBoolean(viewSpecification.fShowEnvironment);
      }
      fOut.flush();

      setSimulatorButtonStates();
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToUploadScenario"),
                                           MessageBox.kError);
      setDefaultControls();
    }
  }

  private boolean loadScenarioFile()
  {
    FileDialog fd = new FileDialog(this,
                                   fMessages.lookup("textLoadScenarioDialogTitle"),
                                   FileDialog.LOAD);
    fd.show();
    String Dir = fd.getDirectory();
    String Filename = fd.getFile();
    if ((Dir != null) && (Filename != null)) {
      String ScenarioFilename = Dir + Filename;

      try {
        // let the fileparser open the file containing the scenario-description
        fScenarioFile = new FileParser(fMessages,ScenarioFilename);

        // retrieve the simulator's settings (network, starttime, parameters, ...)
        readScenarioSimulatorSection();

        // retrieve the routing table
        readScenarioRoutingTableSection();

        // reserve room for 20 loopdetectors and a load factor of 80%
        fResources.fsimLoopDetectors = new Hashtable(20,0.8f);

        // retrieve the loopdetectors
        readScenarioLoopDetectorsSection();

        // reserve room for 20 view-specifications and a load factor of 80%
        fResources.fvisViewSpecifications = new Hashtable(20,0.8f);

        // retrieve the view-specifications
        readScenarioViewSpecificationsSection();

        // let the fileparser close the datafile
        fScenarioFile.close();
      }
      catch (GeneralError err) {
        MessageBox ErrorBox = new MessageBox(this,err.message(),MessageBox.kError);
        return false;
      }

      return true;
    }
    else {
      return false;
    }
  }

  private void readScenarioSimulatorSection() throws GeneralError
  {
   fScenarioFile.readSectionHeader("Simulator");

   fScenarioFile.readNextLineFromFile();

    // retrieve the scenario's network-file
   String NetworkFilename = fScenarioFile.separatePrefix();

   // create and check the network
   fResources.fNetwork.clear();
   fResources.fNetwork.loadFromFile(NetworkFilename);

   // check the network
   Report ResultReport = fResources.fNetwork.analyse();
   if (!ResultReport.isEmpty()) {
     throw (new GeneralError(fMessages.lookup("errorNetworkDoesntComply"),
                             GeneralError.kNetworkError));
   }

   fResources.fNetwork.calcGeometry();

   // retrieve the simulator's starttime
   String StartTime = fScenarioFile.separatePrefix();
   if (StartTime.length() != 5) {
     throw (new GeneralError(fMessages.lookup("errorInvalidStartTimeSpecified",
                             StartTime,String.valueOf(fScenarioFile.fLineNr)),
                             GeneralError.kParseError));
   }

   int Hours = 0;
   int Minutes = 0;
   try {
     Hours = (Integer.valueOf(StartTime.substring(0,2))).intValue();
     Minutes = (Integer.valueOf(StartTime.substring(3))).intValue();
   }
   catch (NumberFormatException exc) {
     throw (new GeneralError(fMessages.lookup("errorInvalidStartTimeSpecified",
                             StartTime,String.valueOf(fScenarioFile.fLineNr)),
                             GeneralError.kParseError));
   }

   // convert the starttime to a value expressed in
   // tenths of seconds (= the simulator's stepsize)
   fResources.fsimStartTime = (Minutes * 10 * 60) +
                              (Hours * 10 * 60 * 60);

   // retrieve the percentage middle-lane drivers
   String PercentageMiddleLaneDriversString = fScenarioFile.separatePrefix();
   try {
     fResources.fsimPercentageMiddleLaneDrivers = (Integer.valueOf(PercentageMiddleLaneDriversString)).intValue();
     if ((fResources.fsimPercentageMiddleLaneDrivers < 0) ||
         (fResources.fsimPercentageMiddleLaneDrivers > 100)) {
       throw (new NumberFormatException());
     }
   }
   catch (NumberFormatException exc) {
     throw (new GeneralError(fMessages.lookup("errorInvalidPercentageMiddleLaneDriversSpecified",
                             PercentageMiddleLaneDriversString,
                             String.valueOf(fScenarioFile.fLineNr)),
                             GeneralError.kParseError));
   }

   // retrieve the other simulation parameters
   try {
     fResources.fsimReactionTimeMean = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimReactionTimeStdDev = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimReactionTimeMin = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimReactionTimeMax = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimInitialVehicleSpeed = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimInitialVehicleSpeed < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCycleHoldTime = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
     if (fResources.fsimCycleHoldTime < 0) {
       throw (new NumberFormatException());
     }

     fResources.fsimMinSpaceAtEntrance = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimMinSpaceAtEntrance < 0.0f) {
       throw (new NumberFormatException());
     }

     String CoupleExcessAndAggression = fScenarioFile.separatePrefix();
     if (CoupleExcessAndAggression.equalsIgnoreCase("YES")) {
       fResources.fsimAggressionAndExcessCoupled = true;
     }
     else if (CoupleExcessAndAggression.equalsIgnoreCase("NO")) {
       fResources.fsimAggressionAndExcessCoupled = false;
     }
     else {
       throw (new GeneralError(fMessages.lookup("errorInvalidCoupleExcessAndAggression",
                               CoupleExcessAndAggression,
                               String.valueOf(fScenarioFile.fLineNr)),
                               GeneralError.kParseError));
     }

     fResources.fsimGravitationalConstant = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingMinSpaceHeadway = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMinSpaceHeadway < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingMinSpaceHeadwayExcess = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMinSpaceHeadwayExcess < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingMinTimeHeadway = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMinTimeHeadway < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingMinTimeHeadwayExcess = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMinTimeHeadwayExcess < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingMaxTimeHeadway = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMaxTimeHeadway < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingMaxTimeHeadwayExcess = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimCarFollowingMaxTimeHeadwayExcess < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimCarFollowingAlphaMin = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingBetaMin = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingGammaMin = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingAlphaPlus = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingBetaPlus = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCarFollowingGammaPlus = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();

     fResources.fsimCriticalRampDistance = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if ((fResources.fsimCriticalRampDistance < 0.0f) ||
         (fResources.fsimCriticalRampDistance > 1.0f)) {
       throw (new NumberFormatException());
     }

     fResources.fsimPowerOfMLCProbability = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
     if (fResources.fsimPowerOfMLCProbability < 0) {
       throw (new NumberFormatException());
     }

     fResources.fsimLaneChangingMinLeadGap = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimLaneChangingMinLeadGap < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimLaneChangingMinLeadGapExcess = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimLaneChangingMinLeadGapExcess < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimLaneChangingMinLagGap = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimLaneChangingMinLagGap < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimLaneChangingMinLagGapExcess = (Float.valueOf(fScenarioFile.separatePrefix())).floatValue();
     if (fResources.fsimLaneChangingMinLagGapExcess < 0.0f) {
       throw (new NumberFormatException());
     }

     fResources.fsimAllowTrucksOnLanes = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
     if (fResources.fsimAllowTrucksOnLanes < 1) {
       throw (new NumberFormatException());
     }

     fResources.fsimMinTimeLeftDuringOvertaking = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
     if (fResources.fsimMinTimeLeftDuringOvertaking < 0) {
       throw (new NumberFormatException());
     }
   }
   catch (NumberFormatException exc) {
     throw (new GeneralError(fMessages.lookup("errorInvalidNumber",
                                              String.valueOf(fScenarioFile.fLineNr)),
                             GeneralError.kParseError));
   }

   fScenarioFile.readNextLineFromFile();

   // does the section-block ends with '}' ?
   fScenarioFile.checkForSectionClosure();
  }

  private void readScenarioRoutingTableSection() throws GeneralError
  {
    fScenarioFile.readSectionHeader("RoutingTable");

    fResources.fsimRoutingTable = new Vector();

    boolean ContinueReading = true;
    do {
      fScenarioFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fScenarioFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {

          RoutingTableEntry routingTableEntry = new RoutingTableEntry();

          int OriginNodeID = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

          routingTableEntry.fOriginNode = (Node) fResources.fNetwork.fNodes.get(new Integer(OriginNodeID));
          if (routingTableEntry.fOriginNode == null) {
            throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",
                                    String.valueOf(OriginNodeID),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          if (routingTableEntry.fOriginNode.fType != Node.kOrigin) {
            throw (new GeneralError(fMessages.lookup("errorInvalidFirstNodeTypeInRoute",
                                    String.valueOf(routingTableEntry.fOriginNode.fID),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          routingTableEntry.fFlowRate = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          if (routingTableEntry.fFlowRate < 0) {
            throw (new GeneralError(fMessages.lookup("errorInvalidFlowRateSpecified",
                                    String.valueOf(routingTableEntry.fFlowRate),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          routingTableEntry.fDistributionFactor = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          if ((routingTableEntry.fDistributionFactor < 0) ||
              (routingTableEntry.fDistributionFactor > 100)) {
            throw (new GeneralError(fMessages.lookup("errorInvalidDistributionFactorSpecified",
                                    String.valueOf(routingTableEntry.fDistributionFactor),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          routingTableEntry.fFleetMix = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          if ((routingTableEntry.fFleetMix < 0) ||
              (routingTableEntry.fFleetMix > 100)) {
            throw (new GeneralError(fMessages.lookup("errorInvalidFleetMixSpecified",
                                    String.valueOf(routingTableEntry.fFleetMix),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          routingTableEntry.fNrOfRoutes = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          if (routingTableEntry.fNrOfRoutes < 1) {
            throw (new GeneralError(fMessages.lookup("errorInvalidNrOfRoutesSpecified",
                                    String.valueOf(routingTableEntry.fNrOfRoutes),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          routingTableEntry.fPercentages = new int[routingTableEntry.fNrOfRoutes];
          routingTableEntry.fRoutes = new Node[routingTableEntry.fNrOfRoutes][];

          for (int i = 0; i < routingTableEntry.fNrOfRoutes; ++i) {
            routingTableEntry.fPercentages[i] = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

            int NrOfNodes = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
            routingTableEntry.fRoutes[i] = new Node[NrOfNodes];

            for (int j = 0; j < NrOfNodes; ++j) {
              int NodeNr = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

              routingTableEntry.fRoutes[i][j] = (Node) fResources.fNetwork.fNodes.get(new Integer(NodeNr));
              if (routingTableEntry.fRoutes[i][j] == null) {
                throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",
                                        String.valueOf(NodeNr),
                                        String.valueOf(fScenarioFile.fLineNr)),
                                        GeneralError.kParseError));
              }

              // check the node's type
              if (j < (NrOfNodes - 1)) {
                if (routingTableEntry.fRoutes[i][j].fType != Node.kJunction) {
                  throw (new GeneralError(fMessages.lookup("errorInvalidNodeTypeInRoute",
                                          String.valueOf(routingTableEntry.fRoutes[i][j].fID),
                                          String.valueOf(fScenarioFile.fLineNr)),
                                          GeneralError.kParseError));
                }
              }
              else if (routingTableEntry.fRoutes[i][j].fType != Node.kDestination) {
                throw (new GeneralError(fMessages.lookup("errorInvalidLastNodeTypeInRoute",
                                        String.valueOf(routingTableEntry.fRoutes[i][j].fID),
                                        String.valueOf(fScenarioFile.fLineNr)),
                                        GeneralError.kParseError));
              }
            }

            // check the connectivity of the route
            for (int j = 0; j < NrOfNodes; ++j) {
              Node PreviousNode = null;
              if (j == 0) {
                PreviousNode = routingTableEntry.fOriginNode;
              }
              else{
                PreviousNode = routingTableEntry.fRoutes[i][j - 1];
              }
              Node CurrentNode = routingTableEntry.fRoutes[i][j];

              // find the link to which the previous node belongs
              boolean Found = false;
              Enumeration LinkEnum = fResources.fNetwork.fLinks.elements();
              while (LinkEnum.hasMoreElements() && (!Found)) {
                Link link = (Link) LinkEnum.nextElement();

                if (link.fNodes[0].fID == PreviousNode.fID) {

                  // try to find a link which fits this piece of the route
                  Enumeration LinkFindEnum = fResources.fNetwork.fLinks.elements();
                  while (LinkFindEnum.hasMoreElements() && (!Found)) {
                    Link TestLink = (Link) LinkFindEnum.nextElement();

                    if (TestLink.fNodes[TestLink.fNrOfNodes - 1].fID == CurrentNode.fID) {
                      Found = true;
                    }
                  }

                  if (!Found) {
                    throw (new GeneralError(fMessages.lookup("errorRouteNotConnected",
                                            String.valueOf(PreviousNode.fID),
                                            String.valueOf(link.fID),
                                            String.valueOf(fScenarioFile.fLineNr)),
                                            GeneralError.kParseError));
                  }
                }
              }
            }
          }

          // check if all the percentages sum up to 100
          int Sum = 0;
          for (int i = 0; i < routingTableEntry.fNrOfRoutes; ++i) {
            Sum += routingTableEntry.fPercentages[i];
          }
          if (Sum != 100) {
            throw (new GeneralError(fMessages.lookup("errorPercentagesDontSumUp",
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          // add the routing table entry to the routing table
          fResources.fsimRoutingTable.addElement(routingTableEntry);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",
                                  String.valueOf(fScenarioFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readScenarioLoopDetectorsSection() throws GeneralError
  {
    fScenarioFile.readSectionHeader("LoopDetectors");

    boolean ContinueReading = true;
    do {
      fScenarioFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fScenarioFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          int NodeID = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          Node Location = (Node) fResources.fNetwork.fNodes.get(new Integer(NodeID));
          if (Location == null) {
            throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",
                                    String.valueOf(NodeID),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }
          if ((Location.fType != Node.kDestination) &&
              (Location.fType != Node.kJunction)) {
            throw (new GeneralError(fMessages.lookup("errorInvalidTypeForLoopDetectorLocation",
                                    String.valueOf(NodeID),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          // check if the location isn't already occupied by a loopdetector
          LoopDetector FindLoopDetector = (LoopDetector) fResources.fsimLoopDetectors.get(new Integer(Location.fID));
          if (FindLoopDetector != null) {
            throw (new GeneralError(fMessages.lookup("errorNodeAlreadyOccupiedByLoopDetector",
                                    String.valueOf(NodeID),
                                    String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          int TimeForMeasurement = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          if (TimeForMeasurement < 0) {
            throw (new NumberFormatException());
          }

          String Name = fScenarioFile.separatePrefix();

          // create a loopdetector based on the information just retrieved
          LoopDetector loopDetector = new LoopDetector(Name,Location,TimeForMeasurement);

          // add the loopdetector to the table of loopdetectors
          fResources.fsimLoopDetectors.put(new Integer(Location.fID),loopDetector);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",
                                  String.valueOf(fScenarioFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readScenarioViewSpecificationsSection() throws GeneralError
  {
    fScenarioFile.readSectionHeader("ViewSpecifications");

    boolean ContinueReading = true;
    do {
      fScenarioFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fScenarioFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          ViewSpecification viewSpecification = new ViewSpecification();

          viewSpecification.fName = fScenarioFile.separatePrefix();

          String ProjectionMode = fScenarioFile.separatePrefix();
          if (ProjectionMode.equalsIgnoreCase("PARALLEL")) {
            viewSpecification.fPerspectiveMode = false;
          }
          else if (ProjectionMode.equalsIgnoreCase("PERSPECTIVE")) {
            viewSpecification.fPerspectiveMode = true;
          }
          else {
            throw (new GeneralError(fMessages.lookup("errorInvalidProjectionMode",
                                    ProjectionMode,String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          viewSpecification.fViewPointX = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fViewPointY = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fViewPointZ = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

          viewSpecification.fEyePointRho = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fEyePointPhi = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fEyePointTheta = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

          viewSpecification.fBoundingBoxMinX = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fBoundingBoxMinY = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fBoundingBoxMaxX = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();
          viewSpecification.fBoundingBoxMaxY = (Integer.valueOf(fScenarioFile.separatePrefix())).intValue();

          String ShowEnvironment = fScenarioFile.separatePrefix();
          if (ShowEnvironment.equalsIgnoreCase("YES")) {
            viewSpecification.fShowEnvironment = true;
          }
          else if (ShowEnvironment.equalsIgnoreCase("NO")) {
            viewSpecification.fShowEnvironment = false;
          }
          else {
            throw (new GeneralError(fMessages.lookup("errorInvalidShowEnvironment",ShowEnvironment,String.valueOf(fScenarioFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          // add the view-specification to the table of view-specifications
          fResources.fvisViewSpecifications.put(new String(viewSpecification.fName),viewSpecification);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fScenarioFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void sendSimulationCommandToServer(int Command, String ErrorMessage)
  {
    try {
      fOut.writeInt(Command);
      fOut.flush();
      boolean Ok = fIn.readBoolean();
      if (!Ok) {
        throw (new IOException());
      }

      setSimulatorButtonStates();
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,ErrorMessage,MessageBox.kError);
      setDefaultControls();
    }
  }

  private void shutdownServer()
  {
    try {
      fOut.writeInt(RemoteControlProtocol.kShutdownServer);
      fOut.flush();
      boolean Ok = fIn.readBoolean();
      if (!Ok) {
        throw (new IOException());
      }

      setDefaultControls();
    }
    catch (IOException exc) {
      MessageBox ErrorBox = new MessageBox(this,
                                fMessages.lookup("errorFailedToShutdownServer"),MessageBox.kError);
      setDefaultControls();
    }
  }

  private void setSimulatorButtonStates() throws IOException
  {
    fOut.writeInt(RemoteControlProtocol.kAcquireState);
    fOut.flush();
    boolean fSimulatorReady = fIn.readBoolean();
    boolean fSimulatorRunning = fIn.readBoolean();

    if (!fSimulatorReady) {
      fControlButtons[kStartContinueSimulatorButton].setEnabled(false);
      fControlButtons[kStopSimulatorButton].setEnabled(false);
      fControlButtons[kResetSimulatorButton].setEnabled(false);
    }
    else {
      if (!fSimulatorRunning) {
        fControlButtons[kStartContinueSimulatorButton].setEnabled(true);
        fControlButtons[kStopSimulatorButton].setEnabled(false);
      }
      else {
        fControlButtons[kStartContinueSimulatorButton].setEnabled(false);
        fControlButtons[kStopSimulatorButton].setEnabled(true);
      }
      fControlButtons[kResetSimulatorButton].setEnabled(true);
    }
  }
}
