// ----------------------------------------
// Filename      : ServerSocketHandler.java
// Created       : 16/02/2000
// Last modified : 02/05/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------------

package mitrasim.server;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import mitrasim.resources.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class ServerSocketHandler implements Runnable
{
  // internal datastructures
  private Resources        fResources;
  private Messages         fMessages;
  private Socket           fSocket;
  private InetAddress      fIP;
  private DataInputStream  fIn;
  private DataOutputStream fOut;
  private Protocol         fProtocol;
  private boolean          fProtocolError;
  private static boolean   fRemoteControlled;

  static {
    fRemoteControlled = false;
  }

  public ServerSocketHandler(Socket socket)
  {
    fResources = new Resources();
    fMessages = fResources.fMessages;
    fSocket = socket;
    fIP = fSocket.getInetAddress();      
    fProtocolError = true;

    try {
      fIn = new DataInputStream(fSocket.getInputStream());
      fOut = new DataOutputStream(fSocket.getOutputStream());
    }
    catch (IOException exc) {
      showMessage(fMessages.lookup("errorServerFailedToRetrieveDataStreams"));
      return;
    }

    try {
      // send the identification-string to the client
      showMessage(fMessages.lookup("messageSendingIDString"));
      fOut.writeUTF(fMessages.lookup("messageServerIDString") + ", " +
                                     DateFormat.getDateTimeInstance(DateFormat.LONG,DateFormat.MEDIUM).format(new Date()));
      fOut.flush();

      // get the requested protocol from the client
      showMessage(fMessages.lookup("messageRequestingClientProtocol"));
      int RequestedProtocol = fIn.readInt();
      if (RequestedProtocol == Protocol.kRemoteControlProtocol) {
        // only allow 1 remote control connected to the server
        if (!fRemoteControlled) {
          fProtocol = new RemoteControlProtocol(fIP,fIn,fOut);
          fRemoteControlled = true;
        }
        else {
          showMessage(fMessages.lookup("errorAlreadyRemoteControlled"));
          // send protocol-failure to the client
          fOut.writeInt(Protocol.kProtocolAlreadyInUse);
          fOut.flush();
          fSocket.close();
          return;
        }
      }
      else if (RequestedProtocol == Protocol.kVisualizerProtocol) {
        fProtocol = new VisualizerProtocol(fIP,fIn,fOut);
      }
      else {
        showMessage(fMessages.lookup("errorInvalidProtocolRequested",
                    String.valueOf(RequestedProtocol)));
        // send protocol-failure to the client
        fOut.writeInt(Protocol.kInvalidProtocolRequested);
        fOut.flush();
        fSocket.close();
        return;
      }

      showMessage(fMessages.lookup("textUsedProtocol") + " : " +
                  fProtocol.translate(fMessages) + ".");

      // send protocol-confirmation to the client
      fProtocolError = false;
      fOut.writeInt(Protocol.kProtocolOk);
      fOut.flush();
    }
    catch (IOException exc) {
      showMessage(fMessages.lookup("errorCommunicationFailure"));
      return;
    }
  }

  public boolean isValidProtocol()
  {
    return (!fProtocolError);
  }

  public int getProtocolCode()
  {
    return fProtocol.getProtocol();
  }

  public String getProtocolAbbreviation()
  {
    return fProtocol.getProtocolAbbreviation();
  }

  synchronized public void run()
  {
    if (isValidProtocol()) {
      while (fProtocol.isOperational()) {
        fProtocol.communicateWithClient();
      }
    }

    // when the remote-control protocol has finished, release it
    if (fProtocol.getProtocol() == Protocol.kRemoteControlProtocol) {
      // allow another remote control to connect
      fRemoteControlled = false;
    }
  }

  private void showMessage(String message)
  {
    System.out.println("[" + String.valueOf(fIP) + "] " + message);
  }
}
