// ----------------------------------
// Filename      : Driver.java
// Created       : 17/05/2000
// Last modified : 04/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.simulator;

import mitrasim.resources.*;
import mitrasim.simulator.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Driver
{
  // the driver's characteristics
  public int     fReactionTime;
  public int     fReactionTimeLeft;
  public float   fAggression;
  public boolean fMiddleLaneDriver;
  public float   fExcessForSpeedLimits;
  public float   fDesiredSpeed;
  public float   fSpeedInvariance;
  public float   fMinSpaceHeadway;
  public float   fMinTimeHeadway;
  public float   fMaxTimeHeadway;
  public float   fMinLeadGap;
  public float   fMinLagGap;

  // internal datastructure
  private Resources fResources;

  public Driver()
  {
  }

  public void setCharacteristics()
  {
    fResources = new Resources();

    // calculate the reaction time
    float ReactionTime = 0.0f;
    do {
      ReactionTime = (float) Math.exp((fResources.fsimReactionTimeStdDev *
                                       fResources.fsimRNG.nextGaussian()) +
                                       fResources.fsimReactionTimeMean);
    } while ((ReactionTime < fResources.fsimReactionTimeMin) ||
             (ReactionTime > fResources.fsimReactionTimeMax));

    // convert the reaction time (expressed in seconds) to tenths of seconds (= clockticks)
    fReactionTime = (int) Math.round(ReactionTime * 10.0f);
    fReactionTimeLeft = fReactionTime;

    // calculate the driver's aggression
    fAggression = fResources.fsimRNG.nextFloat();

    // determine if the driver is a middle-lane driver
    if ((fResources.fsimRNG.nextFloat() * 100.0f) < fResources.fsimPercentageMiddleLaneDrivers) {
      fMiddleLaneDriver = true;
    }
    else {
      fMiddleLaneDriver = false;
    }

    // determine the excess with which the driver crosses the speedlimits
    int Percentile = 0;
    if (fResources.fsimAggressionAndExcessCoupled) {
      Percentile = (int) Math.round(fAggression * 100.0f);
    }
    else {
      Percentile = (int) Math.round(fResources.fsimRNG.nextFloat() * 100.0f);
    }

    if (Percentile < 5) {
      fExcessForSpeedLimits = -16.8f;
    }
    else if (Percentile < 15) {
      fExcessForSpeedLimits = -6.2f;
    }
    else if (Percentile < 25) {
      fExcessForSpeedLimits = 0.7f;
    }
    else if (Percentile < 35) {
      fExcessForSpeedLimits = 3.5f;
    }
    else if (Percentile < 45) {
      fExcessForSpeedLimits = 7.1f;
    }
    else if (Percentile < 55) {
      fExcessForSpeedLimits = 10.1f;
    }
    else if (Percentile < 65) {
      fExcessForSpeedLimits = 14.6f;
    }
    else if (Percentile < 75) {
      fExcessForSpeedLimits = 19.2f;
    }
    else if (Percentile < 85) {
      fExcessForSpeedLimits = 24.2f;
    }
    else if (Percentile < 95) {
      fExcessForSpeedLimits = 32.2f;
    }
    else {
      fExcessForSpeedLimits = 40.0f;
    }

    // note that the desired speed depends on the link the driver is currently driving on
    fDesiredSpeed = 0.0f;

    // calculate the speed-invariance
    fSpeedInvariance = (1.0f - fAggression) * 20.0f;

    // calculate the car-following parameters
    fMinSpaceHeadway = fResources.fsimCarFollowingMinSpaceHeadway +
                       ((1.0f - fAggression) * fResources.fsimCarFollowingMinSpaceHeadwayExcess);
    fMinTimeHeadway = fResources.fsimCarFollowingMinTimeHeadway +
                      ((1.0f - fAggression) * fResources.fsimCarFollowingMinTimeHeadwayExcess);
    fMaxTimeHeadway = fResources.fsimCarFollowingMaxTimeHeadway +
                      ((1.0f - fAggression) * fResources.fsimCarFollowingMaxTimeHeadwayExcess);

    // calculate the lane-changing parameters
    fMinLeadGap = fResources.fsimLaneChangingMinLeadGap +
                  ((1.0f - fAggression) * fResources.fsimLaneChangingMinLeadGapExcess);
    fMinLagGap = fResources.fsimLaneChangingMinLagGap +
                 ((1.0f - fAggression) * fResources.fsimLaneChangingMinLagGapExcess);
  }

  public void cloneTo(Driver duplicate)
  {
    duplicate.fReactionTime = fReactionTime;
    duplicate.fReactionTimeLeft = fReactionTimeLeft;
    duplicate.fAggression = fAggression;
    duplicate.fMiddleLaneDriver = fMiddleLaneDriver;
    duplicate.fExcessForSpeedLimits = fExcessForSpeedLimits;
    duplicate.fDesiredSpeed = fDesiredSpeed;
    duplicate.fSpeedInvariance = fSpeedInvariance;
    duplicate.fMinSpaceHeadway = fMinSpaceHeadway;
    duplicate.fMinTimeHeadway = fMinTimeHeadway;
    duplicate.fMaxTimeHeadway = fMaxTimeHeadway;
    duplicate.fMinLeadGap = fMinLeadGap;
    duplicate.fMinLagGap = fMinLagGap;
  }

  public void setDesiredSpeed(int SpeedLimit, float MaximumSpeedPossible)
  {
    // calculate the desired speed from the speedlimit and the driver's excess for speedlimits
    fDesiredSpeed = SpeedLimit + fExcessForSpeedLimits;

    // truncate the desired speed according to the vehicle class
    // and grade of the current segment
    if (fDesiredSpeed > MaximumSpeedPossible) {
      fDesiredSpeed = MaximumSpeedPossible;
    }
  }
}
