// ----------------------------------
// Filename      : Simulator.java
// Created       : 26/10/1999
// Last modified : 25/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.simulator;

import java.io.*;
import java.util.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.simulator.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Simulator implements Runnable
{
  // internal datastructures
  private static boolean          fSimulatorReady;
  private static boolean          fSimulatorRunning;
  private static Resources        fResources;
  private static Messages         fMessages;
  private static Network          fNetwork;
  private static Vector           fLinkProcessingOrder;
  private static MasterClock      fMasterClock;
  private static Vector           fVehicleGenerators;
  private static Link             fStartLink;

  // only allow one instance of this class
  private static boolean fInstanceCreated;

  static {
    fInstanceCreated = false;
  }

  public Simulator()
  {
    if (fInstanceCreated == false) {
      fSimulatorReady = false;
      fSimulatorRunning = false;
      fResources = new Resources();
      fMessages = fResources.fMessages;
      fResources.fNetwork = new Network(fMessages);
      fNetwork = fResources.fNetwork;
      fLinkProcessingOrder = null;
      fMasterClock = new MasterClock();
      fVehicleGenerators = null;

      fInstanceCreated = true;
    }
  }

  public void prepareForScenarioUpload()
  {
    fSimulatorReady = false;
    fSimulatorRunning = false;
  }

  public void readNetworkFromStream(DataInputStream in)
  {
    fNetwork.readFromStream(in,Network.kShowTickMarks);
  }

  public void writeNetworkToStream(DataOutputStream out)
  {
    fNetwork.writeToStream(out);
  }

  public void writeTimeStampToStream(DataOutputStream out)
  {
    fNetwork.writeTimeStampToStream(out);
  }

  public void setTimeStamp()
  {
    fNetwork.fTimeStamp = System.currentTimeMillis();
  }

  public boolean checkNetwork()
  {
    Report ResultReport = fResources.fNetwork.analyse();
    fSimulatorReady = ResultReport.isEmpty();
    return isSimulatorReady();
  }

  public void calcNetworkGeometry()
  {
    fNetwork.calcGeometry();
  }

  public void topologicallySortNetwork()
  {
    fLinkProcessingOrder = new Vector();

    // initialize all the links before sorting
    Enumeration LinkEnum = fNetwork.fLinks.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) LinkEnum.nextElement();
      link.fTopologicallySorted = false;
    }

    // process all links in the network
    int fLinksProcessed = 0;
    while (fLinksProcessed < fNetwork.fLinks.size()) {

      // find a link which is not processed yet
      // and currently resides at the edge of the network
      boolean LinkFound = false;
      LinkEnum = fNetwork.fLinks.elements();
      while ((!LinkFound) && (LinkEnum.hasMoreElements())) {
        Link link = (Link) LinkEnum.nextElement();

        // skip the link if it is already sorted
        if (!link.fTopologicallySorted) {

          // a terminator automatically resides at the edge of the network
          if (link.fIsTerminator) {
            LinkFound = true;
          }
          else {
            // check if the link is an offramp
            if ((link.fType == Link.kSeparatedOffRamp) ||
                (link.fType == Link.kMergedOffRamp)) {
              // check if the link currently resides at the edge of the network
              if ((link.fMainLinksNextLink.fTopologicallySorted) &&
                  (link.fOffRampsNextLink.fTopologicallySorted)) {
                LinkFound = true;
              }
            }
            else { // the link is not an offramp
              // check if the link currently resides at the edge of the network
              if (link.fMainLinksNextLink.fTopologicallySorted) {
                LinkFound = true;
              }
            }
          }
        }

        // if the link is found, mark it and add it to the sorted list
        if (LinkFound) {
          link.fTopologicallySorted = true;
          fLinkProcessingOrder.addElement(link);
          ++fLinksProcessed;
        }
      }

      // are we dealing with a cycle in the graph ?
      if (!LinkFound) {

        // find an arbitrary link (that hasn't been processed yet) and cut it
        LinkFound = false;
        LinkEnum = fNetwork.fLinks.elements();
        while ((!LinkFound) && (LinkEnum.hasMoreElements())) {
          Link link = (Link) LinkEnum.nextElement();

          // skip the link if it is already sorted
          if (!link.fTopologicallySorted) {

            // if the link is found, mark it and add it to the sorted list
            LinkFound = true;
            link.fTopologicallySorted = true;
            fLinkProcessingOrder.addElement(link);
            ++fLinksProcessed;
          }
        }
      }
    }
  }

  public int getMasterClock()
  {
    return fMasterClock.get();
  }

  public String getMasterClockString()
  {
    return fMasterClock.translate(fMasterClock.get());
  }

  public void createVehicleGenerators()
  {
    fVehicleGenerators = new Vector();

    // there is a vehicle generator for every entry in the routing table
    Enumeration RoutingTableEnum = fResources.fsimRoutingTable.elements();
    while (RoutingTableEnum.hasMoreElements()) {
      RoutingTableEntry routingTableEntry = (RoutingTableEntry) RoutingTableEnum.nextElement();

      VehicleGenerator vehicleGenerator = new VehicleGenerator();

      vehicleGenerator.setRoutingTableEntry(routingTableEntry);

      // add it to the list of vehicle generators
      fVehicleGenerators.addElement(vehicleGenerator);
    }
  }

  public boolean isSimulatorReady()
  {
    return fSimulatorReady;
  }

  public boolean isSimulatorRunning()
  {
    return fSimulatorRunning;
  }

  public boolean startContinueSimulator()
  {
    if (fSimulatorReady && !fSimulatorRunning) {
      fSimulatorRunning = true;
      return true;
    }
    else {
      return false;
    }
  }

  public boolean stopSimulator()
  {
    if (fSimulatorReady && fSimulatorRunning) {
      fSimulatorRunning = false;
      return true;
    }
    else {
      return false;
    }
  }

  public boolean resetSimulator()
  {
    if (fSimulatorReady) {
      fSimulatorRunning = false;

      // reset the masterclock
      fMasterClock.set(fResources.fsimStartTime);

      // reset the random-generator
      fResources.fsimRNG = new Random();

      // purge the network from all vehicles
      Enumeration LinkEnum = fNetwork.fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        link.fVehicleQueue = new Vector();
      }

      // reset the vehicle generators
      Enumeration VehicleGeneratorEnum = fVehicleGenerators.elements();
      while (VehicleGeneratorEnum.hasMoreElements()) {
        VehicleGenerator vehicleGenerator = (VehicleGenerator) VehicleGeneratorEnum.nextElement();

        vehicleGenerator.reset();
      }

      // reset all loopdetectors
      Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
      while (LoopDetectorEnum.hasMoreElements()) {
        LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

        loopDetector.reset();
      }

      return true;
    }
    else {
      return false;
    }
  }

  synchronized public void run()
  {
    while (true) {
      if (fSimulatorReady && fSimulatorRunning) {

        generateVehicles();
        invokeModels();
        moveVehicles();
        updateLoopDetectors();
        fMasterClock.advance();

        // if necessary, wait some time
        if (fResources.fsimCycleHoldTime != 0) {
          try {
            Thread.sleep(fResources.fsimCycleHoldTime);
          }
          catch (InterruptedException exc) {
          }
        }
      }
    }
  }

  synchronized private void generateVehicles()
  {
    // generate vehicles at network entrances
    Enumeration VehicleGeneratorEnum = fVehicleGenerators.elements();
    while (VehicleGeneratorEnum.hasMoreElements()) {
      VehicleGenerator vehicleGenerator = (VehicleGenerator) VehicleGeneratorEnum.nextElement();

      vehicleGenerator.generate(fMasterClock.get());
    }
  }

  synchronized private void invokeModels()
  {
    // process all the topologically sorted links
    Enumeration LinkEnum = fLinkProcessingOrder.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) LinkEnum.nextElement();

      // process all the vehicles on the link
      Enumeration VehicleEnum = link.fVehicleQueue.elements();
      while (VehicleEnum.hasMoreElements()) {
        Vehicle vehicle = (Vehicle) VehicleEnum.nextElement();

        // set the driver's desired speed based on the speedlimit
        // and the vehicle's maximum speed possible (based on the grade of the current segment)
        vehicle.fDriver.setDesiredSpeed(vehicle.fCurrentLink.fSpeedLimit,
                                        vehicle.getMaximumSpeed(
                                          vehicle.fCurrentLink.fGrades[vehicle.fCurrentSegment]));

        invokeLaneChangeModel(vehicle);
        invokeCarFollowingModel(vehicle);

      } // vehicle enumeration
    } // link enumeration
  }

  synchronized private void invokeLaneChangeModel(Vehicle vehicle)
  {
    // decrease the time left when the driver is executing a DLC
    // to the left (and thus is overtaking)
    if (vehicle.fTimeLeftDuringOvertaking > 0) {
      --vehicle.fTimeLeftDuringOvertaking;
    }

    boolean DriverWantsToExecuteMLC = false;
    boolean DriverWantsToExecuteDLC = false;

    // has the driver's reaction-time elapsed ?
    if (vehicle.fDriver.fReactionTimeLeft <= 0) {

      vehicle.fTargetLane = vehicle.fCurrentLane;
      int NrOfMLCsToExecute = 0;

      // check if the driver needs to execute a mandatory lane-change (MLC)
      if (!vehicle.fMandatoryLocationReached) {

        // determine the targetlane
        if (vehicle.fMandatoryLocationOnLink == Vehicle.kMainLink) {

          // the vehicle is currently driving on an onramp
          if (vehicle.fCurrentSegment == (vehicle.fCurrentLink.fNrOfNodes - 2)) {
            // the vehicle is in the last segment (where the onramp starts merging
            // with the mainlink) => failsafe
            vehicle.fMandatoryLocationReached = true;
          }
          else {
            // where is the vehicle to be moved to ?
            if (vehicle.fCurrentLink.fLocation == Link.kLeft) {
              --vehicle.fTargetLane;
              NrOfMLCsToExecute = vehicle.fCurrentLane - vehicle.fCurrentLink.fNrOfLanes + 1;
            }
            else { // location of the onramp is right
              ++vehicle.fTargetLane;
              NrOfMLCsToExecute = vehicle.fCurrentLink.fNrOfLanesOnRamp - vehicle.fCurrentLane;
            }
            DriverWantsToExecuteMLC = true;
          }
        }
        else { // the vehicle's mandatory location is the offramp
           // the vehicle is currently driving on the mainlink of an offramp
          if (vehicle.fCurrentSegment == 0) {
            // the vehicle is in the first segment (where the offramp starts diverging
            // from the mainlink), it must wait until it is in the next segment
          }
          else {
            // where is the vehicle to be moved to ?
            if (vehicle.fCurrentLink.fLocation == Link.kLeft) {
              ++vehicle.fTargetLane;
              NrOfMLCsToExecute = vehicle.fCurrentLink.fNrOfLanes - vehicle.fCurrentLane;
            }
            else { // location of the onramp is right
              --vehicle.fTargetLane;
              NrOfMLCsToExecute = vehicle.fCurrentLane - vehicle.fCurrentLink.fNrOfLanesOnRamp + 1;
            }
            DriverWantsToExecuteMLC = true;
          }
        }

        // calculate the probability with which the driver executes a MLC
        if (DriverWantsToExecuteMLC) {

          DriverWantsToExecuteMLC = false;

          // determine the treshold at which a driver MUST execute a MLC
          float Treshold = vehicle.fDriver.fAggression *
                           fResources.fsimCriticalRampDistance *
                           vehicle.fCurrentLink.fRampLength;

          // modify this treshold by considering the number of MLCs that has to be
          // executed in order to reach the vehicle's mandatory location on the link
          if (NrOfMLCsToExecute == 2) {
            Treshold /= 2.0f;
          }
          else if (NrOfMLCsToExecute > 2) {
            Treshold /= 4.0f;
          }

          float LowerTreshold = (vehicle.fCurrentLink.fRampLength / 16.0f);
          if (LowerTreshold >= Treshold) {
            LowerTreshold = (Treshold / 2.0f);
          }

          if (vehicle.fDistanceTraveledOnRamp >= Treshold) {
            DriverWantsToExecuteMLC = true;
          }
          else {
            // don't create bottlenecks at the very start of an on-/offramp
            if (vehicle.fDistanceTraveledOnRamp >= LowerTreshold) {
              float ProbabilityToExecuteMLC = (float) Math.pow(vehicle.fDistanceTraveledOnRamp / Treshold,
                                                               fResources.fsimPowerOfMLCProbability);
              DriverWantsToExecuteMLC = (fResources.fsimRNG.nextFloat() <= ProbabilityToExecuteMLC);
            }
          }
        }      
      } // mandatory location not reached

      // check if the driver wants to execute a discretionary lane-change (DLC)
      // this is only possible if the driver isn't executing a mandatory lane-change
      if (!DriverWantsToExecuteMLC) {

        // has the driver a reason to change to the adjacent left-lane ?
        if ((!vehicle.fDesiredSpeedReached) &&
            (!vehicle.fFreeFlowing) &&
            (vehicle.fMandatoryLocationReached)) {

          boolean DLCPossible = false;

          // check if there is a valid adjacent left lane and if possible, set the target lane
          if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
              (vehicle.fCurrentLink.fType == Link.kHighway)) {

            if (vehicle.fCurrentLane < (vehicle.fCurrentLink.fNrOfLanes - 1)) {
              vehicle.fTargetLane = vehicle.fCurrentLane + 1;
              DLCPossible = true;
            }
          }
          else if ((vehicle.fCurrentLink.fType == Link.kMergedOnRamp) &&
                   (vehicle.fCurrentSegment < (vehicle.fCurrentLink.fNrOfNodes - 2))) {

            if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                (vehicle.fCurrentLane < (vehicle.fCurrentLink.fNrOfLanes - 1))) {
              vehicle.fTargetLane = vehicle.fCurrentLane + 1;
              DLCPossible = true;
            }
            else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                     (vehicle.fCurrentLane < (vehicle.fCurrentLink.fNrOfLanesOnRamp +
                                              vehicle.fCurrentLink.fNrOfLanes - 1))) {
              vehicle.fTargetLane = vehicle.fCurrentLane + 1;
              DLCPossible = true;
            }
          }
          else if ((vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) &&
                   (vehicle.fCurrentSegment > 0)) {

            if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                (vehicle.fCurrentLane != (vehicle.fCurrentLink.fNrOfLanes - 1)) &&
                (vehicle.fCurrentLane != (vehicle.fCurrentLink.fNrOfLanes +
                                          vehicle.fCurrentLink.fNrOfLanesOnRamp - 1))) {
              vehicle.fTargetLane = vehicle.fCurrentLane + 1;
              DLCPossible = true;
            }
            else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                     (vehicle.fCurrentLane != (vehicle.fCurrentLink.fNrOfLanesOnRamp - 1)) &&
                     (vehicle.fCurrentLane != (vehicle.fCurrentLink.fNrOfLanesOnRamp +
                                               vehicle.fCurrentLink.fNrOfLanes - 1))) {
              vehicle.fTargetLane = vehicle.fCurrentLane + 1;
              DLCPossible = true;
            }
          }

          if (DLCPossible) {

            // determine the lead-vehicle
            vehicle.fLeadVehicle = findVehicleInFront(vehicle,vehicle.fTargetLane);

            // check if the speed in the adjacent left lane is sufficiently higher
            // than the vehicle's current speed
            if (vehicle.fLeadVehicle == null) {
              DriverWantsToExecuteDLC = true;
            }
            else {
              DriverWantsToExecuteDLC = (vehicle.fLeadVehicle.fSpeed >=
                                         (vehicle.fSpeed + vehicle.fDriver.fSpeedInvariance));
            }
          }

          if (DriverWantsToExecuteDLC) {

            // check if the lane-change is allowed by examining the lane-marks on the link
            boolean LaneChangeOk = true;
            try {
              LaneChangeOk = (vehicle.fCurrentLink.fLaneChangePermissions[2 * vehicle.fCurrentLane] == Link.kAllowed);
            }
            catch (ArrayIndexOutOfBoundsException exc) {
              // failsafe
              LaneChangeOk = true;
            }

            if (!LaneChangeOk) {
              DriverWantsToExecuteDLC = false;
            }
            else { // the lane-change is allowed

              // check if trucks are allowed to make the lane-change
              if (vehicle.fType == Vehicle.kTruck) {

                int TargetLane = vehicle.fTargetLane;

                if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
                    (vehicle.fCurrentLink.fType == Link.kHighway)) {
                  DriverWantsToExecuteDLC = (TargetLane < fResources.fsimAllowTrucksOnLanes);
                  vehicle.fBusyOvertaking = DriverWantsToExecuteDLC;
                }
                else if (vehicle.fCurrentLink.fType == Link.kMergedOnRamp) {
                  // only consider the restrictions when the vehicle is driving on the mainlink
                  if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                      (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanes)) {
                    DriverWantsToExecuteDLC = (TargetLane < fResources.fsimAllowTrucksOnLanes);
                    vehicle.fBusyOvertaking = DriverWantsToExecuteDLC;
                  }
                  else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                           (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
                    TargetLane -= vehicle.fCurrentLink.fNrOfLanesOnRamp;
                    DriverWantsToExecuteDLC = (TargetLane < fResources.fsimAllowTrucksOnLanes);
                    vehicle.fBusyOvertaking = DriverWantsToExecuteDLC;
                  }
                }
                else if (vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) {
                  // only consider the restrictions when the vehicle is driving on the mainlink
                  if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                      (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanes)) {
                    DriverWantsToExecuteDLC = (TargetLane < fResources.fsimAllowTrucksOnLanes);
                    vehicle.fBusyOvertaking = DriverWantsToExecuteDLC;
                  }
                  else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                           (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
                    TargetLane -= vehicle.fCurrentLink.fNrOfLanesOnRamp;
                    DriverWantsToExecuteDLC = (TargetLane < fResources.fsimAllowTrucksOnLanes);
                    vehicle.fBusyOvertaking = DriverWantsToExecuteDLC;
                  }
                }
              }
              else {
                vehicle.fBusyOvertaking = true;
              }
            }
          } // driver currently wants to execute DLC
        } // desired speed not reached and vehicle not free-flowing
        else { // the driver might consider moving to the adjacent right lane

          boolean DLCPossible = false;

          // check if there is a valid adjacent right lane and if possible, set the target lane
          if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
              (vehicle.fCurrentLink.fType == Link.kHighway)) {

            if (vehicle.fCurrentLane > 0) {
              vehicle.fTargetLane = vehicle.fCurrentLane - 1;
              DLCPossible = true;
            }
          }
          else if ((vehicle.fCurrentLink.fType == Link.kMergedOnRamp) &&
                   (vehicle.fCurrentSegment < (vehicle.fCurrentLink.fNrOfNodes - 2))) {

            if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                (vehicle.fCurrentLane > 0)) {
              vehicle.fTargetLane = vehicle.fCurrentLane - 1;
              DLCPossible = true;
            }
            else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                     (vehicle.fCurrentLane > vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
              vehicle.fTargetLane = vehicle.fCurrentLane - 1;
              DLCPossible = true;
            }
          }
          else if ((vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) &&
                   (vehicle.fCurrentSegment > 0)) {

            if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                (vehicle.fCurrentLane != 0) &&
                (vehicle.fCurrentLane != vehicle.fCurrentLink.fNrOfLanes)) {
              vehicle.fTargetLane = vehicle.fCurrentLane - 1;
              DLCPossible = true;
            }
            else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                     (vehicle.fCurrentLane != 0) &&
                     (vehicle.fCurrentLane != vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
              vehicle.fTargetLane = vehicle.fCurrentLane - 1;
              DLCPossible = true;
            }
          }

          // check if the driver isn't currently overtaking another vehicle
          if (vehicle.fTimeLeftDuringOvertaking > 0) {
            DLCPossible = false;
          }

          if (DLCPossible) {

            // determine the lead-vehicle
            vehicle.fLeadVehicle = findVehicleInFront(vehicle,vehicle.fTargetLane);

            // check if the speed in the adjacent right lane is higher
            // than the vehicle's current speed
            if (vehicle.fLeadVehicle == null) {
              DriverWantsToExecuteDLC = true;
            }
            else {
              // subtract the minimum space-headway from the calculated space-headway
              // to allow for small gaps between the vehicles
              float SpaceHeadway = calcSpaceHeadwayInFront(vehicle,vehicle.fLeadVehicle) -
                                   fResources.fsimCarFollowingMinSpaceHeadway;
              float TimeHeadway = calcTimeHeadway(SpaceHeadway,vehicle.fSpeed);

              if (vehicle.fSpeed == 0.0f) {
                DriverWantsToExecuteDLC = false;
              }
              else {
                if (TimeHeadway < vehicle.fDriver.fMinTimeHeadway) {
                  DriverWantsToExecuteDLC = (vehicle.fLeadVehicle.fSpeed >=
                                             (vehicle.fSpeed - vehicle.fDriver.fSpeedInvariance));
                }
                else {
                  DriverWantsToExecuteDLC = true;
                }
              }
            }
          }

          if (DriverWantsToExecuteDLC) {

            // check if the lane-change is allowed by examining the lane-marks on the link
            boolean LaneChangeOk = true;
            try {
              LaneChangeOk = (vehicle.fCurrentLink.fLaneChangePermissions[(2 * vehicle.fTargetLane) + 1] == Link.kAllowed);
            }
            catch (ArrayIndexOutOfBoundsException exc) {
              // failsafe
              LaneChangeOk = true;
            }

            if (!LaneChangeOk) {
              DriverWantsToExecuteDLC = false;
            }
            else { // the lane-change is allowed

              // check if the driver isn't a middle-lane-driver
              if (vehicle.fDriver.fMiddleLaneDriver) {

                int TargetLane = vehicle.fTargetLane;

                if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
                    (vehicle.fCurrentLink.fType == Link.kHighway)) {
                  DriverWantsToExecuteDLC = (TargetLane > 0);
                }
                else if (vehicle.fCurrentLink.fType == Link.kMergedOnRamp) {
                  // only consider the restrictions when the vehicle is driving on the mainlink
                  if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                      (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanes)) {
                    DriverWantsToExecuteDLC = (TargetLane > 0);
                  }
                  else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                           (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
                    TargetLane -= vehicle.fCurrentLink.fNrOfLanesOnRamp;
                    DriverWantsToExecuteDLC = (TargetLane > 0);
                  }
                }
                else if (vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) {
                  // only consider the restrictions when the vehicle is driving on the mainlink
                  if ((vehicle.fCurrentLink.fLocation == Link.kLeft) &&
                      (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanes)) {
                    DriverWantsToExecuteDLC = (TargetLane > 0);
                  }
                  else if ((vehicle.fCurrentLink.fLocation == Link.kRight) &&
                           (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesOnRamp)) {
                    TargetLane -= vehicle.fCurrentLink.fNrOfLanesOnRamp;
                    DriverWantsToExecuteDLC = (TargetLane > 0);
                  }
                }
              }
            }
          } // driver currently wants to execute DLC
        } // the driver might consider moving to the adjacent right lane
      } // driver doesn't want to execute MLC
    } // reaction-time elapsed

    // does the driver need/want to change lanes ?
    if (vehicle.fBusyForcedMerging || DriverWantsToExecuteMLC || DriverWantsToExecuteDLC) {

      tryToExecuteLaneChange(vehicle,DriverWantsToExecuteMLC);
    }
  }

  synchronized private void tryToExecuteLaneChange(Vehicle vehicle, boolean DriverWantsToExecuteMLC)
  {
    vehicle.fLeftBlinkerOn = false;
    vehicle.fRightBlinkerOn = false;

    // only allow lane-changes when the vehicle is moving
    if (vehicle.fSpeed == 0.0f) {
      return;
    }

    // determine the lead-vehicle
    vehicle.fLeadVehicle = findVehicleInFront(vehicle,vehicle.fTargetLane);

    // subtract the minimum space-headway from the calculated space-headway
    // to allow for small gaps between the vehicles
    float LeadSpaceHeadway = calcSpaceHeadwayInFront(vehicle,vehicle.fLeadVehicle) -
                             fResources.fsimCarFollowingMinSpaceHeadway;

    // determine the lag-vehicle
    vehicle.fLagVehicle = findVehicleBehind(vehicle,vehicle.fTargetLane);

    // subtract the minimum space-headway from the calculated space-headway
    // to allow for small gaps between the vehicles
    float LagSpaceHeadway = calcSpaceHeadwayBehind(vehicle,vehicle.fLagVehicle) -
                            fResources.fsimCarFollowingMinSpaceHeadway;

    boolean LeadGapAcceptable = false;
    boolean LagGapAcceptable = false;

    if (vehicle.fLeadVehicle == null) {
      // the vehicle has a clear lead-headway
      LeadGapAcceptable = true;
    }
    else {
      if (LeadSpaceHeadway >= vehicle.fDriver.fMinLeadGap) {
        LeadGapAcceptable = true;
      }
    }

    if (vehicle.fLagVehicle == null) {
      // the vehicle has a clear lag-headway
      LagGapAcceptable = true;
    }
    else {
      if (LagSpaceHeadway >= vehicle.fDriver.fMinLagGap) {
        LagGapAcceptable = true;
      }
    }

    if (LeadGapAcceptable && LagGapAcceptable) {

      // execute the lane-change
      vehicle.fCurrentLane = vehicle.fTargetLane;

      if (vehicle.fBusyOvertaking) {
        vehicle.fTimeLeftDuringOvertaking = fResources.fsimMinTimeLeftDuringOvertaking;
        vehicle.fBusyOvertaking = false;
      }

      if (!vehicle.fMandatoryLocationReached) {

        // check if the vehicle has reached it's mandatory location after executing
        // the current lane-change (which is a mandatory lane-change)
        if (vehicle.fMandatoryLocationOnLink == Vehicle.kMainLink) {

          // the vehicle is or was currently driving on an onramp
          if (vehicle.fCurrentSegment < (vehicle.fCurrentLink.fNrOfNodes - 2)) {
            if (vehicle.fCurrentLink.fLocation == Link.kLeft) {

              if (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanes) {
                vehicle.fMandatoryLocationReached = true;
              }
            }
            else { // the location of the onramp is right

              if (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesOnRamp) {
                vehicle.fMandatoryLocationReached = true;
              }
            }
          }
          else {
            // the vehicle is driving on the last segment of the onramp
            vehicle.fMandatoryLocationReached = true;
          }
        }
        else { // the vehicle's mandatory location is the offramp

          // the vehicle is or was currently driving on the mainlink of an offramp
          if (vehicle.fCurrentSegment > 0) {
            if (vehicle.fCurrentLink.fLocation == Link.kLeft) {

              if (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanes) {
                vehicle.fMandatoryLocationReached = true;
              }
            }
            else { // the location of the offramp is right

              if (vehicle.fCurrentLane < vehicle.fCurrentLink.fNrOfLanesOnRamp) {
                vehicle.fMandatoryLocationReached = true;
              }
            }
          } // vehicle's currentsegment > 0
        } // vehicle's mandatory location is offramp
      } // vehicle hasn't reached mandatory location

      vehicle.fBusyForcedMerging = false;

      if (vehicle.fLagVehicle != null) {
        vehicle.fLagVehicle.fYielding = false;
        vehicle.fLagVehicle = null;
      }

    } // lead- and lag-gaps are acceptable
    else {

      // if the driver wants to execute a DLC and it is currently not possible, then
      // that's too bad so the driver has to try again at a later time

      // if the driver wants to execute a MLC and it is currently not possible, then
      // forced merging is necessary
      if (DriverWantsToExecuteMLC || vehicle.fBusyForcedMerging) {

        vehicle.fBusyForcedMerging = true;
        if (vehicle.fTargetLane > vehicle.fCurrentLane) {
          vehicle.fLeftBlinkerOn = true;
          vehicle.fRightBlinkerOn = false;
        }
        else {
          vehicle.fLeftBlinkerOn = false;
          vehicle.fRightBlinkerOn = true;
        }
        if (vehicle.fLagVehicle != null) {
          vehicle.fLagVehicle.fYielding = true;
          vehicle.fLagVehicle.fVehicleToYieldTo = vehicle;
        }
      }
    }
  }

  synchronized private void invokeCarFollowingModel(Vehicle vehicle)
  {
    // find the vehicle in front and calculate (if possible) the space- and timeheadway
    vehicle.fVehicleInFront = findVehicleInFront(vehicle,vehicle.fCurrentLane);
    // subtract the minimum space-headway from the calculated space-headway
    // to allow for small gaps between the vehicles
    vehicle.fSpaceHeadway = calcSpaceHeadwayInFront(vehicle,vehicle.fVehicleInFront) -
                            fResources.fsimCarFollowingMinSpaceHeadway;
    vehicle.fTimeHeadway = calcTimeHeadway(vehicle.fSpaceHeadway,vehicle.fSpeed);

    if (vehicle.fBusyForcedMerging) {

      if (vehicle.fLeadVehicle != null) {
        float LeadSpaceHeadway = calcSpaceHeadwayInFront(vehicle,vehicle.fLeadVehicle) -
                                 fResources.fsimCarFollowingMinSpaceHeadway;
        float LeadTimeHeadway = calcTimeHeadway(LeadSpaceHeadway,vehicle.fSpeed);

        // determine the most constraining vehicle
        if ((vehicle.fVehicleInFront == null) || (LeadSpaceHeadway < vehicle.fSpaceHeadway)) {
          vehicle.fVehicleInFront = vehicle.fLeadVehicle;
          vehicle.fSpaceHeadway = LeadSpaceHeadway;
          vehicle.fTimeHeadway = LeadTimeHeadway;
        }
      }
    }
    else if (vehicle.fYielding) {

      if (vehicle.fVehicleToYieldTo != null) {
        float LeadSpaceHeadway = calcSpaceHeadwayInFront(vehicle,vehicle.fVehicleToYieldTo) -
                                 fResources.fsimCarFollowingMinSpaceHeadway;
        float LeadTimeHeadway = calcTimeHeadway(LeadSpaceHeadway,vehicle.fSpeed);

        // determine the most constraining vehicle
        if ((vehicle.fVehicleInFront == null) || (LeadSpaceHeadway < vehicle.fSpaceHeadway)) {
          vehicle.fVehicleInFront = vehicle.fVehicleToYieldTo;
          vehicle.fSpaceHeadway = LeadSpaceHeadway;
          vehicle.fTimeHeadway = LeadTimeHeadway;
        }
      }
      
      vehicle.fYielding = false;
      vehicle.fVehicleToYieldTo = null;
    }

    // check the car-following regime every clocktick for more realistic behaviour
    // (this is needed due to the discrete nature of the simulator's time)
    if (vehicle.fSpeed != 0.0f) {
      if ((vehicle.fTimeHeadway > vehicle.fDriver.fMinTimeHeadway) &&
          (vehicle.fTimeHeadway < vehicle.fDriver.fMaxTimeHeadway)) {

        // car-following regime
        float SpeedDifference = vehicle.fVehicleInFront.fSpeed - vehicle.fSpeed;
        float Alpha = 0.0f;
        float Beta = 0.0f;
        float Gamma = 0.0f;

        if (vehicle.fSpeed < vehicle.fVehicleInFront.fSpeed) {
          Alpha = fResources.fsimCarFollowingAlphaPlus;
          Beta = fResources.fsimCarFollowingBetaPlus;
          Gamma = fResources.fsimCarFollowingGammaPlus;
        }
        else {
          Alpha = fResources.fsimCarFollowingAlphaMin;
          Beta = fResources.fsimCarFollowingBetaMin;
          Gamma = fResources.fsimCarFollowingGammaMin;
        }

        vehicle.fNewAcceleration = (float) (Alpha *
                                                 (Math.pow((vehicle.fSpeed / 3.6f),Beta) /
                                                  Math.pow(vehicle.fSpaceHeadway,Gamma)) *
                                                 SpeedDifference / 3.6f);

        vehicle.fFreeFlowing = false;

        // apply this acceleration immediately
        vehicle.fCurrentAcceleration = vehicle.fNewAcceleration;
      }
    }

    // has the driver's reaction-time elapsed ?
    if (vehicle.fDriver.fReactionTimeLeft <= 0) {

      // the driver's reaction time has elapsed so he applies the new acceleration
      vehicle.fCurrentAcceleration = vehicle.fNewAcceleration;

      handleFreeFlowingRegime(vehicle);

      vehicle.fDriver.fReactionTimeLeft = vehicle.fDriver.fReactionTime;
    }
    else {
      // the driver must wait some more
      --vehicle.fDriver.fReactionTimeLeft;
    }

    vehicle.fDesiredSpeedReached = false;
 
    if (!handleCarFollowingEmergencyRegime(vehicle)) {

      // has the driver's desired speed been reached ?
      // take into account the driver's speed-invariance
      if ((vehicle.fSpeed >= (vehicle.fDriver.fDesiredSpeed - vehicle.fDriver.fSpeedInvariance)) &&
          (vehicle.fSpeed <= (vehicle.fDriver.fDesiredSpeed + vehicle.fDriver.fSpeedInvariance))) {

        // cruise at this speed
        vehicle.fCurrentAcceleration = 0.0f;
        vehicle.fDesiredSpeedReached = true;
      }
      else {
        // convert and apply the current acceleration (expressed in m/s^2) to km/h
        vehicle.fSpeed += (vehicle.fCurrentAcceleration * 3.6f) / 10.0f;

        // crop a negative speed
        if (vehicle.fSpeed < 0.0f) {
          vehicle.fSpeed = 0.0f;
        }
      }
    }

    // prevent collisions
    if (vehicle.fVehicleInFront != null) {

      // calculate the distance to travel (expressed in m)
      // by converting the speed (expressed in km/h) to m/clocktick

      float DistanceToTravelInOneClocktick = (vehicle.fSpeed / 3.6f) / 10.0f;
      float VehicleInFrontDistanceToTravelInOneClocktick = (vehicle.fVehicleInFront.fSpeed / 3.6f) / 10.0f;

      // calculate the space-headway for the next clocktick to check for a collision
      if (((vehicle.fSpaceHeadway - DistanceToTravelInOneClocktick) +
           VehicleInFrontDistanceToTravelInOneClocktick) <= 0.0f) {

        vehicle.fSpeed = 0.0f;
        vehicle.fNewAcceleration = 0.0f;
        vehicle.fCurrentAcceleration = vehicle.fNewAcceleration;
      }
    } // prevent collisions
  }

  synchronized private void handleFreeFlowingRegime(Vehicle vehicle)
  {
    if (vehicle.fVehicleInFront == null) {

      // free-flow regime
      if (vehicle.fSpeed < (vehicle.fDriver.fDesiredSpeed - vehicle.fDriver.fSpeedInvariance)) {
        vehicle.fNewAcceleration = vehicle.getMaximumAcceleration(
                                     vehicle.fCurrentLink.fGrades[vehicle.fCurrentSegment]);
      }
      else if (vehicle.fSpeed > (vehicle.fDriver.fDesiredSpeed + vehicle.fDriver.fSpeedInvariance)) {
        vehicle.fNewAcceleration = vehicle.getNormalDeceleration();
      }
      else {
        vehicle.fNewAcceleration = 0.0f;
      }

      vehicle.fFreeFlowing = true;
    }
    else {

      if (vehicle.fSpeed != 0) {
        if (vehicle.fTimeHeadway > vehicle.fDriver.fMaxTimeHeadway) {

          // free-flow regime
          if (vehicle.fSpeed < (vehicle.fDriver.fDesiredSpeed - vehicle.fDriver.fSpeedInvariance)) {
            vehicle.fNewAcceleration = vehicle.getMaximumAcceleration(
                                         vehicle.fCurrentLink.fGrades[vehicle.fCurrentSegment]);
          }
          else if (vehicle.fSpeed > (vehicle.fDriver.fDesiredSpeed + vehicle.fDriver.fSpeedInvariance)) {
            vehicle.fNewAcceleration = vehicle.getNormalDeceleration();
          }
          else {
            vehicle.fNewAcceleration = 0.0f;
          }
        }
      }
      else {  // speed == 0.0f
        if (vehicle.fSpaceHeadway < vehicle.fDriver.fMinSpaceHeadway) {
          vehicle.fNewAcceleration = 0.0f;
        }
        else {
          vehicle.fNewAcceleration = vehicle.getMaximumAcceleration(
                                       vehicle.fCurrentLink.fGrades[vehicle.fCurrentSegment]);
        }
      }

      vehicle.fFreeFlowing = true;
    }
  }

  synchronized private boolean handleCarFollowingEmergencyRegime(Vehicle vehicle)
  {
    // check if the vehicle is in the emergency regime
    boolean EmergencyRegime = false;

    if (vehicle.fVehicleInFront != null) {

      if (vehicle.fSpeed != 0.0f) {

        if (vehicle.fTimeHeadway < vehicle.fDriver.fMinTimeHeadway) {
          EmergencyRegime = true;
        }
      }
      else // speed == 0.0f

      if (vehicle.fSpaceHeadway < vehicle.fDriver.fMinSpaceHeadway) {
        EmergencyRegime = true;
      }
    }

    if (EmergencyRegime) {

      // apply a necessary deceleration to safely extend the vehicle's headway
      if (vehicle.fSpeed > vehicle.fVehicleInFront.fSpeed) {
        vehicle.fNewAcceleration = vehicle.getMaximumDeceleration();
        float SpeedDifference = (vehicle.fSpeed - vehicle.fVehicleInFront.fSpeed) / 3.6f;
        float Deceleration = vehicle.fVehicleInFront.fCurrentAcceleration -
                             ((0.5f * SpeedDifference * SpeedDifference) /
                              vehicle.fSpaceHeadway);
        if (Deceleration < vehicle.fNewAcceleration) {
          vehicle.fNewAcceleration = Deceleration;
        }
      }
      else {
        vehicle.fNewAcceleration = vehicle.getMaximumDeceleration();
        float Deceleration = vehicle.fVehicleInFront.fCurrentAcceleration +
                             (0.25f * vehicle.getMaximumDeceleration());
        if (Deceleration < vehicle.fNewAcceleration) {
          vehicle.fNewAcceleration = Deceleration;
        }
      }

      vehicle.fCurrentAcceleration = vehicle.fNewAcceleration;

      // convert and apply the current acceleration (expressed in m/s^2) to km/h
      vehicle.fSpeed += (vehicle.fCurrentAcceleration * 3.6f) / 10.0f;

      // has the vehicle stopped ?
      if (vehicle.fSpeed <= 0.0f) {

        vehicle.fSpeed = 0.0f;

        // don't accelerate or decelerate anymore
        vehicle.fNewAcceleration = 0.0f;
        vehicle.fCurrentAcceleration = vehicle.fNewAcceleration;
      }

      vehicle.fFreeFlowing = false;

      return true;
    }
    else {
      return false;
    }
  }

  synchronized private void moveVehicles()
  {
    // process all the topologically sorted links
    Enumeration LinkEnum = fLinkProcessingOrder.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) LinkEnum.nextElement();

      // process all the vehicles on the link
      Enumeration VehicleEnum = link.fVehicleQueue.elements();
      while (VehicleEnum.hasMoreElements()) {
        Vehicle vehicle = (Vehicle) VehicleEnum.nextElement();

        Link OldLinkVehicleBelongsTo = null;

        try {
          // update the vehicle's position
          // calculate the distance to travel (expressed in m)
          // by converting the speed (expressed in km/h) to m/clocktick
          float DistanceToTravel = (vehicle.fSpeed / 3.6f) / 10.0f;
          vehicle.fDistanceTraveled += DistanceToTravel;

          // calculate the distance the vehicle could have traveled on a ramp (needed for MLC)
          if ((((vehicle.fCurrentLink.fType == Link.kSeparatedOnRamp) ||
                (vehicle.fCurrentLink.fType == Link.kMergedOnRamp)) &&
               (vehicle.fCurrentSegment < (vehicle.fCurrentLink.fNrOfNodes - 2))) ||
              (((vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) ||
                (vehicle.fCurrentLink.fType == Link.kMergedOffRamp)) &&
               (vehicle.fCurrentSegment > 0))) {

              vehicle.fDistanceTraveledOnRamp += DistanceToTravel;
          }

          // failsafe in case the vehicle is in an incorrect segment and/or lane
          if (vehicle.fCurrentSegment < 0) {
            vehicle.fCurrentSegment = 0;
          }
          else if (vehicle.fCurrentSegment >= (vehicle.fCurrentLink.fNrOfNodes - 1)) {
            vehicle.fCurrentSegment = vehicle.fCurrentLink.fNrOfNodes - 2;
          }

          if (vehicle.fCurrentLane < 0) {
            vehicle.fCurrentLane = 0;
          }
          else if (vehicle.fCurrentLane >= (vehicle.fCurrentLink.fNrOfLanesAtNode[vehicle.fCurrentSegment])) {
            vehicle.fCurrentSegment = vehicle.fCurrentLink.fNrOfLanesAtNode[vehicle.fCurrentSegment];
          }

          // handle the vehicle's routing

          // check if the vehicle can stay on the current segment
          if ((vehicle.fPositionOnCurrentSegment + DistanceToTravel) <
              vehicle.fCurrentLink.fLengths[vehicle.fCurrentSegment]) {

            vehicle.fPositionOnCurrentSegment += DistanceToTravel;
          }
          else // check whether the vehicle crosses a segment or a link
          if (vehicle.fCurrentSegment < (vehicle.fCurrentLink.fNrOfNodes - 2)) {

            // select the correct lane in the next segment that lies
            // after the current lane in this segment
            vehicle.fCurrentLane = calcLaneNrInNextSegment(vehicle.fCurrentLane,
                                                           vehicle.fCurrentSegment,
                                                           vehicle.fCurrentLink);

            // failsafe for when the vehicle resides in an onramp's merging area
            if ((vehicle.fCurrentLink.fType == Link.kMergedOnRamp) &&
                (vehicle.fCurrentLane == -1)) {

              // place the vehicle in the lane adjacent to the onramp
              if (vehicle.fCurrentLink.fLocation == Link.kLeft) {
                vehicle.fCurrentLane = vehicle.fCurrentLink.fNrOfLanes - 1;
              }
              else { // the location of the onramp is right
                vehicle.fCurrentLane = 0;
              }
            }

            // stay on the link but advance to the next segment
            vehicle.fPositionOnCurrentSegment = DistanceToTravel -
                                                (vehicle.fCurrentLink.fLengths[vehicle.fCurrentSegment] -
                                                 vehicle.fPositionOnCurrentSegment);
            ++vehicle.fCurrentSegment;
          }
          else { // the vehicles crosses the end of a link

            // check if the vehicle passes over a loopdetector
            Node NodeCrossed = vehicle.fCurrentLink.fNodes[link.fNrOfNodes - 1];
            LoopDetector loopDetector = (LoopDetector) fResources.fsimLoopDetectors.get(new Integer(NodeCrossed.fID));
            if (loopDetector != null) {
              // count the vehicle
              loopDetector.countVehicle(vehicle);
            }

            // check if the vehicle has reached the perimeter of the network
            if (vehicle.fCurrentLink.fIsTerminator) {

              // remove the vehicle from the link's vehicle queue
              vehicle.fCurrentLink.fVehicleQueue.removeElement(vehicle);
              vehicle = null;
            }
            else {

              boolean VehiclesTakesOffRamp = false;

              // temporarily save the current link
              OldLinkVehicleBelongsTo = vehicle.fCurrentLink;

              // find the link that follows after the current link and lies
              // on the vehicle's route

              // only offramps have two outgoing links so find out which
              // one of them lies on the vehicle's route
              if ((vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) ||
                  (vehicle.fCurrentLink.fType == Link.kMergedOffRamp)) {

                // find the next link on the vehicle's route
                int SecondNextJunctionNodeID = 0;

                if (vehicle.fIndexOfLastJunctionNodePassed == -1) {
                  SecondNextJunctionNodeID = (vehicle.fRTE.fRoutes[vehicle.fRouteNr][1]).fID;
                }
                else {
                  SecondNextJunctionNodeID = (vehicle.fRTE.fRoutes[vehicle.fRouteNr]
                                              [vehicle.fIndexOfLastJunctionNodePassed + 2]).fID;
                }

                // find out to which of the two outgoing links this junction-node belongs
                if (SecondNextJunctionNodeID == vehicle.fCurrentLink.fOffRampsNextLink.fNodes
                                                [vehicle.fCurrentLink.fOffRampsNextLink.fNrOfNodes - 1].fID) {
                  vehicle.fCurrentLink = vehicle.fCurrentLink.fOffRampsNextLink;
                  VehiclesTakesOffRamp = true;
                }
                else {
                  // built-in failsafe in case the vehicle is not any longer on it's
                  // route : divert it to the next link connected to this mainlink
                  vehicle.fCurrentLink = vehicle.fCurrentLink.fMainLinksNextLink;
                }
              }
              else { // all the other linktypes have only one outgoing link
                // built-in failsafe : default to this route if vehicle is off-route
                vehicle.fCurrentLink = vehicle.fCurrentLink.fMainLinksNextLink;
              }

              // indicate transfer to the next link
              ++vehicle.fIndexOfLastJunctionNodePassed;

              // check whether or not the vehicle is already in the correct part of the link

              if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
                  (vehicle.fCurrentLink.fType == Link.kHighway)) {

                vehicle.fMandatoryLocationOnLink = Vehicle.kMainLink;
                vehicle.fMandatoryLocationReached = true;
              }
              else if (vehicle.fCurrentLink.fType == Link.kMergedOnRamp) {

                vehicle.fMandatoryLocationOnLink = Vehicle.kMainLink;

                // where does the vehicle come from :
                // is it from a connection with this link's mainlink or onramp ?
                if (OldLinkVehicleBelongsTo.fID ==
                    vehicle.fCurrentLink.fMainLinksPreviousLink.fID) {
                  // the vehicle enters this link at the mainlink
                  vehicle.fMandatoryLocationReached = true;
                }
                else {
                  // the vehicle enters this link at the onramp
                  vehicle.fMandatoryLocationReached = false;
                }
              }
              else if (vehicle.fCurrentLink.fType == Link.kSeparatedOffRamp) {

                // to which outgoing link does the vehicle need to be routed ?

                // find the next link on the vehicle's route
                int SecondNextJunctionNodeID = 0;

                if (vehicle.fIndexOfLastJunctionNodePassed == -1) {
                  SecondNextJunctionNodeID = (vehicle.fRTE.fRoutes[vehicle.fRouteNr][1]).fID;
                }
                else {
                  SecondNextJunctionNodeID = (vehicle.fRTE.fRoutes[vehicle.fRouteNr]
                                              [vehicle.fIndexOfLastJunctionNodePassed + 2]).fID;
                }

                // find out to which of the two outgoing links this junction-node belongs
                if (SecondNextJunctionNodeID == vehicle.fCurrentLink.fOffRampsNextLink.fNodes
                                                [vehicle.fCurrentLink.fOffRampsNextLink.fNrOfNodes - 1].fID) {
                  // the vehicle is routed to the offramp's next link
                  vehicle.fMandatoryLocationOnLink = Vehicle.kOffRamp;
                  vehicle.fMandatoryLocationReached = false;
                }
                else {
                  // the vehicle is routed to the mainlink's next link
                  vehicle.fMandatoryLocationOnLink = Vehicle.kMainLink;
                  vehicle.fMandatoryLocationReached = true;
                }
              }

              vehicle.fDistanceTraveledOnRamp = 0.0f;
              vehicle.fBusyForcedMerging = false;

              if (vehicle.fLagVehicle != null) {
                vehicle.fLagVehicle.fYielding = false;
                vehicle.fLagVehicle = null;
              }

              // temporarily save the current segment
              int OldSegment = vehicle.fCurrentSegment;

              // place vehicle on the first segment of the next link
              vehicle.fCurrentSegment = 0;

              // place the vehicle on the correct position on the new link
              vehicle.fPositionOnCurrentSegment = DistanceToTravel -
                                                  (OldLinkVehicleBelongsTo.fLengths[OldSegment] -
                                                   vehicle.fPositionOnCurrentSegment);

              // failsafe for when the vehicle still resides on the mainlink
              // and has to take the offramp
              if ((OldLinkVehicleBelongsTo.fType == Link.kSeparatedOffRamp) &&
                  (VehiclesTakesOffRamp)) {

                if (OldLinkVehicleBelongsTo.fLocation == Link.kLeft) {

                  if (vehicle.fCurrentLane < OldLinkVehicleBelongsTo.fNrOfLanes) {
                    // place the vehicle in the right-most lane
                    // on the link connected to the offramp
                    vehicle.fCurrentLane = OldLinkVehicleBelongsTo.fNrOfLanes;
                  }
                }
                else { // the location of the offramp is right

                  if (vehicle.fCurrentLane >= OldLinkVehicleBelongsTo.fNrOfLanesOnRamp) {
                    // place the vehicle in the left-most lane
                    // on the link connected to the offramp
                    vehicle.fCurrentLane = OldLinkVehicleBelongsTo.fNrOfLanesOnRamp - 1;
                  }
                }
              }

              // place the vehicle in the correct new lane
              vehicle.fCurrentLane = calcLaneNrInNextLink(vehicle.fCurrentLane,
                                                          OldLinkVehicleBelongsTo,
                                                          vehicle.fCurrentLink);

              // failsafe in case the vehicle is located in an illegal lane
              if (vehicle.fCurrentLane >= vehicle.fCurrentLink.fNrOfLanesAtNode[0]) {
                vehicle.fCurrentLane = vehicle.fCurrentLink.fNrOfLanesAtNode[0];
              }
              else if (vehicle.fCurrentLane < 0) {
                vehicle.fCurrentLane = 0;
              }

              // remove the vehicle from the vehicle's previous link's vehicle queue
              OldLinkVehicleBelongsTo.fVehicleQueue.removeElement(vehicle);

              // place the vehicle on this link by
              // adding it at the end of the queue (so it will be processed LAST)
              vehicle.fCurrentLink.fVehicleQueue.addElement(vehicle);
            } // vehicle crosses a link and is not at the network's perimeter
          }
        }

        // just in case ...
        catch (ArrayIndexOutOfBoundsException exc) {
          Link LinkOfVehicle = vehicle.fCurrentLink;

          if (OldLinkVehicleBelongsTo != null) {
            OldLinkVehicleBelongsTo.fVehicleQueue.removeElement(vehicle);
          }

          if (LinkOfVehicle != null) {
            LinkOfVehicle.fVehicleQueue.removeElement(vehicle);
          }
        }
      }

      // sort the vehicles (most downstream ones are placed first in the queue)
      quicksort(link.fVehicleQueue,0,link.fVehicleQueue.size() - 1);
    }
  }

  private static void quicksort(Vector VehicleQueue, int from, int to)
  {
    if (to > from) {

      int pivot = from;
      for (int i = from + 1; (i <= to); ++i) {

        Vehicle Temp = new Vehicle();
        ((Vehicle) VehicleQueue.elementAt(i)).cloneTo(Temp);

        if (Temp.isGreaterThan((Vehicle) VehicleQueue.elementAt(pivot))) {

          // swap the two vehicles in the link's queue
          ((Vehicle) VehicleQueue.elementAt(pivot + 1)).cloneTo((Vehicle) VehicleQueue.elementAt(i));
          ((Vehicle) VehicleQueue.elementAt(pivot)).cloneTo((Vehicle) VehicleQueue.elementAt(pivot + 1));
          Temp.cloneTo((Vehicle) VehicleQueue.elementAt(pivot));

          ++pivot;
        }
      }

      quicksort(VehicleQueue,from,pivot - 1);
      quicksort(VehicleQueue,pivot + 1,to);
    }
  }

  synchronized private void updateLoopDetectors()
  {
    // process all loopdetectors
    Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
    while (LoopDetectorEnum.hasMoreElements()) {
      LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

      loopDetector.update();
    }
  }

  synchronized private Vehicle findVehicleInFront(Vehicle vehicle, int LaneToSearchIn)
  {
    Vehicle VehicleInFront = null;

    int LaneNr = LaneToSearchIn;
    int CurrentSegment = vehicle.fCurrentSegment;

    // search in the current link
    int IndexOfSubjectVehicle = vehicle.fCurrentLink.fVehicleQueue.indexOf(vehicle);
    int IndexOfVehicleInFront = (IndexOfSubjectVehicle - 1);
    boolean Found = false;
    while ((!Found) && (IndexOfVehicleInFront >= 0)) {

      VehicleInFront = (Vehicle) vehicle.fCurrentLink.fVehicleQueue.elementAt(IndexOfVehicleInFront);

      // are we looking in another segment ?
      if (VehicleInFront.fCurrentSegment != CurrentSegment) {

        for (int SegmentNr = CurrentSegment;
             SegmentNr < VehicleInFront.fCurrentSegment; ++SegmentNr) {

          LaneNr = calcLaneNrInNextSegment(LaneNr,SegmentNr,vehicle.fCurrentLink);

          // if we reached a merging area, bail-out
          if (LaneNr == -1) {
            return null;
          }
        }

        CurrentSegment = VehicleInFront.fCurrentSegment;
      }

      if (VehicleInFront.fCurrentLane == LaneNr) {
        Found = true;
      }
      else {
        // the vehicle in front is located more downstream
        // so it resides at the beginning of the queue
        --IndexOfVehicleInFront;
      }
    }

    if (Found) {
      return VehicleInFront;
    }

    // there was no vehicle in front found in the current link

    // if the current link is a terminator then there is no vehicle in front
    if (vehicle.fCurrentLink.fIsTerminator) {
      return null;
    }

    // calculate the lanenumber at the end of this link
    for (int SegmentNr = CurrentSegment;
         SegmentNr < ((vehicle.fCurrentLink.fNrOfNodes - 1) - 1); ++SegmentNr) {

      LaneNr = calcLaneNrInNextSegment(LaneNr,SegmentNr,vehicle.fCurrentLink);

      // if we reached a merging area, bail-out
      if (LaneNr == -1) {
        return null;
      }
    }

    // find out which link lies in front on the vehicle's current path
    // (formed by the extension of the current lane)

    Link OldLink = vehicle.fCurrentLink;
    Link NewLink = null;

    // only offramps have two outgoing links so find out which
    // of them lies on the vehicle's path
    if (OldLink.fType == Link.kSeparatedOffRamp) {

      if (OldLink.fLocation == Link.kLeft) {

        if (LaneNr < OldLink.fNrOfLanes) {
          NewLink = OldLink.fMainLinksNextLink;
        }
        else {
          NewLink = OldLink.fOffRampsNextLink;
        }
      }
      else { // the location of the offramp is right

        if (LaneNr < OldLink.fNrOfLanesOnRamp) {
          NewLink = OldLink.fOffRampsNextLink;
        }
        else {
          NewLink = OldLink.fMainLinksNextLink;
        }
      }
    }
    else { // all the other supported linktypes have only one outgoing link
      NewLink = OldLink.fMainLinksNextLink;
    }

    // calculate the lanenumber at the beginning of the next link
    LaneNr = calcLaneNrInNextLink(LaneNr,OldLink,NewLink);

    // search for the vehicle in front through the segments, starting at the first one

    if (NewLink.fVehicleQueue.size() == 0) {
      // the new link contains no vehicles
      return null;
    }

    // start searching at the first vehicle on the link (located last in the vehicle-queue)
    IndexOfVehicleInFront = NewLink.fVehicleQueue.size() - 1;
    VehicleInFront = (Vehicle) NewLink.fVehicleQueue.elementAt(IndexOfVehicleInFront);
    CurrentSegment = VehicleInFront.fCurrentSegment;

    Found = false;
    while ((!Found) && (IndexOfVehicleInFront >= 0)) {

      VehicleInFront = (Vehicle) NewLink.fVehicleQueue.elementAt(IndexOfVehicleInFront);

      // are we looking in another segment ?
      if (VehicleInFront.fCurrentSegment != CurrentSegment) {

        for (int SegmentNr = CurrentSegment;
             SegmentNr < VehicleInFront.fCurrentSegment; ++SegmentNr) {

          LaneNr = calcLaneNrInNextSegment(LaneNr,SegmentNr,NewLink);

          // if we reached a merging area, bail-out
          if (LaneNr == -1) {
            return null;
          }
        }

        CurrentSegment = VehicleInFront.fCurrentSegment;
      }

      if (VehicleInFront.fCurrentLane == LaneNr) {
        Found = true;
      }
      else {
        // the vehicle in front is located more downstream
        // so it resides at the beginning of the queue
        --IndexOfVehicleInFront;
      }
    }

    if (Found) {
      return VehicleInFront;
    }

    // no vehicle was found in the new link
    return null;
  }

  synchronized private int calcLaneNrInNextSegment(int LaneNr, int CurrentSegment, Link link)
  {
    int NrOfSegments = link.fNrOfNodes - 1;

    // are we in the last segment ?
    if (CurrentSegment == (NrOfSegments - 1)) {
      return LaneNr;
    }

    if ((link.fType == Link.kTunnel) ||
        (link.fType == Link.kHighway)) {
      return LaneNr;
    }
    else if (link.fType == Link.kMergedOnRamp) {

      if (CurrentSegment < (NrOfSegments - 2)) {
        // all the segments (except for the two last ones) are fully lane-connected
        return LaneNr;
      }
      else {
        // we are in the last but one segment (where the onramp merges)
        if (link.fLocation == Link.kLeft) {
          if (LaneNr < link.fNrOfLanes) {
            return LaneNr;
          }
          else {
            // we are in the onramp's merging area so there is no lane connected
            // to the current lane
            return -1;
          }
        }
        else { // location is right
          if (LaneNr < link.fNrOfLanesOnRamp) {
            // we are in the onramp's merging area so there is no lane connected
            // to the current lane
            return -1;
          }
          else {
            return (LaneNr - link.fNrOfLanesOnRamp);
          }
        }
      }
    }
    else if (link.fType == Link.kSeparatedOffRamp) {

      if (CurrentSegment > 0) {
        // all the segments (except for the first one) are fully lane-connected
        return LaneNr;
      }
      else {
        // we are in the first segment (where the offramp diverges)
        if (link.fLocation == Link.kLeft) {
          return LaneNr;
        }
        else { // location is right
          return (LaneNr + link.fNrOfLanesOnRamp);
        }
      }
    }

    // unsupported linktype !
    return -1;
  }

  synchronized private int calcLaneNrInNextLink(int LaneNr, Link OldLink, Link NewLink)
  {
    if ((OldLink.fType == Link.kTunnel) ||
        (OldLink.fType == Link.kHighway) ||
        (OldLink.fType == Link.kMergedOnRamp)) {

      if ((NewLink.fType == Link.kTunnel) ||
          (NewLink.fType == Link.kHighway) ||
          (NewLink.fType == Link.kSeparatedOffRamp)) {

        // the vehicle stays in the current lane
        return LaneNr;
      }
      else if (NewLink.fType == Link.kMergedOnRamp) {

        // is the previous link connected to this link's mainlink or onramp ?
        if (OldLink.fMainLinkConnectedToPart == Link.kConnectedToMainLink) {
          // the connection is made with this link's mainlink

          // where does the onramp lie ?
          if (NewLink.fLocation == Link.kLeft) {
            // the vehicle stays in the current lane
            return LaneNr;
          }
          else { // the location of the onramp is right
            return (LaneNr + NewLink.fNrOfLanesOnRamp);
          }
        }
        else {
          // the connection is made with this link's onramp

          // where does the onramp lie ?
          if (NewLink.fLocation == Link.kLeft) {
            return (LaneNr + NewLink.fNrOfLanes);
          }
          else { // the location of the onramp is right
            // the vehicle stays in the current lane
            return LaneNr;
          }
        }
      } // if this link is a merged onramp
    } // if old link's type is a tunnel, a highway or a merged onramp

    else if (OldLink.fType == Link.kSeparatedOffRamp) {

      if (OldLink.fLocation == Link.kLeft) {

        // is the vehicle located on the offramp or the mainlink
        if (LaneNr < OldLink.fNrOfLanes) {

          // the vehicle is located on the mainlink
          // the vehicle stays in the current lane
          return LaneNr;
        }
        else { // the vehicle is located on the offramp
          return (LaneNr - OldLink.fNrOfLanes);
        }
      }
      else { // the location of the offramp is right

        // is the vehicle located on the offramp or the mainlink
        if (LaneNr < OldLink.fNrOfLanesOnRamp) {

          // the vehicle is located on the offramp
          // the vehicle stays in the current lane
          return LaneNr;
        }
        else { // the vehicle is located on the mainlink
          return (LaneNr - OldLink.fNrOfLanesOnRamp);
        }
      }
    }

    // unsupported linktype !
    return -1;
  }

  synchronized private float calcSpaceHeadwayInFront(Vehicle vehicle, Vehicle VehicleInFront)
  {
    // failsafe
    if (VehicleInFront == null) {
      return 0.0f;
    }

    // do the vehicle and the vehicle in front belong to the same link ?
    if (vehicle.fCurrentLink.fID == VehicleInFront.fCurrentLink.fID) {

      // do the vehicle and the vehicle in front belong to the same segment ?
      if (VehicleInFront.fCurrentSegment == vehicle.fCurrentSegment) {
        return (VehicleInFront.fPositionOnCurrentSegment -
                VehicleInFront.fLength -
                vehicle.fPositionOnCurrentSegment);
      }
      else {
        float SpaceHeadway = vehicle.fCurrentLink.fLengths[vehicle.fCurrentSegment] -
                             vehicle.fPositionOnCurrentSegment;
        for (int SegmentNr = (vehicle.fCurrentSegment + 1);
             SegmentNr < VehicleInFront.fCurrentSegment; ++SegmentNr) {
          SpaceHeadway += vehicle.fCurrentLink.fLengths[SegmentNr];
        }
        return (SpaceHeadway + (VehicleInFront.fPositionOnCurrentSegment -
                                VehicleInFront.fLength));
      }
    }
    else { // the vehicle and the vehicle behind belong to different connected links

      float SpaceHeadway = vehicle.fCurrentLink.fLengths[vehicle.fCurrentSegment] -
                           vehicle.fPositionOnCurrentSegment;
      for (int SegmentNr = (vehicle.fCurrentSegment + 1);
           SegmentNr < (vehicle.fCurrentLink.fNrOfNodes - 1);
           ++SegmentNr) {
        SpaceHeadway += vehicle.fCurrentLink.fLengths[SegmentNr];
      }
      for (int SegmentNr = 0; SegmentNr < VehicleInFront.fCurrentSegment; ++SegmentNr) {
        SpaceHeadway += VehicleInFront.fCurrentLink.fLengths[SegmentNr];
      }
      return (SpaceHeadway + (VehicleInFront.fPositionOnCurrentSegment -
                              VehicleInFront.fLength));
    }
  }

  synchronized private float calcTimeHeadway(float SpaceHeadway, float Speed)
  {
    if (Speed != 0.0f) {
      return (SpaceHeadway / (Speed / 3.6f));
    }
    else {
      return 0.0f;
    }
  }

  synchronized private Vehicle findVehicleBehind(Vehicle vehicle, int LaneToSearchIn)
  {
    Vehicle VehicleBehind = null;

    int LaneNr = LaneToSearchIn;
    int CurrentSegment = vehicle.fCurrentSegment;

    // search in the current link
    int IndexOfSubjectVehicle = vehicle.fCurrentLink.fVehicleQueue.indexOf(vehicle);
    int IndexOfVehicleBehind = (IndexOfSubjectVehicle + 1);
    boolean Found = false;
    while ((!Found) && (IndexOfVehicleBehind < vehicle.fCurrentLink.fVehicleQueue.size())) {

      VehicleBehind = (Vehicle) vehicle.fCurrentLink.fVehicleQueue.elementAt(IndexOfVehicleBehind);

      // are we looking in another segment ?
      if (VehicleBehind.fCurrentSegment != CurrentSegment) {

        for (int SegmentNr = CurrentSegment;
             SegmentNr > VehicleBehind.fCurrentSegment; --SegmentNr) {

          LaneNr = calcLaneNrInPreviousSegment(LaneNr,SegmentNr,vehicle.fCurrentLink);

          // if we reached a diverging area, bail-out
          if (LaneNr == -1) {
            return null;
          }
        }

        CurrentSegment = VehicleBehind.fCurrentSegment;
      }

      if (VehicleBehind.fCurrentLane == LaneNr) {
        Found = true;
      }
      else {
        // the vehicle behind is located more upstream
        // so it resides at the end of the queue
        ++IndexOfVehicleBehind;
      }
    }

    if (Found) {
      return VehicleBehind;
    }

    // there was no vehicle behind found in the current link

    // if the current link is an entrance to the network then there is no vehicle behind
    // only tunnels and highways are allowed as network-entrances
    if ((vehicle.fCurrentLink.fType == Link.kTunnel) ||
        (vehicle.fCurrentLink.fType == Link.kHighway)) {
      if (vehicle.fCurrentLink.fMainLinksPreviousLink == null) {
        return null;
      }
    }

    // calculate the lanenumber at the beginning of this link
    for (int SegmentNr = CurrentSegment; SegmentNr > 0; --SegmentNr) {

      LaneNr = calcLaneNrInPreviousSegment(LaneNr,SegmentNr,vehicle.fCurrentLink);

      // if we reached a diverging area, bail-out
      if (LaneNr == -1) {
        return null;
      }
    }

    // find out which link lies behind on the vehicle's current path
    // (formed by the extension of the current lane)

    Link OldLink = vehicle.fCurrentLink;
    Link NewLink = null;

    // only onramps have two incoming links so find out which
    // of them lies on the vehicle's path
    if (OldLink.fType == Link.kMergedOnRamp) {

      if (OldLink.fLocation == Link.kLeft) {

        if (LaneNr < OldLink.fNrOfLanes) {
          NewLink = OldLink.fMainLinksPreviousLink;
        }
        else {
          NewLink = OldLink.fOnRampsPreviousLink;
        }
      }
      else { // the location of the onramp is right

        if (LaneNr < OldLink.fNrOfLanesOnRamp) {
          NewLink = OldLink.fOnRampsPreviousLink;
        }
        else {
          NewLink = OldLink.fMainLinksPreviousLink;
        }
      }
    }
    else { // all the other supported linktypes have only one incoming link
      NewLink = OldLink.fMainLinksPreviousLink;
    }

    // calculate the lanenumber at the end of the previous link
    LaneNr = calcLaneNrInPreviousLink(LaneNr,OldLink,NewLink);

    // search for the vehicle behind through the segments, starting at the last one

    if (NewLink.fVehicleQueue.size() == 0) {
      // the new link contains no vehicles
      return null;
    }

    // start searching at the last vehicle on the link (located first in the vehicle-queue)

    IndexOfVehicleBehind = 0;
    VehicleBehind = (Vehicle) NewLink.fVehicleQueue.elementAt(IndexOfVehicleBehind);
    CurrentSegment = VehicleBehind.fCurrentSegment;
    Found = false;
    while ((!Found) && (IndexOfVehicleBehind < NewLink.fVehicleQueue.size())) {

      VehicleBehind = (Vehicle) NewLink.fVehicleQueue.elementAt(IndexOfVehicleBehind);

      // are we looking in another segment ?
      if (VehicleBehind.fCurrentSegment != CurrentSegment) {

        for (int SegmentNr = CurrentSegment;
             SegmentNr > VehicleBehind.fCurrentSegment; --SegmentNr) {

          LaneNr = calcLaneNrInPreviousSegment(LaneNr,SegmentNr,NewLink);

          // if we reached a diverging area, bail-out
          if (LaneNr == -1) {
            return null;
          }
        }

        CurrentSegment = VehicleBehind.fCurrentSegment;
      }

      if (VehicleBehind.fCurrentLane == LaneNr) {
        Found = true;
      }
      else {
        // the vehicle behind is located more upstream
        // so it resides at the end of the queue
        ++IndexOfVehicleBehind;
      }
    }

    if (Found) {
      return VehicleBehind;
    }

    // no vehicle was found in the new link
    return null;
  }

  synchronized private int calcLaneNrInPreviousSegment(int LaneNr, int CurrentSegment, Link link)
  {
    int NrOfSegments = link.fNrOfNodes - 1;

    // are we in the first segment ?
    if (CurrentSegment == 0) {
      return LaneNr;
    }

    if ((link.fType == Link.kTunnel) ||
        (link.fType == Link.kHighway)) {
      return LaneNr;
    }
    else if (link.fType == Link.kMergedOnRamp) {

      if (CurrentSegment == (NrOfSegments - 1)) {

        // we are in the last segment (where the onramp merges)
        if (link.fLocation == Link.kLeft) {
          return LaneNr;
        }
        else { // location is right
          if (LaneNr < link.fNrOfLanesOnRamp) {
            return (LaneNr + link.fNrOfLanesOnRamp);
          }
        }
      }
      else {
        // all the other segments are fully lane-connected
        return LaneNr;
      }
    }
    else if (link.fType == Link.kSeparatedOffRamp) {

      if (CurrentSegment > 1) {
        // all the segments (except for the first two ones) are fully lane-connected
        return LaneNr;
      }
      else {
        // we are in the second segment
        if (link.fLocation == Link.kLeft) {
          if (LaneNr < link.fNrOfLanes) {
            return LaneNr;
          }
          else {
            //  the offramp diverges from the first segment
            return -1;
          }
        }
        else { // location is right
          if (LaneNr < link.fNrOfLanesOnRamp) {
            //  the offramp diverges from the first segment
            return -1;
          }
          else {
            return (LaneNr - link.fNrOfLanesOnRamp);
          }
        }
      }
    }

    // unsupported linktype !
    return -1;
  }

  synchronized private int calcLaneNrInPreviousLink(int LaneNr, Link OldLink, Link NewLink)
  {
    // ***************************************************************
    // ** Due to the complexity of lane-change operations and       **
    // ** vehicle transfers between links, I decided not to support **
    // ** all connections between links.                            **
    // **                                                           **
    // ** The following connections are not supported :             **
    // **                                                           **
    // **   offramp -> onramp                                       **
    // **   onramp  -> onramp                                       **
    // ***************************************************************

    if ((OldLink.fType == Link.kTunnel) ||
        (OldLink.fType == Link.kHighway) ||
        (OldLink.fType == Link.kSeparatedOffRamp)) {

      if ((NewLink.fType == Link.kTunnel) ||
          (NewLink.fType == Link.kHighway) ||
          (NewLink.fType == Link.kMergedOnRamp)) {

        return LaneNr;
      }
      else { // new link is a separated offramp

        if (OldLink.fID == NewLink.fMainLinksNextLink.fID) {
          // the connection is made between an offramp's mainlink and this link

          if (NewLink.fLocation == Link.kLeft) {
            return LaneNr;
          }
          else { // location is right
            return (LaneNr + NewLink.fNrOfLanesOnRamp);
          }
        }
        else { // the connection is made between an offramp's offramp and this link

          if (NewLink.fLocation == Link.kLeft) {
            return (LaneNr + NewLink.fNrOfLanes);
          }
          else { // the location of the offramp is right
            return LaneNr;
          }
        }
      }
    }
    else if (OldLink.fType == Link.kMergedOnRamp) {

      if (OldLink.fLocation == Link.kLeft) {

        if (LaneNr < OldLink.fNrOfLanes) {
          return LaneNr;
        }
        else {
          return (LaneNr - OldLink.fNrOfLanes);
        }
      }
      else { // the location of the onramp is right

        if (LaneNr < OldLink.fNrOfLanesOnRamp) {
          return LaneNr;
        }
        else {
          return (LaneNr - OldLink.fNrOfLanesOnRamp);
        }
      }
    }
    else {
      // unsupported linktype !
      return -1;
    }
  }

  synchronized private float calcSpaceHeadwayBehind(Vehicle vehicle, Vehicle VehicleBehind)
  {
    // failsafe
    if (VehicleBehind == null) {
      return 0.0f;
    }

    // do the vehicle and the vehicle behind belong to the same link ?
    if (vehicle.fCurrentLink.fID == VehicleBehind.fCurrentLink.fID) {

      if (VehicleBehind.fCurrentSegment == vehicle.fCurrentSegment) {
        return (vehicle.fPositionOnCurrentSegment -
                vehicle.fLength - 
                VehicleBehind.fPositionOnCurrentSegment);
      }
      else {
        float SpaceHeadwayBehind = vehicle.fPositionOnCurrentSegment - vehicle.fLength;
        for (int SegmentNr = (vehicle.fCurrentSegment - 1);
             SegmentNr > VehicleBehind.fCurrentSegment; --SegmentNr) {
          SpaceHeadwayBehind += vehicle.fCurrentLink.fLengths[SegmentNr];
        }
        return (SpaceHeadwayBehind +
                VehicleBehind.fCurrentLink.fLengths[VehicleBehind.fCurrentSegment] -
                VehicleBehind.fPositionOnCurrentSegment);
      }
    }
    else { // the vehicle and the vehicle behind belong to different connected links

      float SpaceHeadwayBehind = vehicle.fPositionOnCurrentSegment - vehicle.fLength;
      for (int SegmentNr = (vehicle.fCurrentSegment - 1); SegmentNr >= 0; --SegmentNr) {
        SpaceHeadwayBehind += vehicle.fCurrentLink.fLengths[SegmentNr];
      }
      for (int SegmentNr = (VehicleBehind.fCurrentLink.fNrOfNodes - 2);
           SegmentNr > VehicleBehind.fCurrentSegment; --SegmentNr) {
        SpaceHeadwayBehind += VehicleBehind.fCurrentLink.fLengths[SegmentNr];
      }
      return (SpaceHeadwayBehind +
              VehicleBehind.fCurrentLink.fLengths[VehicleBehind.fCurrentSegment] -
              VehicleBehind.fPositionOnCurrentSegment);
    }
  }
}
