// ----------------------------------
// Filename      : Vehicle.java
// Created       : 17/05/2000
// Last modified : 21/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.simulator;

import mitrasim.network.*;
import mitrasim.resources.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Vehicle
{
  // the two vehicle-types
  public static final int kCar   = 1;
  public static final int kTruck = 2;

  // the two possible mandatory locations on a link
  public static final int kOffRamp  = 1;
  public static final int kMainLink = 2;

  // the vehicle's characteristics
  public int    fType;
  public String fLicensePlate;
  public float  fLength;
  public float  fWidth;
  public float  fHeight;
  public float  fSpeed;
  public float  fCurrentAcceleration;
  public float  fNewAcceleration;
  public Driver fDriver;
  public float  fDistanceTraveled;

  // routing information and miscellaneous
  public RoutingTableEntry fRTE;
  public int               fRouteNr;
  public int               fIndexOfLastJunctionNodePassed;
  public Link              fCurrentLink;
  public int               fCurrentSegment;
  public float             fPositionOnCurrentSegment;
  public int               fCurrentLane;
  public boolean           fFreeFlowing;
  public float             fSpaceHeadway;
  public float             fTimeHeadway;
  public Vehicle           fVehicleInFront;
  public boolean           fDesiredSpeedReached;
  public int               fMandatoryLocationOnLink;
  public boolean           fMandatoryLocationReached;
  public int               fTargetLane;
  public float             fDistanceTraveledOnRamp;
  public Vehicle           fLeadVehicle;
  public Vehicle           fLagVehicle;
  public boolean           fBusyForcedMerging;
  public boolean           fYielding;
  public Vehicle           fVehicleToYieldTo;
  public int               fTimeLeftDuringOvertaking;
  public boolean           fBusyOvertaking;
  public boolean           fLeftBlinkerOn;
  public boolean           fRightBlinkerOn;

  // internal datastructure
  protected Resources fResources;

  public Vehicle()
  {
    fResources = new Resources();

    fType = kCar;
    fLicensePlate = "MV";
    fLength = 0.0f;
    fWidth = 0.0f;
    fHeight = 0.0f;
    fSpeed = 0.0f;
    fCurrentAcceleration = 0.0f;
    fNewAcceleration = 0.0f;
    fDriver = null;
    fDistanceTraveled = 0.0f;
    fRTE = null;
    fRouteNr = 0;
    // indicate that the first node (= origin) on the route has already been passed
    fIndexOfLastJunctionNodePassed = -1;
    fCurrentLink = null;
    fCurrentSegment = 0;
    fPositionOnCurrentSegment = 0.0f;
    fCurrentLane = 0;
    fFreeFlowing = true;
    fSpaceHeadway = 0.0f;
    fTimeHeadway = 0.0f;
    fVehicleInFront = null;
    fDesiredSpeedReached = false;
    fMandatoryLocationOnLink = kMainLink;
    fMandatoryLocationReached = true; // always true at a network-entrance !
    fTargetLane = 0;
    fDistanceTraveledOnRamp = 0.0f;
    fLeadVehicle = null;
    fLagVehicle = null;
    fBusyForcedMerging = false;
    fYielding = false;
    fVehicleToYieldTo = null;
    fTimeLeftDuringOvertaking = 0;
    fBusyOvertaking = false;
    fLeftBlinkerOn = false;
    fRightBlinkerOn = false;
  }

  public void cloneTo(Vehicle duplicate)
  {
    duplicate.fType = fType;
    duplicate.fLicensePlate = fLicensePlate;
    duplicate.fLength = fLength;
    duplicate.fWidth = fWidth;
    duplicate.fHeight = fHeight;
    duplicate.fSpeed = fSpeed;
    duplicate.fCurrentAcceleration = fCurrentAcceleration;
    duplicate.fNewAcceleration = fNewAcceleration;
    duplicate.fDriver = new Driver();
    fDriver.cloneTo(duplicate.fDriver);
    duplicate.fDistanceTraveled = fDistanceTraveled;
    duplicate.fRTE = fRTE;
    duplicate.fRouteNr = fRouteNr;
    duplicate.fIndexOfLastJunctionNodePassed = fIndexOfLastJunctionNodePassed;
    duplicate.fCurrentLink = fCurrentLink;
    duplicate.fCurrentSegment = fCurrentSegment;
    duplicate.fPositionOnCurrentSegment = fPositionOnCurrentSegment;
    duplicate.fCurrentLane = fCurrentLane;
    duplicate.fFreeFlowing = fFreeFlowing;
    duplicate.fSpaceHeadway = fSpaceHeadway;
    duplicate.fTimeHeadway = fTimeHeadway;
    duplicate.fVehicleInFront = fVehicleInFront;
    duplicate.fDesiredSpeedReached = fDesiredSpeedReached;
    duplicate.fMandatoryLocationOnLink = fMandatoryLocationOnLink;
    duplicate.fMandatoryLocationReached = fMandatoryLocationReached;
    duplicate.fTargetLane = fTargetLane;
    duplicate.fDistanceTraveledOnRamp = fDistanceTraveledOnRamp;
    duplicate.fLeadVehicle = fLeadVehicle;
    duplicate.fLagVehicle = fLagVehicle;
    duplicate.fBusyForcedMerging = fBusyForcedMerging;
    duplicate.fYielding = fYielding;
    duplicate.fVehicleToYieldTo = fVehicleToYieldTo;
    duplicate.fTimeLeftDuringOvertaking = fTimeLeftDuringOvertaking;
    duplicate.fBusyOvertaking = fBusyOvertaking;
    duplicate.fLeftBlinkerOn = fLeftBlinkerOn;
    duplicate.fRightBlinkerOn = fRightBlinkerOn;
  }

  public void setType(int Type)
  {
    fType = Type;

    if (fType == kCar) {
      fLength = fResources.ksimCarLength;
      fWidth = fResources.ksimCarWidth;
      fHeight = fResources.ksimCarLength;
    }
    else {
      fLength = fResources.ksimTruckLength;
      fWidth = fResources.ksimTruckWidth;
      fHeight = fResources.ksimTruckHeight;
    }
  }

  public float getMaximumSpeed(float Grade)
  {
    if (fType == kCar) {
      return getMaximumSpeedOfCar(Grade);
    }
    else {
      return getMaximumSpeedOfTruck(Grade);
    }
  }

  public float getMaximumAcceleration(float Grade)
  {
    return (getMaximumAccelerationOnLevelTerrain() -
            ((Grade * fResources.fsimGravitationalConstant) / 100.0f));
  }

  public float getMaximumAccelerationOnLevelTerrain()
  {
    if (fType == kCar) {
      return getMaximumAccelerationOfCarOnLevelTerrain();
    }
    else {
      return getMaximumAccelerationOfTruckOnLevelTerrain();
    }
  }

  public float getNormalDeceleration()
  {
    if (fType == kCar) {
      return getNormalDecelerationOfCar();
    }
    else {
      return getNormalDecelerationOfTruck();
    }
  }

  public float getMaximumDeceleration()
  {
    if (fType == kCar) {
      return getMaximumDecelerationOfCar();
    }
    else {
      return getMaximumDecelerationOfTruck();
    }
  }

  private float getMaximumSpeedOfCar(float Grade)
  {
    return 180.0f;
  }

  private float getMaximumAccelerationOfCarOnLevelTerrain()
  {
    if (fSpeed < 20.0f) {
      return 2.7f;
    }
    else if (fSpeed < 45.0f) {
      return 1.8f;
    }
    else if (fSpeed < 65.0f) {
      return 1.5f;
    }
    else if (fSpeed < 85.0f) {
      return 1.0f;
    }
    return 0.9f;
  }

  private float getNormalDecelerationOfCar()
  {
    if (fSpeed < 20.0f) {
      return -2.4f;
    }
    else if (fSpeed < 45.0f) {
      return -2.0f;
    }
    else if (fSpeed < 65.0f) {
      return -1.5f;
    }
    else if (fSpeed < 85.0f) {
      return -1.5f;
    }
    return -1.5f;
  }

  private float getMaximumDecelerationOfCar()
  {
    if (fSpeed < 20.0f) {
      return -3.0f;
    }
    else if (fSpeed < 45.0f) {
      return -2.9f;
    }
    else if (fSpeed < 65.0f) {
      return -2.7f;
    }
    else if (fSpeed < 85.0f) {
      return -2.6f;
    }
    return -2.5f;
  }

  private float getMaximumSpeedOfTruck(float Grade)
  {
    if (Grade < -2.0f) {
      return 140.0f;
    }
    else if (Grade < -1.0f) {
      return 130.0f;
    }
    else if (Grade < 0.0f) {
      return 120.0f;
    }
    else if (Grade < 1.0f) {
      return 110.0f;
    }
    else if (Grade < 2.0f) {
      return 80.0f;
    }
    else if (Grade < 4.0f) {
      return 70.0f;
    }
    return 50.0f;
  }

  private float getMaximumAccelerationOfTruckOnLevelTerrain()
  {
    if (fSpeed < 20.0f) {
      return 2.0f;
    }
    else if (fSpeed < 45.0f) {
      return 1.5f;
    }
    else if (fSpeed < 65.0f) {
      return 1.0f;
    }
    else if (fSpeed < 85.0f) {
      return 0.5f;
    }
    return 0.4f;
  }

  private float getNormalDecelerationOfTruck()
  {
    if (fSpeed < 20.0f) {
      return -1.5f;
    }
    else if (fSpeed < 45.0f) {
      return -1.0f;
    }
    else if (fSpeed < 65.0f) {
      return -0.7f;
    }
    else if (fSpeed < 85.0f) {
      return -0.6f;
    }
    return -0.5f;
  }

  private float getMaximumDecelerationOfTruck()
  {
    if (fSpeed < 20.0f) {
      return -2.0f;
    }
    else if (fSpeed < 45.0f) {
      return -1.5f;
    }
    else if (fSpeed < 65.0f) {
      return -1.0f;
    }
    else if (fSpeed < 85.0f) {
      return -0.7f;
    }
    return -0.6f;
  }

  public boolean isGreaterThan(Vehicle vehicle)
  {
    return ((fCurrentSegment > vehicle.fCurrentSegment) ||
            ((fCurrentSegment == vehicle.fCurrentSegment) &&
             (fPositionOnCurrentSegment > vehicle.fPositionOnCurrentSegment)) ||
            ((fCurrentSegment == vehicle.fCurrentSegment) &&
             (fPositionOnCurrentSegment == vehicle.fPositionOnCurrentSegment) &&
             (fCurrentLane > vehicle.fCurrentLane)));
  }
}
