// -------------------------------------
// Filename      : VehicleGenerator.java
// Created       : 17/05/2000
// Last modified : 04/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// -------------------------------------

package mitrasim.simulator;

import mitrasim.resources.*;
import mitrasim.simulator.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class VehicleGenerator
{
  // internal datastructures
  private Resources         fResources;
  private RoutingTableEntry fRTE;
  private int               fPreviousDepartureTime;
  private int               fCurrentInterDepartureTime;
  private int               fCurrentClocktick;
  private int[]             fPercentageMap;
  private int               fLastLaneUsed;

  public VehicleGenerator()
  {
    fResources = new Resources();
    fRTE = null;
    fPreviousDepartureTime = 0;
    fCurrentInterDepartureTime = 0;
    fCurrentClocktick = 0;
    fPercentageMap = null;
    fLastLaneUsed = 0;
  }

  public void setRoutingTableEntry(RoutingTableEntry RTE)
  {
    fRTE = RTE;

    fPercentageMap = new int[100];
    int LastMapIndex = 0;

    // create the map filled with percentages
    for (int i = 0; i < fRTE.fNrOfRoutes; ++i) {

      if (fRTE.fPercentages[i] != 0) {
        for (int j = 0; j < fRTE.fPercentages[i]; ++j) {
          fPercentageMap[j + LastMapIndex] = i; // store the routenr
        }
        LastMapIndex += fRTE.fPercentages[i] - 1;
      }
    }

    reset();
  }

  public void reset()
  {
    fPreviousDepartureTime = 0;
    fCurrentInterDepartureTime = 0;
    fLastLaneUsed = 0;
  }

  public void generate(int CurrentClocktick)
  {
    fCurrentClocktick = CurrentClocktick;

    // has the current inter-departure-time elapsed and is there
    // enough space on the track to accomodate a new vehicle ?
    if ((Math.abs(fCurrentClocktick - fPreviousDepartureTime) >= fCurrentInterDepartureTime) &&
        enoughSpaceOnLink()) {

      int fNewInterDepartureTime = 0;

      // choose between departing with constant headways or
      // according to a Poisson-distribution
      float distributionFactor = fResources.fsimRNG.nextFloat();

      if (distributionFactor < (fRTE.fDistributionFactor / 100.0f)) {
        // depart with constant headways

        // flowrate is expressed in vehicles/hour
        // translate it to tenths of seconds (= clockticks)
        fNewInterDepartureTime = (int) Math.ceil((1.0f / fRTE.fFlowRate) * 36000.0f);

      }
      else {
        // depart according to a Poisson-distribution

        float RandomNumber = 0.0f;
        while (RandomNumber == 0.0f) {
          RandomNumber = fResources.fsimRNG.nextFloat();
        }

        fNewInterDepartureTime = fCurrentInterDepartureTime -
                                 (int) Math.ceil((Math.log(RandomNumber) / fRTE.fFlowRate) *
                                                 36000.0f);
      }

      fCurrentInterDepartureTime = fNewInterDepartureTime;
      fPreviousDepartureTime = fCurrentClocktick;

      createAndPlaceVehicle();
    }
    else {
      // put vehicle to generate on hold
      // or simulate the behaviour of a queue when there is not enough space on the track
      return;
    }
  }

  private boolean enoughSpaceOnLink()
  {
    if (fRTE.fNetworkEntranceLink.fVehicleQueue.size() == 0) {
      // there are no vehicles on the link, so there is always enough space
      return true;
    }
    else {
      // examine the position and length of the most upstream vehicle on the link
      // this vehicle is located last in the vehicle queue
      Vehicle vehicle = (Vehicle) fRTE.fNetworkEntranceLink.fVehicleQueue.lastElement();

      // check if this vehicle is on the first segment
      if (vehicle.fCurrentSegment == 0) {

        // check if there is room for a new vehicle
        if ((vehicle.fPositionOnCurrentSegment - vehicle.fLength) > fResources.fsimMinSpaceAtEntrance) {
          return true;
        }
        else {
          return false;
        }
      }
      else {
        // the vehicle is on the next segment, so there should be enough room for a new vehicle
        return true;
      }
    }
  }

  private void createAndPlaceVehicle()
  {
    Vehicle vehicle = new Vehicle();

    // determine the type of the vehicle, based on the fleetmix
    int RandomInt = (int) (fResources.fsimRNG.nextFloat() * 100.0f);
    if (RandomInt < fRTE.fFleetMix) {
      vehicle.setType(Vehicle.kCar);
    }
    else {
      vehicle.setType(Vehicle.kTruck);
    }

    // set the licenseplate according to the current clocktime and the node-of-origin
    vehicle.fLicensePlate = "MV-" +
                            String.valueOf(fRTE.fOriginNode.fID) + "-" +
                            String.valueOf(fCurrentClocktick);

    // set an initial speed for this vehicle
    vehicle.fSpeed = fResources.fsimInitialVehicleSpeed;
    vehicle.fCurrentAcceleration = 0.0f;
    vehicle.fNewAcceleration = 0.0f;

    // create a driver for this vehicle
    vehicle.fDriver = new Driver();
    vehicle.fDriver.setCharacteristics();

    // the vehicle has not traveled any distance
    vehicle.fDistanceTraveled = 0.0f;

    // select a route that this vehicle must follow
    vehicle.fRTE = fRTE;
    vehicle.fRouteNr = fPercentageMap[(int) Math.floor(fResources.fsimRNG.nextFloat() * 100.0f)];
    vehicle.fIndexOfLastJunctionNodePassed = -1; // !

    // position the vehicle at the beginning of the first segment of this link
    vehicle.fCurrentLink = fRTE.fNetworkEntranceLink;
    vehicle.fCurrentSegment = 0;
    vehicle.fPositionOnCurrentSegment = 0.0f;

    // alternate the lanes in which the vehicles are placed
    vehicle.fCurrentLane = fLastLaneUsed;
    ++fLastLaneUsed;
    if (fLastLaneUsed >= fRTE.fNetworkEntranceLink.fNrOfLanes) {
      fLastLaneUsed = 0;
    }

    vehicle.fSpaceHeadway = 0.0f;
    vehicle.fTimeHeadway = 0.0f;
    vehicle.fVehicleInFront = null;

    vehicle.fDesiredSpeedReached = false;
    vehicle.fMandatoryLocationOnLink = Vehicle.kMainLink;
    vehicle.fMandatoryLocationReached = true;

    // place the vehicle on this link by
    // adding it at the end of the queue (so it will be processed LAST)
    fRTE.fNetworkEntranceLink.fVehicleQueue.addElement(vehicle);
  }
}
